//////////////////////////////////////////////////////////////////////
//   Module    : sx16core.c                                                                         
//
//   Function  : OS-independent core for model 516/616 drivers     
//
//   Author(s) : Sensoray Company Inc.  Copyright Sensoray Company 2003-2008
//   Revision :                                                                              
//   Copyright : (C) Sensoray 2003 - 2006      
//////////////////////////////////////////////////////////////////////
// sx16core MUST not contain any printk's, EVER.
// use dbgprint from the module file

#define I2C_LOCK
#define  SX16_KERNEL                            // For user/kernel shared headers.

#ifdef WIN32
#define INLINE __inline
#include "stddcls.h"
#define dbgprint( _x_ ) KdPrint( ( _x_ ))
#else
#define INLINE
#endif


#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
/* 2.4 kernel stuff */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#endif

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0)
#include <linux/vmalloc.h>
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
#include <linux/wait.h>
#endif
#include <linux/fs.h>
#include <linux/proc_fs.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
#include <linux/seq_file.h>
#endif
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/pagemap.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 17)
#include <linux/uaccess.h>
#include <linux/io.h>
#endif

#include "vidcap_link.h"
#include "sx16mod.h"                                                    // Wrappers for kernel functions.
#include "sx16core.h"                                                   // Declarations and definitions for the 516.
#include "font516.h"                                                    // 516 caption font.
#include "font616.h"                                                    // 616 caption font.
#ifndef WIN32   // sensoray write(playback) not implemented yet in WIN driver
//MPEG2 prefix for synchronization during playback
#include "hdr_m2m.h"
//MPEG1 CIF prefix for synchronization during playback
#include "hdr_m1m.h"
#endif


#define TRUE                1
#define FALSE               0

#ifndef WIN32
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/uaccess.h>
#endif

#define BPP_32  0x77 // magic number for BPP 1.5
#define DP dbgprint("sx16core:%d\n", __LINE__)

///////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// DEBUG CONSTANTS  //////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

                                                // Log debug messages to system console:
#define DEBUG_PACKET_INFO   0x00000001          //   SM packet info.
#define DEBUG_STREAM_CMD    0x00000002          //   Stream control command issued.
#define DEBUG_PARAM_UPLOAD  0x00000004          //   SM parameter upload failed.
#define DEBUG_UCODE_UPLOAD  0x00000008          //   SM microcode upload failed.
#define DEBUG_DEBI_ERROR    0x00000010          //   DEBI error (616 only).

#define DFLT_DEBUG_FLAGS    0                   // Default debug flags.

///////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////// BEHAVIORAL CONFIGURATION CONSTANTS  ////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////


// MANIFEST constants.  moved to sx16core.h


///////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  FORWARD FUNCTION DECLARATIONS  //////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

static void     sm_upload_params        ( PVPAR v );
static BOOL     sm_exec_cmd             ( PVPAR v, USHORT opcode, enum STREAM_STATE next_state, ULONG wait_msec );

static void     set_bass_treble         ( PVPAR v );
static void     set_mute                ( PVPAR v );
static void     set_volume              ( PVPAR v );
static void     set_brightness          ( PVPAR v );
static void     set_contrast            ( PVPAR v );
static void     set_saturation          ( PVPAR v );
static void     set_colorbars_enable    ( PVPAR v );

static void     set_video_in_516        ( PVPAR v );
static void     set_video_in_616        ( PVPAR v );

static void     set_audio_agc_ingain    ( PVPAR v );


// see comments below for WDM driver and capture code
static void     capture_init_params     ( PVPAR v );


///////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  CONSTANT PARAMETER ATTRIBUTES  /////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// These MUST be in same order as ATTR_ identifiers.

static ATTR_ITEM attr_516[NUM_ATTR_IDS] = {
    { set_bass_treble,      0,                      3,                      DFLT_BASS           },      // ATTR_audio_bass.
    { 0,                    0,                      14,                     DFLT_A_BITRATE      },      // ATTR_audio_bitrate.
    { 0,                    AFMT_MPEG1_LAYER1,      AFMT_AC3,               DFLT_AFORMAT        },      // ATTR_audio_format.
    { set_mute,             FALSE,                  TRUE,                   DFLT_MUTE           },      // ATTR_audio_mute.
    { 0,                    AUDIO_PCM_RES_16,       AUDIO_PCM_RES_24,       DFLT_PCM            },      // ATTR_audio_pcm.
    { 0,                    AUDIO_SAMPLERATE_48KHZ, AUDIO_SAMPLERATE_32KHZ, DFLT_ASR            },      // ATTR_audio_sample_rate.// 44.1 not supported for 516
    { 0,                    STEREOMODE_STEREO,      STEREOMODE_SINGLE,      DFLT_STEREO_MODE    },      // ATTR_audio_stereo_mode.
    { set_bass_treble,      0,                      3,                      DFLT_TREBLE         },      // ATTR_audio_treble.
    { set_volume,           0,                      63,                     DFLT_VOLUME         },      // ATTR_audio_volume.
    { 0,                    AVSEL_BOTH,             AVSEL_AUDIO_ONLY,       DFLT_AV_SELECT      },      // ATTR_av_select.
    { 0,                    BITRATE_MIN,            BITRATE_MAX,            DFLT_V_BITRATE      },      // ATTR_video_bitrate.
    { set_brightness,       0,                      255,                    DFLT_BRIGHTNESS     },      // ATTR_video_brightness.
    { set_video_in_516,     0,                      3,                      DFLT_CHAN           },      // ATTR_video_chan.
    { set_colorbars_enable, FALSE,                  TRUE,                   DFLT_COLORBARS      },      // ATTR_video_colorbars_enable.
    { set_contrast,         0,                      127,                    DFLT_CONTRAST       },      // ATTR_video_contrast.
    { 0,                    FALSE,                  TRUE,                   DFLT_FILTER         },      // ATTR_video_filter_enable.
    { 0,                    1,                      31,                     DFLT_GOP            },      // ATTR_video_m (GOP).
    { 0,                    1,                      3,                      DFLT_IP             },      // ATTR_video_n.
    { 0,                    SIZE704_2,              SIZE480_2_2,            DFLT_PICSIZE        },      // ATTR_video_picsize.
    { set_saturation,       0,                      127,                    DFLT_SATURATION     },      // ATTR_video_saturation.
    { set_video_in_516,     CVIDEO,                 SVIDEO,                 DFLT_INPUT_TYPE     },      // ATTR_video_type.
    { 0,                    SYS_PAL,                SYS_NTSC,               DFLT_VSYSTEM        },      // ATTR_video_system.
    { 0,                    SYSFMT_SVCD,            SYSFMT_OTHER,           DFLT_SYS_FORMAT     },      // ATTR_video_system_format.
    { 0,                    FALSE,                  TRUE,                   DFLT_VBR_MODE       },      // ATTR_video_vbr_enable.
    { 0,                    MUX_FORMAT_SVCD,        MUX_FORMAT_TRANSPORT,   DFLT_MUX_FORMAT     },      // ATTR_mux_format.
    { 0,                    0,                      255,                    DFLT_PROG_NUM       },      // ATTR_program_number.
    { 0,                    SIZE704_2,              SIZE480_2_2,            DFLT_PICSIZE        },      // ATTR_mpeg.
    { 0,                    STREAMTYPE_MUX,         STREAMTYPE_TRANSPORT,   DFLT_STREAMTYPE     },      // ATTR_stream_type.
    { set_audio_agc_ingain, 0,                      0x7f,                   DFLT_AUDIO_INAMPGAIN},      // ATTR_audio_inampgain.
    { set_audio_agc_ingain, 0,                      1,                      DFLT_AUDIO_AGC_ONOFF},      // ATTR_audio_agc_onoff.
    { set_audio_agc_ingain, 0,                      6,                      DFLT_AUDIO_AGC_SENS }       // ATTR_audio_agc_sens.
};

static ATTR_ITEM attr_616[NUM_ATTR_IDS] = {
    { set_bass_treble,      0,                      3,                      DFLT_BASS           },      // ATTR_audio_bass.
    { 0,                    0,                      14,                     DFLT_A_BITRATE      },      // ATTR_audio_bitrate.
    { 0,                    AFMT_MPEG1_LAYER1,      AFMT_AC3,               DFLT_AFORMAT        },      // ATTR_audio_format.
    { set_mute,             FALSE,                  TRUE,                   DFLT_MUTE           },      // ATTR_audio_mute.
    { 0,                    AUDIO_PCM_RES_16,       AUDIO_PCM_RES_24,       DFLT_PCM            },      // ATTR_audio_pcm.
    { 0,                    AUDIO_SAMPLERATE_44KHZ, AUDIO_SAMPLERATE_32KHZ, DFLT_ASR            },      // ATTR_audio_sample_rate.
    { 0,                    STEREOMODE_STEREO,      STEREOMODE_SINGLE,      DFLT_STEREO_MODE    },      // ATTR_audio_stereo_mode.
    { set_bass_treble,      0,                      3,                      DFLT_TREBLE         },      // ATTR_audio_treble.
    { set_volume,           0,                      63,                     DFLT_VOLUME         },      // ATTR_audio_volume.
    { 0,                    AVSEL_BOTH,             AVSEL_AUDIO_ONLY,       DFLT_AV_SELECT      },      // ATTR_av_select.
    { 0,                    BITRATE_MIN,            BITRATE_MAX,            DFLT_V_BITRATE      },      // ATTR_video_bitrate.
    { set_brightness,       0,                      255,                    DFLT_BRIGHTNESS     },      // ATTR_video_brightness.
    { set_video_in_616,     0,                      0,                      DFLT_CHAN           },      // ATTR_video_chan.
    { set_colorbars_enable, FALSE,                  TRUE,                   DFLT_COLORBARS      },      // ATTR_video_colorbars_enable.
    { set_contrast,         0,                      127,                    DFLT_CONTRAST       },      // ATTR_video_contrast.
    { 0,                    FALSE,                  TRUE,                   DFLT_FILTER         },      // ATTR_video_filter_enable.
    { 0,                    1,                      31,                     DFLT_GOP            },      // ATTR_video_m (GOP).
    { 0,                    1,                      3,                      DFLT_IP             },      // ATTR_video_n.
    { 0,                    SIZE704_2,              SIZE640_2,              DFLT_PICSIZE        },      // ATTR_video_picsize.
    { set_saturation,       0,                      127,                    DFLT_SATURATION     },      // ATTR_video_saturation.
    { set_video_in_616,     CVIDEO,                 SVIDEO,                 DFLT_INPUT_TYPE     },      // ATTR_video_type.
    { 0,                    SYS_PAL,                SYS_NTSC,               DFLT_VSYSTEM        },      // ATTR_video_system.
    { 0,                    SYSFMT_SVCD,            SYSFMT_OTHER,           DFLT_SYS_FORMAT     },      // ATTR_video_system_format.
    { 0,                    FALSE,                  TRUE,                   DFLT_VBR_MODE       },      // ATTR_video_vbr_enable.
    { 0,                    MUX_FORMAT_SVCD,        MUX_FORMAT_TRANSPORT,   DFLT_MUX_FORMAT     },      // ATTR_mux_format.
    { 0,                    0,                      255,                    DFLT_PROG_NUM       },      // ATTR_program_number.
    { 0,                    SIZE704_2,              SIZE640_2,              DFLT_PICSIZE        },      // ATTR_mpeg.
    { 0,                    STREAMTYPE_MUX,         STREAMTYPE_TRANSPORT,   DFLT_STREAMTYPE     },      // ATTR_stream_type.
    { set_audio_agc_ingain, 0,                      0x7f,                   DFLT_AUDIO_INAMPGAIN},      // ATTR_audio_inampgain.
    { set_audio_agc_ingain, 0,                      1,                      DFLT_AUDIO_AGC_ONOFF},      // ATTR_audio_agc_onoff.
    { set_audio_agc_ingain, 0,                      6,                      DFLT_AUDIO_AGC_SENS }       // ATTR_audio_agc_sens.
};


///////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  CONSTANT CAPTION PARAMETERS   //////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// These MUST be in same order as CAPTN_PARAM identifiers.

static CAPTN_PARAM captn_prm_516[NUM_CAPTN_IDS] = {
    { 1,                    2,                      DFLT_CAPTN_WINDOW   },      // CAPTN_window.
    { 0,                    MAX_XSTART,             DFLT_CAPTN_XSTART   },      // CAPTN_xstart.
    { 0,                    MAX_XSTOP,              DFLT_CAPTN_XSTOP    },      // CAPTN_xstop.
    { 0,                    MAX_YSTART_PAL,         DFLT_CAPTN_YSTART   },      // CAPTN_ystart.
    { 0,                    MAX_YSTOP_PAL,          DFLT_CAPTN_YSTOP    },      // CAPTN_ystop.
    { OVL_SCALE_X1,         OVL_SCALE_X4,           DFLT_CAPTN_SCALE    },      // CAPTN_scale.
    { OVL_WHITE,            OVL_MAGENTA,            DFLT_CAPTN_COLOR    },      // CAPTN_color.
    { OVL_TRANSP_T,         OVL_TRANSP_B_BKGRD,     DFLT_CAPTN_TRANSP   },      // CAPTN_transp.
    { '\0',                 '\0',                   '\0'                },      // CAPTN_text.
    { CAPTION_OFF,          CAPTION_ON,             DFLT_CAPTN_ON       }       // CAPTN_on.
};

static CAPTN_PARAM captn_prm_616[NUM_CAPTN_IDS] = {
    { 1,                    2,                      DFLT_CAPTN_WINDOW   },      // CAPTN_window.
    { 0,                    MAX_XSTART,             DFLT_CAPTN_XSTART   },      // CAPTN_xstart.
    { 0,                    MAX_XSTOP,              DFLT_CAPTN_XSTOP    },      // CAPTN_xstop.
    { 0,                    MAX_YSTART_PAL,         DFLT_CAPTN_YSTART   },      // CAPTN_ystart.
    { 0,                    MAX_YSTOP_PAL,          DFLT_CAPTN_YSTOP    },      // CAPTN_ystop.
    { OVL_SCALE_X1,         OVL_SCALE_X1,           DFLT_CAPTN_SCALE    },      // CAPTN_scale.
    { OVL_WHITE,            OVL_MAGENTA,            DFLT_CAPTN_COLOR    },      // CAPTN_color.
    { OVL_TRANSP_T,         OVL_TRANSP_B_BKGRD,     DFLT_CAPTN_TRANSP   },      // CAPTN_transp.
    { '\0',                 '\0',                   '\0'                },      // CAPTN_text.
    { CAPTION_OFF,          CAPTION_ON,             DFLT_CAPTN_ON       }       // CAPTN_on.
};


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////  UTILITY FUNCTIONS  ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////
// emulate memcpy() function implementation in core.
static INLINE void cmemcpy( void *dst, void *src, int len ) 
{
    int i;
    char *pdst;
    char *psrc;
    
    pdst = (char *) dst;
    psrc = (char *) src;
    for ( i = 0; i < len; i++)
    {
        *pdst = *psrc;
        pdst++;
        psrc++;
    }
}

/////////////////////////////////////////////////////////////
// emulate strlen() function implementation in core.
static INLINE int cstrlen( char *string )
{
    int i;
    for ( i = 0; *string != '\0'; i++, string++ );  //lint !e722
    return i;
}

/////////////////////////////////////////////////////////////
// Place a one digit number in a string
static INLINE void num2txt( int num, char *txt )
{
    if ( num < 0 || num > 9 )
        return;

    txt[0] = (char)num + 0x30;
    txt[1] = 0;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////  REGISTER I/O FUNCTIONS  ////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////
// 516: Set address used for indirect register addressing.

static INLINE void ga_set_indirect_adrs( PVPAR v, int addr )
{
    // Assume we are indirectly writing to a register.  If the register address differs from that of the previous access ...
    if ( addr != v->lastaddr )
    {
        // Write the target register's address to the gate array's Register Pointer register.
        SX16MOD_write_io16( v->base_adrs + G_INDEX, (unsigned short)addr );

        // Remember the address so we might be able to save some time during the next access.
        v->lastaddr = addr;
    }
}

//////////////////////////////////////////////////////////////////
// 516: Read a word value from an indirectly accessed register.

static INLINE USHORT ga_indirect_reg_read( PVPAR v, int addr )
{
    // Set up register pointer.
    ga_set_indirect_adrs( v, addr );

    // Read value from target register via the Register Pointer register.
    return SX16MOD_read_io16( v->base_adrs + G_DATAIO );
}

/////////////////////////////////////////////////////////////////
// 516: Write a word value to an indirectly accessed register.

static INLINE void ga_indirect_reg_write( PVPAR v, int addr, USHORT data )
{
    // Set up Index register with target register address.
    ga_set_indirect_adrs( v, addr );

    // Write value to target register indirectly via Index.

    SX16MOD_write_io16( v->base_adrs + G_DATAIO, data );
}

////////////////////////////////////////////////////
// 516: Modify an indirectly accessed register.
// Ones in mask specify bits to be retained.

static INLINE void ga_indirect_reg_modify( PVPAR v, int addr, USHORT data, USHORT mask )
{
    // Set up Index register with target register address.
    ga_set_indirect_adrs( v, addr );

    // Write value to target register indirectly via Index.
    SX16MOD_write_io16( v->base_adrs + G_DATAIO, (USHORT)( ( SX16MOD_read_io16( v->base_adrs + G_DATAIO ) & mask ) | data ) );
}

//////////////////////////////////////////////////////////////////
// 616: Execute DEBI transaction.

static int transact_debi_616( PVPAR v, int debi_cmd )
{
    int i;
#ifdef TEST_PCI_616
    return 0;
#endif
    
    // Abort if DEBI timed out.
    if ( v->errflags & ERR_DEBI_TIMEOUT )
        return -1;

    // Prepare DEBI command.
    SX16MOD_write_mem32( v->base_adrs + P_DEBICMD, debi_cmd );

    // Invoke DEBI command upload.
    SX16MOD_write_mem32( v->base_adrs + P_MC2, MC2_DEBI_UPLD_CMD );

    // Wait for the command upload to finish.
    for ( i = 1000; ( SX16MOD_read_mem32( v->base_adrs + P_MC2 ) & MC2_DEBI_UPLD_DONE ) == 0; )
    {
        if ( --i == 0 )
        {
            if ( v->debug_flags & DEBUG_DEBI_ERROR )
                SX16MOD_text_message( "debi upload err\n" );

            return i;
        }
    }

    // Wait for the DEBI command to execute.
    for ( i = 10000; ( SX16MOD_read_mem32( v->base_adrs + P_PSR ) & PSR_DEBI_BUSY ) != 0; )
    {
        if ( --i == 0 )
        {
            if ( v->debug_flags & DEBUG_DEBI_ERROR )
                SX16MOD_text_message( "debi timeout err\n" );

            v->errflags |= ERR_DEBI_TIMEOUT;
            return i;
        }
    }
    return i;
}

///////////////////////////////////////////////////////////////////
// Write data block to SM register.

// 516: Use GA indirect register access interface.
static void sm_reg_block_write_516( PVPAR v, int reg_adrs, USHORT *pdata, int nbytes )
{
    // Set up register pointer.
    ga_set_indirect_adrs( v, SMBASE_516 + reg_adrs );

    // Write word values from data[] to target register via the Register Pointer register.
    SX16MOD_block_write_io16( v->base_adrs + G_DATAIO, pdata, nbytes >> 1 );
}

// 616: Use SAA7146 DEBI interface.
static void sm_reg_block_write_616( PVPAR v, int reg_adrs, USHORT *pdata, int nbytes )
{
    // If we are transferring 4 bytes or less, set up for immediate transfer, otherwise set up for block transfer.
    switch ( nbytes )
    {
        case 2:
            SX16MOD_write_mem32( v->base_adrs + P_DEBIAD, *pdata );
            break;

        case 4:
            SX16MOD_write_mem32( v->base_adrs + P_DEBIAD, *( (ULONG *)pdata ) );
            break;

        default:
            cmemcpy( v->dmabuf_virt, pdata, nbytes );
            SX16MOD_write_mem32( v->base_adrs + P_DEBIAD, v->dmabuf_phys );
            break;
    }
    // Execute the DEBI transfer to the SM.
    transact_debi_616( v, DEBICMD_WRITE( reg_adrs, nbytes ) );
}

static void sm_reg_block_write_616_mpeg( PVPAR v, int reg_adrs, USHORT *pdata, int nbytes )
{
    int rc = 0;
    if (nbytes <= 4) {
        sm_reg_block_write_616(v, reg_adrs, pdata, nbytes);
		transact_debi_616( v, DEBICMD_WRITE( reg_adrs, nbytes ) );
        return;
    }
#ifdef WIN32
    cmemcpy( v->dmabuf_virt, pdata, nbytes );
#else
    rc = copy_from_user( v->dmabuf_virt, pdata, nbytes );
    if (rc != 0) {
        printk(KERN_INFO "failed copy from user\n");
    }
#endif
    SX16MOD_write_mem32( v->base_adrs + P_DEBIAD, v->dmabuf_phys );
    // Execute the DEBI transfer to the SM.
    transact_debi_616( v, DEBICMD_WRITE( reg_adrs, nbytes ) );
}

////////////////////////////////////////////////////////////////////
// Read data block from SM register.

// 616: Use SAA7146 DEBI interface.
static void sm_reg_block_read_616( PVPAR v, int reg_adrs, USHORT *pdata, int nbytes )
{
    int rc = 0;
    // If we are transferring 4 bytes or less, execute immediate transfer, otherwise execute a block transfer.
    switch ( nbytes )
    {
        case 2:
            transact_debi_616( v, DEBICMD_READ( reg_adrs, 2 ) );                            // Execute immediate DEBI read from SM.
            *pdata = (USHORT)SX16MOD_read_mem32( v->base_adrs + P_DEBIAD );                 // Copy retrieved DEBI data to target buffer.
            break;

        case 4:
            transact_debi_616( v, DEBICMD_READ( reg_adrs, 4 ) );                            // Execute immediate DEBI read from SM.
            *( (ULONG *)pdata ) = (USHORT)SX16MOD_read_mem32( v->base_adrs + P_DEBIAD );    // Copy retrieved DEBI data to target buffer.
            break;

        default:
            SX16MOD_write_mem32( v->base_adrs + P_DEBIAD, v->dmabuf_phys );             // Set up block transfer address.
            transact_debi_616( v, DEBICMD_READ( reg_adrs, nbytes ) );                       // Execute DEBI block read from SM to target buffer.
#ifdef WIN32
            cmemcpy( pdata, v->dmabuf_virt, nbytes );
#else
            rc = copy_to_user( pdata, v->dmabuf_virt, nbytes );
            if (rc != 0) {
                printk(KERN_INFO "failed copy to user\n");
            }
#endif
            break;
    }
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////  INTERRUPT FUNCTIONS  ////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////
// 516: Enable 516 SM interrupt.

static void enable_irq_516( PVPAR v )
{
    // Disable all SM interrupts
    v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

    // Enable the 516's SM interrupt.
    ga_indirect_reg_write( v, G_IRQENABLE, GI_SMINT );

    // Route the 516's interrupt request pin to the specified IRQ line.
    ga_indirect_reg_write( v, G_IRQSELECT, (USHORT)v->irq );
}

/////////////////////////////////////////////////////////////
// 516: Disable 516 SM interrupt.

static void disable_irq_516( PVPAR v )
{
    // Disable all SM interrupts
    v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

    // Enable the 516's SM interrupt.
    ga_indirect_reg_write( v, G_IRQENABLE, 0 );
}


/////////////////////////////////////////////////////////////
// 516: Shutdown 516 SM interrupt.

static void shutdown_irq_516( PVPAR v )
{
    // Unoute the 516's interrupt request pin.
    ga_indirect_reg_write( v, G_IRQSELECT, 0 );
}


/////////////////////////////////////////////////////////////
// 616: Enable 616 SM interrupts.

static void enable_irq_616( PVPAR v )
{
    
    // Disable all SM interrupts: will be enabled later.
    v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

    // Enable SAA7146 interrupts. The SM interrupt is routed to GPIO0.
    SX16MOD_write_mem32( v->base_adrs + P_INTEN, ISR_GPIO0 | ISR_RPS_I0 );

    //dbgprint("IRQ enabled in core\n");
}

/////////////////////////////////////////////////////////////
// 616: Disable 616 SM interrupt.

static void disable_irq_616( PVPAR v )
{
    // Disable all SM interrupts
    v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

    // Disable SAA7146 GPIO0 interrupt. The SM interrupt is routed to
    // GPIO0
    SX16MOD_write_mem32( v->base_adrs + P_INTEN, 0 );
}

/////////////////////////////////////////////////////////////
// 616: Shutdown 616 SM interrupt.

static void shutdown_irq_616( PVPAR v )
{
    // Nothing to be done...
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////  QUEUE STATUS  ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Return the level of the specified MPEG queue in bytes.  Q_id: audio=1, mux=2, or video=3.

static ULONG get_Q_level( PVPAR v, int Q_id )
{
    switch ( Q_id )
    {
        case Q_ID_AUD:  return   (ULONG)v->sm_reg_read( v, SM_AQSIZE_REG );                                                             // Return audio queue size.
        case Q_ID_MUX:  return ( (ULONG)v->sm_reg_read( v, SM_MQSIZE_MSW_REG ) << 16 ) + (ULONG)v->sm_reg_read( v, SM_MQSIZE_LSW_REG ); // Return multiplex queue size.
        case Q_ID_VID:  return ( (ULONG)v->sm_reg_read( v, SM_VQSIZE_MSW_REG ) << 16 ) + (ULONG)v->sm_reg_read( v, SM_VQSIZE_LSW_REG ); // Return video queue size.
        default:        return 0;
    }
}

/////////////////////////////////////////////////////////////////////////////////
// Return the size of the data contained by any MPEG queue, in bytes.

ULONG CORE_queue_get_any_size( HBD board, int qid )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    ULONG level;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return 0;
    level =  get_Q_level( (PVPAR)board, qid );
    SX16MOD_release_mutex( v->mutex_object );

    return level;
}

/////////////////////////////////////////////////////////////////////////////////
// Return the number of bytes left in the dma cycle.

ULONG CORE_get_dma_remaining( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;

    return DFLT_DMA_SIZE - v->num_dma_bytes;
}

/////////////////////////////////////////////////////////////////////////////////
// Return the encoder/decoder elapsed time.

UINT64 CORE_get_elapsed_time( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;

    return SX16MOD_time_now() - v->stream_start_time;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////  I2C INTERFACE  /////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////
// Determine whether I2C transaction is in progress.

// 516: I2C controller in GA.
static BOOL i2c_is_busy_516( PVPAR v )
{
    return ( ( ga_indirect_reg_read( v, G_I2CCONTROL ) & I2C_BUSY_516 ) != 0 );
}

// 616: I2C controller in SAA7146.
static BOOL i2c_is_busy_616( PVPAR v )
{
    return ( ( SX16MOD_read_mem32( v->base_adrs + P_I2CCTRL ) & I2C_BUSY_616 ) != 0 );
}

////////////////////////////////////////////////////////////////////////////////////
// Wait for I2C interface idle.  Returns TRUE if idle within timeout interval,
// otherwise return FALSE and set I2C error flag.

#define I2C_TIMEOUT     200         // I2C transaction timeout in milliseconds.

static BOOL poll_I2C_until_idle( PVPAR v )
{
    // Start software timer.
    UINT64 t_start = SX16MOD_time_now();
    // Retry until timeout ...
    do
    {
        // Exit function if I2C transaction is finished.
        if ( !v->i2c_is_busy( v ) )
            return TRUE;
    }
    while ( SX16MOD_time_now() - t_start < (UINT64)I2C_TIMEOUT );
    
    // It's possible that we experienced a lengthly interruption after detecting I2C busy, and the I2C is no
    // longer busy.  To cover this possibility, we must check the I2C one more time to be sure it's still busy.
    return !v->i2c_is_busy( v );
}

static BOOL wait_idle_i2c( PVPAR v )
{
    BOOL idle = FALSE;  // Init return value to FALSE in case earlier I2C fault was detected.
#ifdef TEST_PCI_616
    return TRUE;
#endif
    // If no previous I2C fault was detected ...
    if ( ( v->errflags & ERR_I2C ) == 0 )
    {

        // Wait for I2C idle, or timeout, whichever comes first.  Log I2C error if we timed out.
        if ( ( idle = poll_I2C_until_idle( v ) ) == FALSE )
            v->errflags |= ERR_I2C;
    }

    // Indicate whether I2C is idle.
    return idle;
}

////////////////////////////////////////////////////////////////////////////////////
// 616: Execute I2C transaction.

static void transact_i2c_616( PVPAR v, int ctrlval )
{
#ifdef TEST_PCI_616
	return;
#endif

	if (SX16MOD_acquire_mutex( v->i2c_lock ) < 0)
		return;


        // If I2C interface is ready within max permitted time ...
        if (wait_idle_i2c( v )) {
            // Set up for transfer on SAA7146's I2C interface and invoke command upload.
            SX16MOD_write_mem32( v->base_adrs + P_I2CCTRL, ctrlval );
            SX16MOD_write_mem32( v->base_adrs + P_MC2, MC2_I2C_UPLD_CMD );

            // Wait for upload to execute.
            while ( ( SX16MOD_read_mem32( v->base_adrs + P_MC2 ) & MC2_I2C_UPLD_DONE ) == 0 );

            // Wait for transaction to finish or error to be detected.
            wait_idle_i2c( v );
        }
    SX16MOD_release_mutex( v->i2c_lock );
}

////////////////////////////////////////////////////////////////////////////////////////
// Write value to I2C device.

// 516: I2C interface in GA.
static void i2c_write_516( PVPAR v, int devaddr, int subaddr, int bdata )
{
#ifdef TEST_PCI_616
	return;
#endif
	
#ifdef I2C_LOCK
    // If I2C interface is ready within max permitted time ...
    if ( SX16MOD_acquire_mutex( v->i2c_lock ) < 0 )
        return;
#endif

    if ( wait_idle_i2c( v ) )
    {
        // Invoke transfer on GA's I2C interface.
        ga_indirect_reg_write( v, G_I2CDEVICEADDR, (USHORT)devaddr );
        ga_indirect_reg_write( v, G_I2CSUBADDR,    (USHORT)subaddr );
        ga_indirect_reg_write( v, G_I2CDATA,       (USHORT)bdata   );
        ga_indirect_reg_write( v, G_I2CCONTROL,    I2C_START       );

        // Wait until I2C transaction finishes or I2C fault is detected.
        wait_idle_i2c( v );
    }
#ifdef I2C_LOCK
    SX16MOD_release_mutex( v->i2c_lock );
#endif
}

// 616: I2C interface in SAA7146.
static void i2c_write_616( PVPAR v, int devaddr, int subaddr, int bdata )
{
#ifdef TEST_PCI_616
	return;
#endif
    //if( devaddr >= 0x70 && devaddr <= 0x7f ) dbgprint("Write_i2c: %x %x %x\n", devaddr, subaddr, bdata);
    // Execute I2C write transaction.
    transact_i2c_616( v, I2C_WR_616 | (devaddr << 24) | (subaddr << 16) | (bdata << 8) | (I2C_ATTRSTART << 6) | (I2C_ATTRCONT << 4) | (I2C_ATTRSTOP << 2) );
    if( v->errflags & ERR_I2C )
    {
        //dbgprint("Write_i2c error: %x %x %x\n", devaddr, subaddr, bdata);
        v->errflags &= ~ERR_I2C;
    }
}

////////////////////////////////////////////////////////////////////////////////////
// Read value from I2C device.  Sets error flag if I2C fault detected.

// 516: I2C interface in GA.
static int i2c_read_516( PVPAR v, int devaddr, int subaddr )
{
#ifdef TEST_PCI_616
	return 0;
#endif

#ifdef I2C_LOCK
    if ( SX16MOD_acquire_mutex( v->i2c_lock ) < 0 )
        return -1;
#endif

    // If I2C interface is ready within max permitted time ...
    if ( wait_idle_i2c( v ) )
    {
        // Invoke transfer on GA's I2C interface.
        ga_indirect_reg_write( v, G_I2CDEVICEADDR, (USHORT)(devaddr | 1));
        ga_indirect_reg_write( v, G_I2CSUBADDR,    (USHORT)subaddr );
        ga_indirect_reg_write( v, G_I2CCONTROL,    I2C_START       );

        // Wait until I2C transaction finishes or I2C fault is detected.
        if( wait_idle_i2c( v ) )
        {
#ifdef I2C_LOCK
            SX16MOD_release_mutex( v->i2c_lock );
#endif
            return ga_indirect_reg_read( v, G_I2CDATA );
        }
    
}
#ifdef I2C_LOCK
    SX16MOD_release_mutex( v->i2c_lock );
#endif

    ///return ERR_I2C;
    return -1;
}

// 616: I2C interface in SAA7146.
static int i2c_read_616( PVPAR v, int devaddr, int subaddr )
{
#ifdef TEST_PCI_616
	return 0;
#endif
	// Send Read command to the device.
    transact_i2c_616( v, I2C_WR_616 | (devaddr << 24) | (subaddr << 16) | (I2C_ATTRSTART << 6) | (I2C_ATTRSTOP << 4) | (I2C_ATTRNOP << 2) );

    // Receive data from device.
    transact_i2c_616( v, I2C_RD_616 | (devaddr << 24) | (I2C_ATTRSTART << 6) | (I2C_ATTRSTOP << 4) | (I2C_ATTRNOP << 2) );

  ///   // Return the data that was received from the I2C device, or zero if I2C fault was detected.
  ///   return ( v->errflags & ERR_I2C ) ? 0 : ( ( SX16MOD_read_mem32( v->base_adrs + P_I2CCTRL ) >> 16 ) & 0xFF );

    // Return the data that was received from the I2C device, or -1 if I2C fault was detected.
    return ( v->errflags & ERR_I2C ) ? -1 : ( ( SX16MOD_read_mem32( v->base_adrs + P_I2CCTRL ) >> 16 ) & 0xFF );
}

///////////////////////////////////////////////////////////////
// Upload parameter to video decoder.

static void dset( PVPAR v, ULONG address, ULONG data )
{
    v->i2c_write( v, v->i2caddr_dec, address, data );
}

///////////////////////////////////////////////////////////////
// Upload parameter to video encoder.

static void eset( PVPAR v, ULONG address, ULONG data )
{
    v->i2c_write( v, v->i2caddr_enc, address, data );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  ENCODER & DECODER CONFIGURATION  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define IS_NTSC     ( v->attr_current[ATTR_video_system] == SYS_NTSC )
#define IS_FILT     ( v->attr_current[ATTR_video_filter_enable] != 0 )

///////////////////////////////////////////////////////////////////////////////////////////////////////
// Write all runtime parameters to video encoder (SAA7120/SAA7121).

static void enc_config_516( PVPAR v )
{
    int i;

    for ( i = 0x00; i <= 0x25; i++ )                                        // 0x00 to 0x25: "Null" registers.
        eset( v, i, 0x00 );

    eset( v, E_WIDE_SCREEN_LSB,     0x00                            );      // 0x26: Wide screen signalling output disabled.
    eset( v, E_WIDE_SCREEN_MSB,     0x00                            );      // 0x27: Wide screen signal.
    eset( v, E_RTC_BURSTSTART,      IS_NTSC ? 0x19 : 0x21           );      // 0x28: Starting point of burst in clock cycles.
    eset( v, E_BURST_END,           IS_NTSC ? 0x1D : 0x1D           );      // 0x29: Ending point of burst in clock cycles.
    eset( v, E_COPYGUARD_ODD_0,     0x00                            );      // 0x2A
    eset( v, E_COPYGUARD_ODD_1,     0x00                            );      // 0x2B
    eset( v, E_COPYGUARD_EVEN_0,    0x00                            );      // 0x2C
    eset( v, E_COPYGUARD_EVEN_1,    0x00                            );      // 0x2D
    eset( v, E_COPYGUARD_ENABLE,    0x00                            );      // 0x2E: Copy guard encoding disabled.

    for ( i = 0x2F; i <= 0x39; i++ )                                        // 0x2F to 0x39: "Null" registers.
        eset( v, i, 0x00 );

    set_colorbars_enable( v );                                              // 0x3A: Colorbars control.

    for ( i = 0x3B; i <= 0x59; i++ )                                        // 0x3B to 0x59: Functions not specified, but other drivers write zeros to these regs.
        eset( v, i, 0x00 );

    eset( v, E_CHROMA_PHASE,        IS_NTSC ? 0x67 : 0x3F           );      // 0x5A: 
    eset( v, E_GAIN_U_LSB,          IS_NTSC ? 0x76 : 0x7D           );      // 0x5B: Gain U
    eset( v, E_GAIN_V_LSB,          IS_NTSC ? 0xA5 : 0xAF           );      // 0x5C: Gain V
    eset( v, E_GAIN_U_MSB,          0x40 | 0x2B                     );      // 0x5D: Gain U MSB, Real time control, Black level
    eset( v, E_GAIN_V_MSB,          0x40 | 0x2E                     );      // 0x5E: Gain V MSB, Real time control, Black level
    eset( v, E_CCR_BLANKING_LVL,    0x2E                            );      // 0x5F:

    eset( v, E_NULL_60,             0x00                            );      // 0x60:
    eset( v, E_STANDARD_CTRL,       IS_NTSC ? 0x15 : 0x16           );      // 0x61:
    eset( v, E_RTCEN_BURSTAMPL,     0x80 | (IS_NTSC ? 0x3F : 0x2F)  );      // 0x62: Realtime control enable & Color burst amplitude.
    eset( v, E_SUBCARRIER_0,        IS_NTSC ? 0x1F : 0xCB           );      // 0x63 to 0x66: Subcarrier frequency = 0x21F07C1F (NTSC) or 0x2A098ACB (PAL).
    eset( v, E_SUBCARRIER_1,        IS_NTSC ? 0x7C : 0x8A           );
    eset( v, E_SUBCARRIER_2,        IS_NTSC ? 0xF0 : 0x09           );
    eset( v, E_SUBCARRIER_3,        IS_NTSC ? 0x21 : 0x2A           );
    eset( v, 0x67,                  0x00                            );      // 0x67 to 0x6A: Teletext.
    eset( v, 0x68,                  0x00                            );
    eset( v, 0x69,                  0x00                            );
    eset( v, 0x6A,                  0x00                            );
    eset( v, E_RCV_PORT_CTRL,       0x12                            );      // 0x6B: pins RCV1 and RCV2 both set to output
    eset( v, E_TRIG_CTRL_LSB,       0xFD                            );      // 0x6C:
    eset( v, E_TRIG_CTRL_MSB,       0x07                            );      // 0x6D:
    eset( v, E_MULTI_CTRL,          0x00                            );      // 0x6E: 
    eset( v, E_CC_ENABLE,           0x02                            );      // 0x6F: Closed caption & teletext enable.

    eset( v, E_RCV2_OUT_START,      IS_NTSC ? 0x05 : 0xF1           );      // 0x70: Start NTSC 0x305 : PAL 0x1F1
    eset( v, E_RCV2_OUT_END_LSB,    IS_NTSC ? 0x94 : 0xA7           );      // 0x71: End   NTSC 0x694 : PAL 0x6A7
    eset( v, E_RCV2_OUT_END_MSB,    IS_NTSC ? 0x63 : 0x61           );      // 0x72: MSB (End, Start)
    eset( v, 0x73,                  0x00                            );      // 0x73 to 0x74: Teletext.
    eset( v, 0x74,                  0x00                            );
    eset( v, E_VSYNC_SHIFT,         0x03                            );      // 0x75:
    eset( v, 0x76,                  0x00                            );      // 0x76 to 0x79: Teletext.
    eset( v, 0x77,                  0x00                            );
    eset( v, 0x78,                  0x00                            );
    eset( v, 0x79,                  0x00                            );
    eset( v, E_FIRST_ACTIVE_LINE,   0x11                            );      // 0x7A: First active line.
    eset( v, E_LAST_ACTIVE_LINE,    IS_NTSC ? 0x03 : 0x31           );      // 0x7B: Last active line.

    eset( v, E_MSB_VERTICAL,        0x40                            );      // 0x7C:
    eset( v, E_NULL_7D,             0x00                            );      // 0x7D: Not used, but MUST be programmed on SAA7121 to make video output work.
    eset( v, E_DISAB_TTXLINE_LSB,   0x00                            );      // 0x7E:
    eset( v, E_DISAB_TTXLINE_MSB,   0x00                            );      // 0x7F:
}

static void enc_config_616( PVPAR v )
{
    int i;
#ifdef TEST_PCI_616
	return;
#endif
	
    for ( i = 0x00; i <= 0x25; i++ )                                        // 0x00 to 0x25: "Null" registers.
        eset( v, i, 0x00 );

    eset( v, E_WIDE_SCREEN_LSB,     0x00                            );      // 0x26: Wide screen signalling output disabled.
    eset( v, E_WIDE_SCREEN_MSB,     0x00                            );      // 0x27: Wide screen signal.
    eset( v, E_RTC_BURSTSTART,      IS_NTSC ? 0x19 : 0x21           );      // 0x28: Starting point of burst in clock cycles.
    eset( v, E_BURST_END,           IS_NTSC ? 0x1D : 0x1D           );      // 0x29: Ending point of burst in clock cycles.
    eset( v, E_COPYGUARD_ODD_0,     0x00                            );      // 0x2A
    eset( v, E_COPYGUARD_ODD_1,     0x00                            );      // 0x2B
    eset( v, E_COPYGUARD_EVEN_0,    0x00                            );      // 0x2C
    eset( v, E_COPYGUARD_EVEN_1,    0x00                            );      // 0x2D
    eset( v, E_COPYGUARD_ENABLE,    0x00                            );      // 0x2E: Copy guard encoding disabled.

    for ( i = 0x2F; i <= 0x39; i++ )                                        // 0x2F to 0x39: "Null" registers.
        eset( v, i, 0x00 );

    set_colorbars_enable( v );                                              // 0x3A: Colorbars control.

    for ( i = 0x3B; i <= 0x59; i++ )                                        // 0x3B to 0x59: Functions not specified, but other drivers write zeros to these regs.
        eset( v, i, 0x00 );

    eset( v, E_CHROMA_PHASE,        IS_NTSC ? 0x67 : 0x3F           );      // 0x5A:
    eset( v, E_GAIN_U_LSB,          IS_NTSC ? 0x76 : 0x7D           );      // 0x5B: Gain U
    eset( v, E_GAIN_V_LSB,          IS_NTSC ? 0xA5 : 0xAF           );      // 0x5C: Gain V
    eset( v, E_GAIN_U_MSB,          0x40 | 0x2B                     );      // 0x5D: Gain U MSB, Real time control, Black level
    eset( v, E_GAIN_V_MSB,          0x40 | 0x2E                     );      // 0x5E: Gain V MSB, Real time control, Black level
    eset( v, E_CCR_BLANKING_LVL,    0x2E                            );      // 0x5F:

    eset( v, E_NULL_60,             0x00                            );      // 0x60:
    eset( v, E_STANDARD_CTRL,       IS_NTSC ? 0x15 : 0x16           );      // 0x61:
    eset( v, E_RTCEN_BURSTAMPL,     IS_NTSC ? 0x3F : 0x2F           );      // 0x62: Realtime control enable & Color burst amplitude.
    eset( v, E_SUBCARRIER_0,        IS_NTSC ? 0x1F : 0xCB           );      // 0x63 to 0x66: Subcarrier frequency = 0x21F07C1F (NTSC) or 0x2A098ACB (PAL).
    eset( v, E_SUBCARRIER_1,        IS_NTSC ? 0x7C : 0x8A           );
    eset( v, E_SUBCARRIER_2,        IS_NTSC ? 0xF0 : 0x09           );
    eset( v, E_SUBCARRIER_3,        IS_NTSC ? 0x21 : 0x2A           );
    eset( v, 0x67,                  0x00                            );      // 0x67 to 0x6A: Teletext.
    eset( v, 0x68,                  0x00                            );
    eset( v, 0x69,                  0x00                            );
    eset( v, 0x6A,                  0x00                            );
    //Mod 6/12/07 A.K.
    //E_RCV_PORT_CTRL (0x32) set to generate a proper vert.sync (twice the frame) and proper hor.sync (present during vert. retrace)
    //for the sake of proper capture of full res image via 7146
    eset( v, E_RCV_PORT_CTRL,       0x20 | 0x12                     );      // 0x6B: hsync taken from RCV2, pins RCV1 and RCV2 both set to output
    eset( v, E_TRIG_CTRL_LSB,       0xFD                            );      // 0x6C:
    eset( v, E_TRIG_CTRL_MSB,       0x07                            );      // 0x6D:
    eset( v, E_MULTI_CTRL,          0x30                            );      // 0x6E: phres = reset every 4 fields
    eset( v, E_CC_ENABLE,           0x02                            );      // 0x6F: Closed caption & teletext enable.
    // DA: TODO: note 616 relay clicks on install.  setting below may be incorrect and need changed
    eset( v, E_RCV2_OUT_START,      IS_NTSC ? 0x05 : 0xF1           );      // 0x70: Start NTSC 0x105 PAL 0x0F1
    eset( v, E_RCV2_OUT_END_LSB,    IS_NTSC ? 0x94 : 0xA7           );      // 0x71: End   NTSC 0x694 PAL 0x6A7
    // DA: On capture works with 61 for NTSC for capture.  60 for PAL
    eset( v, E_RCV2_OUT_END_MSB,    IS_NTSC ? 0x61 : 0x60           );      // 0x72: MSB (End, Start)
    eset( v, 0x73,                  0x00                            );      // 0x73 to 0x74: Teletext.
    eset( v, 0x74,                  0x00                            );
    eset( v, E_VSYNC_SHIFT,         0x03                            );      // 0x75:
    eset( v, 0x76,                  0x00                            );      // 0x76 to 0x79: Teletext.
    eset( v, 0x77,                  0x00                            );
    eset( v, 0x78,                  0x00                            );
    eset( v, 0x79,                  0x00                            );
    eset( v, E_FIRST_ACTIVE_LINE,   0x11                            );      // 0x7A: First active line.
    eset( v, E_LAST_ACTIVE_LINE,    IS_NTSC ? 0x03 : 0x31           );      // 0x7B: Last active line.

    eset( v, E_MSB_VERTICAL,        0x40                            );      // 0x7C:
    eset( v, E_NULL_7D,             0x00                            );      // 0x7D: Not used, but MUST be programmed on SAA7121 to make video output work.
    eset( v, E_DISAB_TTXLINE_LSB,   0x00                            );      // 0x7E:
    eset( v, E_DISAB_TTXLINE_MSB,   0x00                            );      // 0x7F:
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Program runtime parameters to decoder (SAA7113/SAA7114).

// 516: SAA7113 decoder.
static void dec_config_516( PVPAR v )
{
    dset( v, D_CHIP_VERSION,    0x00                                );      // 0x00: Chip version.
    dset( v, D_INCREMENT_DELAY, 0x08                                );      // 0x01: Horizontal increment delay.

    set_video_in_516( v );                                                  // 0x02/0x09: AIN control 1: video input / Luminance control.

    dset( v, D_AIN_CTRL_2,      0x10                                );      // 0x03: AIN control 2: agc enabled.
    dset( v, D_AIN_CTRL_3,      0x00                                );      // 0x04: AIN control 3: static gain control ch1 (not used).
    dset( v, D_AIN_CTRL_4,      0x00                                );      // 0x05: AIN control 4. static gain control ch2 (not used).
    dset( v, D_HSYNC_START,     0xE9                                );      // 0x06: Horizontal sync start.
    dset( v, D_HSYNC_STOP,      0x0D                                );      // 0x07: Horizontal sync stop.
    dset( v, D_SYNC_CONTROL,    IS_NTSC ? 0x61 : 0x21               );      // 0x08: Sync control.

    set_brightness( v );                                                    // 0x0A: Luminance brightness.
    set_contrast  ( v );                                                    // 0x0B: Luminance contrast.
    set_saturation( v );                                                    // 0x0C: Chroma saturation.

    dset( v, D_CHROMA_HUE,      0x00                                );      // 0x0D: Chroma hue control.
    dset( v, D_CHROMA_CTRL,     IS_NTSC ? 0x89 : 0x81               );      // 0x0E: Chroma control.
    dset( v, D_CHROMA_GAIN,     IS_NTSC ? 0x24 : 0x2a               );      // 0x0F: Chroma gain control.

    dset( v, D_FORMAT_DELAY_7113,   IS_NTSC ? 0x04 : 0x04           );      // 0x10: Format/delay control.
    dset( v, D_OUT_CTRL_1_7113,     0x0C                            );      // 0x11: Output control 1.
    dset( v, D_OUT_CTRL_2_7113,     0xE7                            );      // 0x12: Output control 2.
    dset( v, D_OUT_CTRL_3_7113,     0x00                            );      // 0x13: Output control 3.
    dset( v, D_RESERVED_14,     0x00                                );      // 0x14:  reserved.
    dset( v, D_V_GATE1_START,   0x10                                );      // 0x15: V_GATE1_START.
    dset( v, D_V_GATE1_STOP,    0x08                                );      // 0x16: V_GATE1_STOP.
    dset( v, D_V_GATE1_MSB,     0x00                                );      // 0x17: V_GATE1_MSB.
}

// 616:  SAA7114 decoder.
// 6/8/2007
// RTS0 on 7114 is HSYNC_DEM into gate array (HS)
// RTS1 on 7114 is VSYNC_DEM into gate array (VS)
// RTC0 is RTC0
static void dec_config_616( PVPAR v )
{
    dset( v, D_CHIP_VERSION,    0x00                                );      // 0x00: Chip version (read only).
    dset( v, D_INCREMENT_DELAY, 0x08                                );      // 0x01: Horizontal increment delay (SAA7114 spec says this MUST be 0x08).

    set_video_in_616( v );                                                  // 0x02/0x09: AIN control 1: video input / Luminance control.

    dset( v, D_AIN_CTRL_2,      0x33                                );      // 0x03: AIN control 2: agc enabled.
    dset( v, D_AIN_CTRL_3,      0x00                                );      // 0x04: AIN control 3: static gain control ch1 (not used).
    dset( v, D_AIN_CTRL_4,      0x00                                );      // 0x05: AIN control 4. static gain control ch2 (not used).
    dset( v, D_HSYNC_START,     0xeb                                );      // 0x06: Horizontal sync start.
    dset( v, D_HSYNC_STOP,      0xe0                                );      // 0x07: Horizontal sync stop.
    dset( v, D_SYNC_CONTROL,    0xd8                                );      // 0x08: Sync control.

    set_brightness( v );                                                    // 0x0A: Luminance brightness.
    set_contrast  ( v );                                                    // 0x0B: Luminance contrast.
    set_saturation( v );                                                    // 0x0C: Chroma saturation.

    dset( v, D_CHROMA_HUE,      0x00                                );      // 0x0D: Chroma hue control.
    dset( v, D_CHROMA_CTRL,     0x89                                );      // 0x0E: Chroma control.
    dset( v, D_CHROMA_GAIN,     0x2a                                );      // 0x0F: Chroma gain control.

    // 6/8/2007 D_FORMAT_DELAY was wrong register!  Adding chroma gain 2
    // Chroma gain 2 is not used because AGC set in 0x0f 
    dset( v, D_CHROMA_GAIN2,    0x0e                                );      

    // fixed 6/8/2007
    dset( v, D_FORMAT_DELAY_7114,   0x00                            );      // 0x11: Format/delay control.
    // 7114 RTS0[12h 0:3] = Href,  RTS1[12h 7:4] = VS
    dset( v, D_OUT_CTRL_1_7114,     0xc8                            );      // 0x12: Output control 1.
    dset( v, D_OUT_CTRL_2_7114,     0xd0                            );      // 0x13: Output control 2.
    dset( v, D_RESERVED_14,     0x00                                );      // 0x14:  reserved.
    dset( v, D_V_GATE1_START,   0x11                                );      // 0x15: V_GATE1_START.
    dset( v, D_V_GATE1_STOP,    0x06                                );      // 0x16: V_GATE1_STOP.
    dset( v, D_V_GATE1_MSB,     0x40                                );      // 0x17: V_GATE1_MSB.
    dset( v, 0x18,              0x40                                );      // 0x18: Raw data gain control.
    dset( v, 0x19,              0x80                                );      // 0x19: Raw data offset control.
    dset( v, 0x1A,              0x00                                );      // 0x1A:  reserved.
    dset( v, 0x1B,              0x00                                );      // 0x1B:  reserved.
    dset( v, 0x1C,              0x20                                );      // 0x1C:  reserved.
    dset( v, 0x1D,              0x00                                );      // 0x1D:  reserved.
    dset( v, 0x1E,              0x00                                );      // 0x1E:  reserved.
    dset( v, 0x1F,              0x00                                );      // 0x1F: Status byte video decoder.

    dset( v, 0x20, 0x00 );      //  reserved.
    dset( v, 0x21, 0x00 );      //  reserved.
    dset( v, 0x22, 0x00 );      //  reserved.
    dset( v, 0x24, 0x00 );      //  reserved.
    dset( v, 0x25, 0x00 );      //  reserved.
    dset( v, 0x26, 0x00 );      //  reserved.
    dset( v, 0x28, 0x00 );      //  reserved.
    dset( v, 0x29, 0x00 );      //  reserved.
    dset( v, 0x2A, 0x00 );      //  reserved.
    dset( v, 0x2B, 0x00 );      //  reserved.
    dset( v, 0x2C, 0x00 );      //  reserved.
    dset( v, 0x2D, 0x00 );      //  reserved.
    dset( v, 0x2E, 0x00 );      //  reserved.
    dset( v, 0x2F, 0x00 );      //  reserved.

    // SAA7114H section 8.6 Table 21
    switch( v->attr_current[ATTR_audio_sample_rate])
    {
        case AUDIO_SAMPLERATE_48KHZ:
            dset( v, 0x30, 0xCD );      // Audio master clock cycles per field <7:0>.
            dset( v, 0x31, 0x20 );      // Audio master clock cycles per field <15:8>.
            dset( v, 0x32, 0x03 );      // Audio master clock cycles per field <17:16>.
            break;
        case AUDIO_SAMPLERATE_44KHZ:
            dset( v, 0x30, 0xBC );      // Audio master clock cycles per field <7:0>.
            dset( v, 0x31, 0xDF );      // Audio master clock cycles per field <15:8>.
            dset( v, 0x32, 0x02 );      // Audio master clock cycles per field <17:16>.
            break;
        case AUDIO_SAMPLERATE_32KHZ:
            dset( v, 0x30, 0xDE );      // Audio master clock cycles per field <7:0>.
            dset( v, 0x31, 0x15 );      // Audio master clock cycles per field <15:8>.
            dset( v, 0x32, 0x02 );      // Audio master clock cycles per field <17:16>.
            break;
    }

    dset( v, 0x33, 0x00 );      //  reserved.

    // SAA7114H section 8.6 Table 21
    switch( v->attr_current[ATTR_audio_sample_rate])
    {
        case AUDIO_SAMPLERATE_48KHZ:
            dset( v, 0x34, 0xFF );      // Audio master clock nominal increment <7:0>.
            dset( v, 0x35, 0xFF );      // Audio master clock nominal increment <15:8>.
            dset( v, 0x36, 0x3F );      // Audio master clock nominal increment <21:16>.
            break;
        case AUDIO_SAMPLERATE_44KHZ:
            dset( v, 0x34, 0xCD );      // Audio master clock nominal increment <7:0>.
            dset( v, 0x35, 0xCC );      // Audio master clock nominal increment <15:8>.
            dset( v, 0x36, 0x3A );      // Audio master clock nominal increment <21:16>.
            break;
        case AUDIO_SAMPLERATE_32KHZ:
            dset( v, 0x34, 0xAB );      // Audio master clock nominal increment <7:0>.
            dset( v, 0x35, 0xAA );      // Audio master clock nominal increment <15:8>.
            dset( v, 0x36, 0x2A );      // Audio master clock nominal increment <21:16>.
            break;
        default:
            dbgprint("invalid sampling rate\n");
            break;
    }


    dset( v, 0x37, 0x00 );      //  reserved.
    dset( v, 0x38, 0x03 );      // Clock ratio AMCLK to ASCLK.
    dset( v, 0x39, 0x10 );      // Clock ratio ASCLK to ALRCLK.
    dset( v, 0x3A, 0x08 );      // Audio clock control.
    dset( v, 0x3B, 0x00 );      //  reserved.
    dset( v, 0x3C, 0x00 );      //  reserved.
    dset( v, 0x3D, 0x00 );      //  reserved.
    dset( v, 0x3E, 0x00 );      //  reserved.
    dset( v, 0x3F, 0x00 );      //  reserved.

    dset( v, 0x40, 0x00 );      // Slicer control 1.
    dset( v, 0x41, 0xFF );      // LCR2.
    dset( v, 0x42, 0xFF );      // LCR3.
    dset( v, 0x43, 0xFF );      // LCR4.
    dset( v, 0x44, 0xFF );      // LCR5.
    dset( v, 0x45, 0xFF );      // LCR6.
    dset( v, 0x46, 0xFF );      // LCR7.
    dset( v, 0x47, 0xFF );      // LCR8.
    dset( v, 0x48, 0xFF );      // LCR9.
    dset( v, 0x49, 0xFF );      // LCR10.
    dset( v, 0x4A, 0xFF );      // LCR11.
    dset( v, 0x4B, 0xFF );      // LCR12.
    dset( v, 0x4C, 0xFF );      // LCR13.
    dset( v, 0x4D, 0xFF );      // LCR14.
    dset( v, 0x4E, 0xFF );      // LCR15.
    dset( v, 0x4F, 0xFF );      // LCR16.

    dset( v, 0x50, 0xFF );      // LCR17.
    dset( v, 0x51, 0xFF );      // LCR18.
    dset( v, 0x52, 0xFF );      // LCR19.
    dset( v, 0x53, 0xFF );      // LCR20.
    dset( v, 0x54, 0x5F );      // LCR21.
    dset( v, 0x55, 0xFF );      // LCR22.
    dset( v, 0x56, 0xFF );      // LCR23.
    dset( v, 0x57, 0xFF );      // LCR24.
    dset( v, 0x58, 0x00 );      // Programmable framing code.
    dset( v, 0x59, 0x47 );      // Horizontal offset for slicer.
    dset( v, 0x5A, 0x06 );      // Vertical offset for slicer.
    dset( v, 0x5B, 0x83 );      // Field offset and MSBs for horizontal and vertical offset.
    dset( v, 0x5C, 0x00 );      //  reserved.
    dset( v, 0x5D, 0x3E );      // Header and data identification (DID) code control.
    dset( v, 0x5E, 0x00 );      // Sliced data identification (SDID) code.
    dset( v, 0x5F, 0x00 );      //  reserved.

    dset( v, 0x60, 0xFF );      // Slicer status byte 0.
    dset( v, 0x61, 0xFF );      // Slicer status byte 1.
    dset( v, 0x62, 0xFF );      // Slicer status byte 2.
    dset( v, 0x63, 0x00 );      //  reserved.
    dset( v, 0x64, 0x00 );      //  reserved.
    dset( v, 0x65, 0x00 );      //  reserved.
    dset( v, 0x66, 0x00 );      //  reserved.
    dset( v, 0x67, 0x00 );      //  reserved.
    dset( v, 0x68, 0x00 );      //  reserved.
    dset( v, 0x69, 0x00 );      //  reserved.
    dset( v, 0x6A, 0x00 );      //  reserved.
    dset( v, 0x6B, 0x00 );      //  reserved.
    dset( v, 0x6D, 0x00 );      //  reserved.
    dset( v, 0x6E, 0x00 );      //  reserved.
    dset( v, 0x6F, 0x00 );      //  reserved.

    dset( v, 0x70, 0x00 );      //  reserved.
    dset( v, 0x71, 0x00 );      //  reserved.
    dset( v, 0x72, 0x00 );      //  reserved.
    dset( v, 0x73, 0x00 );      //  reserved.
    dset( v, 0x74, 0x00 );      //  reserved.
    dset( v, 0x75, 0x00 );      //  reserved.
    dset( v, 0x76, 0x00 );      //  reserved.
    dset( v, 0x77, 0x00 );      //  reserved.
    dset( v, 0x78, 0x00 );      //  reserved.
    dset( v, 0x79, 0x00 );      //  reserved.
    dset( v, 0x7A, 0x00 );      //  reserved.
    dset( v, 0x7B, 0x00 );      //  reserved.
    dset( v, 0x7D, 0x00 );      //  reserved.
    dset( v, 0x7E, 0x00 );      //  reserved.
    dset( v, 0x7F, 0x00 );      //  reserved.

    dset( v, 0x80, 0x10 );      // Global control 1.
    dset( v, 0x81, 0x00 );      //  reserved.
    dset( v, 0x82, 0x00 );      //  reserved.
    dset( v, 0x83, 0x02 );      // X-port I/O enable and output clock phase control.
    dset( v, 0x84, 0x00 );      // I-port signal definitions.
    dset( v, 0x85, 0x00 );      // I-port signal polarities.
    dset( v, 0x86, 0x45 );      // I-port FIFO flag control and arbitration.
    dset( v, 0x87, 0x02 );      // I-port I/O enable, output clock and gated clock phase control.
    dset( v, 0x88, 0xF0 );      // Power-save control.
    dset( v, 0x89, 0x00 );      //  reserved.
    dset( v, 0x8A, 0x00 );      //  reserved.
    dset( v, 0x8B, 0x00 );      //  reserved.
    dset( v, 0x8C, 0x00 );      //  reserved.
    dset( v, 0x8D, 0x00 );      //  reserved.
    dset( v, 0x8E, 0x00 );      //  reserved.
    dset( v, 0x8F, 0x00 );      // Status info scalar part.

    dset( v, 0x90, 0x10 );      // Task handling control.
    dset( v, 0x91, 0x08 );      // X-port formats and configuration.
    dset( v, 0x92, 0x00 );      // X-port input reference signal definition.
    dset( v, 0x93, 0x80 );      // I-port format and configuration.
    dset( v, 0x94, 0x00 );      // Horizontal input window start <7:0>.
    dset( v, 0x95, 0x00 );      // Horizontal input window start <11:8>.
    dset( v, 0x96, 0xD0 );      // Horizontal input window length <7:0>.
    dset( v, 0x97, 0x02 );      // Horizontal input window length <11:8>.
    dset( v, 0x98, 0x00 );      // Vertical input window start <7:0>.
    dset( v, 0x99, 0x00 );      // Vertical input window start <11:8>.
    dset( v, 0x9A, 0xFE );      // Vertical input window length <7:0>.
    dset( v, 0x9B, 0x00 );      // Vertical input window length <11:8>.
    dset( v, 0x9C, 0xD0 );      // Horizontal output window length <7:0>.
    dset( v, 0x9D, 0x02 );      // Horizontal output window length <11:8>.
    dset( v, 0x9E, 0xFE );      // Vertical output window length <7:0>.
    dset( v, 0x9F, 0x00 );      // Vertical output window length <11:8>.

    dset( v, 0xA0, 0x01 );      // Horizontal prescaling.
    dset( v, 0xA1, 0x00 );      // Accumulation length.
    dset( v, 0xA2, 0x00 );      // Prescaler DC gain and FIR prefilter control.
    dset( v, 0xA3, 0x00 );      //  reserved.
    dset( v, 0xA4, 0x80 );      // Luminance video_brightness setting.
    dset( v, 0xA5, 0x40 );      // Luminance video_contrast setting.
    dset( v, 0xA6, 0x40 );      // Chrominance video_saturation setting.
    dset( v, 0xA7, 0x00 );      //  reserved.
    dset( v, 0xA8, 0x00 );      // Horizontal luminance scaling increment <7:0>.
    dset( v, 0xA9, 0x04 );      // Horizontal luminance scaling increment <12:8>.
    dset( v, 0xAA, 0x00 );      // Horizontal luminance phase offset.
    dset( v, 0xAB, 0x00 );      //  reserved.
    dset( v, 0xAC, 0x00 );      // Horizontal chrominance scaling increment <7:0>.
    dset( v, 0xAD, 0x02 );      // Horizontal chrominance scaling increment <12:8>.
    dset( v, 0xAE, 0x00 );      // Horizontal chrominance phase offset.
    dset( v, 0xAF, 0x00 );      //  reserved.

    dset( v, 0xB0, 0x00 );      //
    dset( v, 0xB1, 0x02 );      // 
    dset( v, 0xB2, 0x00 );      //
    dset( v, 0xB3, 0x02 );      // 
    dset( v, 0xB4, 0x00 );      // 
    dset( v, 0xB8, 0x00 );      // 
    dset( v, 0xB9, 0x00 );      //
    dset( v, 0xBA, 0x00 );      // 
    dset( v, 0xBB, 0x00 );      // 
    dset( v, 0xBC, 0x00 );      // 
    dset( v, 0xBD, 0x00 );      // 
    dset( v, 0xBE, 0x00 );      // 
    dset( v, 0xBF, 0x00 );      // 
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////  AUDIO CODEC (UDA1341) INTERFACE  //////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Return current state of L3 bit-bang control bus.  On the 516, the bit-bang register is directly accessed through
// a GA register.  On the 616, the bit-bang register is indirectly accessed over the I2C bus via the PCA9554 device
// (which provides access to an 8-bit general-purpose port via I2C interface).

// Read GA's bit-bang L3 control bus.
static int L3_read_516( PVPAR v )
{
    return ga_indirect_reg_read( v, G_L3CONTROL );          
}

// Read L3 control bus bits from PCA9554's output port register.
static int L3_read_616( PVPAR v )
{
    return i2c_read_616( v, I2CADDR_IO_616, PCA_PORT_OUT ); 
}

////////////////////////////////////////////////
// Write value to L3 bit-bang control bus.

// Write GA's bit-bang L3 control bus.
static void L3_write_516( PVPAR v, int data )
{
    ga_indirect_reg_write( v, G_L3CONTROL, (USHORT)data );
}

// Write L3 control bus bits to PCA9554's output port register.
static void L3_write_616( PVPAR v, int data )
{
    i2c_write_616( v, I2CADDR_IO_616, PCA_PORT_OUT, ( L3_read_616(v) & ~L3MASK_616 ) | ( data & L3MASK_616 ) );
}

////////////////////////////////////////////////////////////////////////////////////
// Send byte to audio codec.  Clk is held high when idle (upon entry and exit).

static void SendByteL3( PVPAR v, int dataval )
{
    int value;
    int bitcnt;
    
    // Get the initial state of the Mode bit (and all non-L3 output bits used by PCA9554 in the
    // case of model 616) so it (they) can be preserved throughout this function.
    int base_image = v->L3_read( v ) & ~( v->L3mask_clk | v->L3mask_data );
    
    // Align data value lsb with the bit-banging bus data bit.
    dataval <<= v->L3bit_data;

    // Transmit 8 bits to the codec ...
    for ( bitcnt = 0; bitcnt < 8; bitcnt++ )
    {
        // Combine data bit with mode control bit.
        value = ( dataval & v->L3mask_data ) | base_image;

        // Generate negative clock pulse.  Data bit is clocked out on positive edge.

        v->L3_write( v, value & ~v->L3mask_clk );
        v->L3_write( v, value | v->L3mask_clk );

        // Align next data bit to be sent.
        dataval >>= 1;
    }
}

////////////////////////////////////////////////////////////////////////////////////
// Send control command to audio codec.

static void L3_Out( PVPAR v, int L3_reg, int data )
{
    // Send target register address to the codec.
    v->L3_write( v, v->L3mask_clk | v->L3mask_data                  );  // Switch Mode bit low to select Address mode.

    SendByteL3( v, ( ADDR_AUDIOCODEC << 2 ) | L3_reg                );  // Shift out the address to codec.
    v->L3_write( v, v->L3mask_clk | v->L3mask_data | v->L3mask_mode );  // Set Mode bit high to terminate Address mode..

    // Send the data byte to the codec.
    SendByteL3( v, data );                                              // Shift out the data to codec.
}

//////////////////////////////////////////////////////////////////////////////////////////
// Initialize the UDA1341 audio codec.

static void AudioInit( PVPAR v )
{
    // Init the audio control bit-banging bus by setting all L3 control lines high.  Preserve all
    // non-L3 control bits in case this is a 616.
    v->L3_write( v, v->L3_read( v ) | v->L3mask_clk | v->L3mask_data | v->L3mask_mode );

    // Pulse the codec's Reset and configure its serial communication interface for compatibility with the SM.
    L3_Out( v, L3REG_STATUS, SEL_STATUS_0 | STAT0_CONFIG | STAT0_RESET_ON  );
    
    L3_Out( v, L3REG_STATUS, SEL_STATUS_0 | STAT0_CONFIG | STAT0_RESET_OFF );

    // Configure fundamental operating parameters.  
    L3_Out( v, L3REG_STATUS, SEL_STATUS_1   // Write to Status Register 1:
        | ( 0 << 6 )                        //   DAC gain = 0 db.
        | ( 0 << 5 )                        //   ADC gain = 0 db.
        | ( 0 << 4 )                        //   ADC polarity = non-inverting.
        | ( 0 << 3 )                        //   DAC polarity = non-inverting.
        | ( 0 << 2 )                        //   Single speed playback.
        | ( 1 << 1 )                        //   Apply power to ADC.
        | ( 1 << 0 )                        //   Apply power to DAC.
    );

    // Program the default settings.        
    L3_Out( v, L3REG_DATA0, SEL_DATA0_VOLUME    | DFLT_VOLUME                       );  // Volume level.
    L3_Out( v, L3REG_DATA0, SEL_DATA0_BASSTREB  | ( DFLT_BASS << 2 ) | DFLT_TREBLE  );  // Bass and audio_treble levels.
    L3_Out( v, L3REG_DATA0, SEL_DATA0_MISC      | ( DFLT_MUTE << 2 )                );  // Mute.
    L3_Out( v, L3REG_DATA0, SEL_DATA0_EXTADDR   | DATA0_EXT_MS_MM                   );  // Select extended reg for MIC sensitivity and mixer mode control.
    L3_Out( v, L3REG_DATA0, SEL_DATA0_EXTDATA   | 2                                 );  // Extended data:  select channel 2.
}

//////////////////////////////////////
// Sets Microphone AGC on/off with microphone sensitivity ATTR_audio_agc_sens
// also sets ATTR_audio_inampgain. used only when AGC is off.  
static void set_audio_agc_ingain( PVPAR v )
{
  /*dbgprint("AGCC %d, AGCS %d, G %d\n", v->attr_current[ATTR_audio_agc_onoff],
            v->attr_current[ATTR_audio_agc_sens],
            v->attr_current[ATTR_audio_inampgain]);
  */
    if( v->attr_current[ATTR_audio_agc_onoff]) {
        // set MIC sensitivity
        L3_Out( v, L3REG_DATA0, SEL_DATA0_EXTADDR | DATA0_EXT_MS_MM );
                                // 0xC0           + EA[2:0] = 0x02 (extended to MS[2:0]+MM[1:0])
        SendByteL3( v, SEL_DATA0_EXTDATA | ( (v->attr_current[ATTR_audio_agc_sens]<<2) & 0x1c ) | 0x02 ); // send DATA0 2nd byte.
                        // 0xE0          +                                 MS[2:0]              + Chan-2 selected
        // turn on AGC
        L3_Out( v, L3REG_DATA0, SEL_DATA0_EXTADDR | DATA0_EXT_AG_IGLO );
                                // 0xC0           + EA[2:0] = 0x04 (extended to AG+0+0+IG[1:0])
        SendByteL3( v, SEL_DATA0_EXTDATA | 0x10 | (v->attr_current[ATTR_audio_inampgain] & 0x02) ); // send DATA0 2nd byte.
                        // 0xE0          + AG=1 ( enable AGC: gain control with manual MIC sensitivity setting )
                        //                      + IG[1:0]
    }
    else {
        // turn off AGC & keep manual inampgain settings.
        L3_Out( v, L3REG_DATA0, SEL_DATA0_EXTADDR | DATA0_EXT_AG_IGLO );
                                // 0xC0           + EA[2:0] = 0x04 (extended to AG+0+0+IG[1:0])
        SendByteL3( v, SEL_DATA0_EXTDATA | 0x00 | (v->attr_current[ATTR_audio_inampgain] & 0x02) ); // send DATA0 2nd byte.
                        // 0xE0          + AG=0 ( disable AGC: gain control with manual gain setting through IG[6:0] )
                        //                      + IG[1:0]
        // high byte IGHI
        L3_Out( v, L3REG_DATA0, SEL_DATA0_EXTADDR | DATA0_EXT_IGHI );
                                // 0xC0           + EA[2:0] = 0x05 (extended to IG[6:2])
        SendByteL3( v, SEL_DATA0_EXTDATA |  (v->attr_current[ATTR_audio_inampgain]>>2 & 0x1f) ); // send DATA0 2nd byte.
                        // 0xE0          +                              IG[6:2]
    }
}


//////////////////////////////////////
// Change audio volume level.

static void set_volume( PVPAR v )
{
    L3_Out( v, L3REG_DATA0, SEL_DATA0_VOLUME | (v->attr_current[ATTR_audio_volume] &0x3F) );
}

///////////////////////////////////////////
// Change audio bass and/or treble level.

static void set_bass_treble( PVPAR v )
{
    L3_Out( v, L3REG_DATA0, SEL_DATA0_BASSTREB | ( (v->attr_current[ATTR_audio_bass] &0x0F) << 2 ) | (v->attr_current[ATTR_audio_treble] &0x03) );
}

//////////////////////////////////
// Change audio mute enable.

static void set_mute( PVPAR v )
{
    L3_Out( v, L3REG_DATA0, SEL_DATA0_MISC | ( v->attr_current[ATTR_audio_mute] << 2 ) );
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////  STREAM MACHINE STATUS  ////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Packet type identifiers.
#define PKT_TYPE_COMMAND_ACK        0x0     // User command ACK.
#define PKT_TYPE_COMMAND_NACK       0x1     // User command NACK.
#define PKT_TYPE_MICROCODE_ACK      0x2     // Microcode upload ACK.
#define PKT_TYPE_STATUS_RPT         0x3     // Status report.
#define PKT_TYPE_USERDATA_ACK       0x4     // User data ACK.
#define PKT_TYPE_EXCEPTION_RPT      0x6     // Exception report.
#define PKT_TYPE_END_OF_STREAM      0xE     // End of stream data reached after STOP cmd (0xEEFF).
#define PKT_TYPE_PARAMS_ACK         0xF     // Parameters upload ACK (0xFEED).

/////////////////////////////////////
// Return TRUE if SM is decoding.

static BOOL stream_is_decoding( PVPAR v )
{
    return ( v->active_microcode == MPEG_PLAY ) && ( v->stream_state == STREAM_RUNNING );
}

BOOL CORE_stream_is_decoding( HBD board )
{
    return stream_is_decoding( (PVPAR)board );
}

/////////////////////////////////////
// Return TRUE if SM is encoding.

static BOOL stream_is_encoding( PVPAR v )
{
    return ( v->active_microcode == MPEG_RECORD ) && ( v->stream_state == STREAM_RUNNING );
}

BOOL CORE_stream_is_encoding( HBD board )
{
    return stream_is_encoding( (PVPAR)board );
}

//////////////////////////////////////////////////////////
// DEBUG FUNCTION:  Display SM response packet info.

static void debug_print_status( PVPAR v, ULONG packet_size )
{
    ULONG i;
    
    switch ( v->sm_responses[0] >> 12 )
    {
        case PKT_TYPE_STATUS_RPT:           SX16MOD_text_message( "SM status report: "      );  break;
        case PKT_TYPE_EXCEPTION_RPT:        SX16MOD_text_message( "SM exception report: "       );  break;
        case PKT_TYPE_USERDATA_ACK:         SX16MOD_text_message( "SM user data report: "       );  break;
        case PKT_TYPE_COMMAND_NACK:         SX16MOD_text_message( "SM cmd nack: "               );  break;
        case PKT_TYPE_MICROCODE_ACK:        SX16MOD_text_message( "SM ucode upload ack: "       );  break;

        case PKT_TYPE_COMMAND_ACK:
            switch ( v->sm_responses[0] )
            {
                case 0x0000:                SX16MOD_text_message( "SM STOP cmd ack: "           );  break;
                case 0x0001:                SX16MOD_text_message( "SM START cmd ack: "      );  break;
                case 0x0002:                SX16MOD_text_message( "SM PAUSE cmd ack: "      );  break;
                case 0x0003:                SX16MOD_text_message( "SM RESUME cmd ack: "     );  break;
                default:                    SX16MOD_text_message( "SM unknown cmd ack: "        );  break;
            }
            break;
            
        case PKT_TYPE_END_OF_STREAM:
            switch ( v->sm_responses[0] )
            {
                case 0xEEFF:                SX16MOD_text_message( "SM end of stream: "      );  break;
                default:                    SX16MOD_text_message( "SM unknown pkt type: "       );  break;
            }
            break;

        case PKT_TYPE_PARAMS_ACK:
            switch ( v->sm_responses[0] )
            {
                case 0xFEED:                SX16MOD_text_message( "SM params upload ack: "  );  break;
                default:                    SX16MOD_text_message( "SM unknown pkt type: "       );  break;
            }
            break;

        default:                            SX16MOD_text_message( "SM unknown pkt type: "       );  break;
    }
    
    for ( i = 0; i < packet_size; i++ )
        SX16MOD_hex_message( 4, v->sm_responses[i] );

    SX16MOD_text_message( ".\n" );
}

//////////////////////////////////////////////////////////////////////////////////////////
// Return TRUE if at least one status word is available in SM command response FIFO.
// Wait up to wait_msec for the status word.  Set wait_msec to zero to exit immediately.

static INLINE BOOL sm_is_status_ready( PVPAR v, ULONG wait_msec )
{
    UINT64 t_start;

    // If a greater than zero wait time is specified ...
    if ( wait_msec != 0 )
    {
        // Cache the start time.
        t_start = SX16MOD_time_now();

        // Wait up to wait_msec for a response word to appear in the SM command response FIFO.
        do
        {
            // Return Success indication if there is at least one word in the response FIFO.
            if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_TXFIFO_NOTMT ) != 0 )
                return TRUE;

            // Allow time for the SM to populate the FIFO: call scheduler in case we are in kernel mode.
            SX16MOD_sleep( 0 );
        }
        while ( SX16MOD_time_now() - t_start < (UINT64)wait_msec );
    }

    // Return true if fifo has content, or false if it is empty (timeout).  In cases where wait_msec is not
    // zero, it's possible that we experienced a lengthly interruption after polling the fifo, and the fifo
    // is not empty now even though it timed out.  To cover this possibility, we must check the fifo level one
    // more time to be sure it's empty.
    return ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_TXFIFO_NOTMT ) != 0 );
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Fetch first word of SM status packet into sm_responses[].  Return packet size if successful,
// or 0 if timeout. Return immediately if SM's status fifo is empty and wait_msec is zero.

static INLINE int sm_fetch_packet_head( PVPAR v, ULONG wait_msec )
{
    ULONG   packet_type;                // Type of received packet.

    // As we are expecting a new response packet, reset response length and expected size to zero.
    v->sm_response_length        = 0;
    v->sm_response_expected_size = 0;

    // If we receive at least one status word before timing out ...
    if ( sm_is_status_ready( v, wait_msec ) )
    {
        // Fetch and cache the first (and possibly only) status word belonging to this packet.
        v->sm_responses[0] = v->sm_reg_read( v, SM_FIFOTX_REG );
        v->sm_response_length = 1;

        // Extract the packet type identifier.
        packet_type = v->sm_responses[0] >> 12;

        // Determine how many more words are expected in this packet.
        switch ( v->sm_responses[0] )
        {
        case 0xFEED:
            // Parameters upload ACK --------
            // No further status words expected for parameters upload ACK.  This is a special case that does
            // not conform to the format rules that apply to all other packet types.
            v->sm_response_expected_size = 0;
            break;

        case 0xEEFF:
            // End of MPEG stream following a Stop command ----------
            // After a Stop command we can continue reading MPEG data from the SM until this status word
            // value is received, at which time the SM will be ready to Start again without the need for
            // a SM reset and microcode upload.
            v->sm_response_expected_size = 0;
            break;

        case 0xAFF4:
            // Undocumented Record messages -----------
            // These status words have been seen, but are undocumented in the SM literature.  We assume
            // here that they are "stand-alone" indicators that occupy a single status word.
            v->sm_response_expected_size = 0;
            break;

        case 0xD002:
        case 0xD101:
            // Undocumented Play messages -----------
            // These status words have been seen, but are undocumented in the SM literature.  We assume
            // here that they are "stand-alone" indicators that occupy a single status word.
            v->sm_response_expected_size = 0;
            break;

        default:
            // All other packet types --------
            // No further words are expected for a command ACK/NACK packet.  In all other cases, extract the
            // remaining word count from the first word, bits 11-8.
            v->sm_response_expected_size = ( packet_type <= PKT_TYPE_COMMAND_NACK ) ? 0 : ( ( v->sm_responses[0] >> 8 ) & 0x000F );
        }

    }

    // Return total packet size in words.
    return v->sm_response_length;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
// Fetch remaining words of SM status packet into sm_responses[].  Return total packet size if successful,
// or 0 if timeout. Return immediately if SM's status fifo is empty and wait_msec is zero.

static INLINE int sm_fetch_packet_tail( PVPAR v, ULONG wait_msec )
{
    // Fetch and cache all remaining words belonging to this packet.
    for ( ; v->sm_response_expected_size > 0; v->sm_response_expected_size-- )
    {
        // Skip remaining words if the word fetch timed out.  We allow up to SM_NEXT_WORD_TOUT milliseconds
        // for each subsequent packet word to arrive in the SM's response fifo.
        if ( !sm_is_status_ready( v, wait_msec ) )
            break;

        // Fetch and cache the next word from SM's status packet FIFO.
        v->sm_responses[v->sm_response_length++] = v->sm_reg_read( v, SM_FIFOTX_REG );
    }

    // Return total packet size in words.
    return v->sm_response_length;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
// Fetch SM status packet into sm_responses[].  Return packet size if successful, or 0 if timeout.
// When encountered, packet types ExceptionReport and StatusReport are automatically processed
// and the packet receiver repeats until some other packet type is received or a timeout occurs.
// Return immediately if SM's status fifo is empty and wait_if_empty is not asserted.  Set wait
// time to zero to exit immediately.

static INLINE int sm_get_status_packet( PVPAR v, ULONG wait_msec )
{
    // Fetch packet into sm_responses[] buffer, allowing up to wait_msec for it to arrive in the SM's response fifo.

    // Get first part of response packet.
    sm_fetch_packet_head( v, wait_msec );

    // Get second part of response packet, if available.
    sm_fetch_packet_tail( v, wait_msec > 0 ? SM_NEXT_WORD_TOUT : 0 );

    // If a packet was received ...
    if ( v->sm_response_length != 0 )
    {
        // Log packet info if debug messages are enabled.
        if ( v->debug_flags & DEBUG_PACKET_INFO )
            debug_print_status( v, v->sm_response_length );

        // Decode the packet type and process the packet accordingly.
        switch ( v->sm_responses[0] >> 12 )
        {
        case PKT_TYPE_EXCEPTION_RPT:    // Exception report.  Since these are unsolicited (?), process these here.
                                        //   *** add exception handler here, followed by recursive call to this function ***

            break;

        case PKT_TYPE_STATUS_RPT:       // Status report.

        case PKT_TYPE_USERDATA_ACK:     // User data ACK.
        case PKT_TYPE_COMMAND_ACK:      // Command ACK.
        case PKT_TYPE_COMMAND_NACK:     // Command NACK.
        case PKT_TYPE_MICROCODE_ACK:    // Microcode upload ACK.
        case PKT_TYPE_PARAMS_ACK:       // Parameters upload ACK.
        case PKT_TYPE_END_OF_STREAM:    // End of mpeg stream data.
        default:                        // All other packet types.
            break;
        }
    }

    // Return the status packet's word count.
    return v->sm_response_length;
}

/////////////////////////////////////////////////////////////////////////
// Fetch SM status packet of desired_size into sm_responses[].
// Return non-zero if successful, or 0 if timeout.

static BOOL sm_get_sized_packet( PVPAR v, int desired_size, ULONG wait_msec )
{
    int packet_size;

    // Repeat until we timeout or receive a packet of desired_size ...
    for ( ; ; )
    {
        // Fetch status packet from SM response FIFO.
        packet_size = sm_get_status_packet( v, wait_msec );

        // Exit with Fail indication if we timed out while waiting for our packet.
        if ( packet_size == 0 )
            return FALSE;

        // Exit with Success indication if we received a packet of the desired size.
        if ( packet_size == desired_size )
            return TRUE;
    }
}

/////////////////////////////////////////////////////////////////////////
// Wait for a SM command ACK or timeout, whichever comes first.
// Return TRUE upon ACK, or FALSE upon NACK or timeout.

static BOOL sm_wait_command_ack( PVPAR v, USHORT opcode, ULONG wait_msec )
{
    // SM BUG WORKAROUND ----------------
    // SM doesn't send us command responses while in the playback mode, so we simply insert a delay instead
    // of waiting for a command response handshake; in this case we must assume the SM executed the command.

    // If player microcode is loaded ...
    if ( v->active_microcode == MPEG_PLAY )
    {
        // Delay to allow time for the SM to execute the command.
        SX16MOD_sleep( wait_msec );

        // Indicate the command executed, even though SM has given us no evidence of this happening.
        return TRUE;
    }

    // Recorder microcode is loaded, so loop until we get an ACK or NACK from the SM or we timeout while waiting for a status packet.
    for ( ; ; )
    {
        // Exit with Fail indication if there is no response (timed out).
        if ( sm_get_sized_packet( v, 1, wait_msec ) == 0 )
            return FALSE;

        // Process packet if ACK or NACK, else loop back for next packet.
        switch ( v->sm_responses[0] >> 12 )
        {
        case PKT_TYPE_COMMAND_ACK:      // Command ACK: return Success indication if response opcode matches the invoked opcode.
            if ( ( v->sm_responses[0] & 0x00FF ) == opcode )
                return TRUE;

            break;  // No match -- keep waiting.

        case PKT_TYPE_COMMAND_NACK:     // Command NACK: return Fail indication.
            return FALSE;
        }
    }
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////  STREAM CONTROL  ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define NUM_CUR_STATES      NUM_STREAM_STATES       // Number of possible current states.

#define NUM_NEXT_STATES     NUM_STREAM_STATES       // Number of possible next states.

// Map SM's current state and target next state into SM stream control opcode.
static const USHORT sm_cmd[ NUM_CUR_STATES ][ NUM_NEXT_STATES ] =
{
//  +----------- TARGET NEXT STATE -----------------+       CURRENT STATE
//  |  HALTED        RUNNING         PAUSED         |
    { SM_NOP,       SM_START_CMD,   SM_NOP          },      // HALTED
    { SM_STOP_CMD,  SM_NOP,         SM_PAUSE_CMD    },      // RUNNING
    { SM_STOP_CMD,  SM_RESUME_CMD,  SM_NOP          },      // PAUSED
};

/////////////////////////////////////////////////////////////////////////////
// Invoke a command on the SM.  Returns TRUE if successful (ack received).

static BOOL sm_exec_cmd( PVPAR v, USHORT opcode, enum STREAM_STATE next_state, ULONG wait_msec )
{
    BOOL rtnval = FALSE;

    // Issue stream control command to the SM.
    v->sm_reg_write( v, SM_FIFORX_REG, opcode );

    // Wait for command ACK, and if ACK is received ...
    if ( sm_wait_command_ack( v, opcode, wait_msec ) )
    {
        // Indicate new stream state is in effect.
        v->stream_state = next_state;

        // Indicate success.
        rtnval = TRUE;
    }

    // Return success/fail indication.
    return rtnval;
}

/////////////////////////////////////////////////////////////////////////////
// Switch SM to new MPEG stream mode.

static void set_stream_state( PVPAR v, enum STREAM_STATE next_state )
{
    // Abort if no microcode is loaded as the SM depends on the microcode to manage parameter downloads.    
    if ( v->active_microcode == MPEG_NONE )
        return;
    
    // Look up the SM opcode required for this state transition, then execute the state transition based on the SM opcode.
    switch ( sm_cmd[ v->stream_state ][ next_state ] )
    {
    case SM_START_CMD:  // START THE MPEG STREAM.

        if ( v->debug_flags & DEBUG_STREAM_CMD )
            SX16MOD_text_message( "CMD_START\n" );

        // Execute the SM command.
        sm_exec_cmd( v, SM_START_CMD, next_state, SM_START_TOUT );

        // If we are starting a Record mode stream ...
        if ( v->active_microcode == MPEG_RECORD )
        {
            // IS THIS NEEDED (FROM WINDOWS DRIVER) ???
//          SX16MOD_sleep( 1000 );

        }
        break;

    case SM_STOP_CMD:   // HALT THE MPEG STREAM.

        if ( v->debug_flags & DEBUG_STREAM_CMD )
            SX16MOD_text_message( "CMD_STOP\n" );

        // Execute the SM command.
        sm_exec_cmd( v, SM_STOP_CMD, next_state, SM_STOP_TOUT );

        // A SM reset and microcode upload is required if we just halted the MPEG stream.  If we don't
        // do this, the SM might be left in an unpredictable state.  Force a microcode reload by indicating
        // that no microcode is currently loaded.
        v->active_microcode = MPEG_NONE;

        break;

    case SM_PAUSE_CMD:  // PAUSE THE MPEG STREAM.

        if ( v->debug_flags & DEBUG_STREAM_CMD )
            SX16MOD_text_message( "CMD_PAUSE\n" );

        // Execute the SM command.
        sm_exec_cmd( v, SM_PAUSE_CMD, next_state, SM_PAUSE_TOUT );

        break;

    case SM_RESUME_CMD: // RESUME THE MPEG STREAM.

        if ( v->debug_flags & DEBUG_STREAM_CMD )
            SX16MOD_text_message( "CMD_RESUME\n" );

        // Execute the SM command.
        sm_exec_cmd( v, SM_RESUME_CMD, next_state, SM_RESUME_TOUT );
        
        break;
    }
}

// Public wrappers.
void CORE_stream_pause( HBD board )
{
    PVPAR v = (PVPAR) board;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;
    set_stream_state( v, STREAM_PAUSED  );
    SX16MOD_release_mutex( v->mutex_object );
}

void CORE_stream_resume( HBD board )
{
    PVPAR v = (PVPAR) board;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;
    set_stream_state( v, STREAM_RUNNING );
    SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_stream_stop( HBD board )
{
    PVPAR v = (PVPAR) board;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;
    set_stream_state( v, STREAM_HALTED  );
    SX16MOD_release_mutex( v->mutex_object ); 
}

/////////////////////////////////////////////////////////////////
// Read word from a Stream Machine register.

static USHORT sm_reg_read_516( PVPAR v, int adrs_os )
{
    return ga_indirect_reg_read( v, SMBASE_516 + adrs_os );
}

static USHORT sm_reg_read_616( PVPAR v, int adrs_os )
{
    // Execute DEBI read from SM.
    transact_debi_616( v, DEBICMD_READ( adrs_os, 2 ) );

    // Return the retrieved DEBI data.
    return (USHORT)SX16MOD_read_mem32( v->base_adrs + P_DEBIAD );
}

/////////////////////////////////////////////////////////////////
// Write word to a Stream Machine register.

static void sm_reg_write_516( PVPAR v, int adrs_os, USHORT data )
{
    ga_indirect_reg_write( v, SMBASE_516 + adrs_os, data );
}

static void sm_reg_write_616( PVPAR v, int adrs_os, USHORT data )
{
#ifdef TEST_PCI_616
	return;
#endif	
    // Set up data value to be transmitted via DEBI.
    SX16MOD_write_mem32( v->base_adrs + P_DEBIAD, data );

    // Execute DEBI write to the SM.
    transact_debi_616( v, DEBICMD_WRITE( adrs_os, 2 ) );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Upload operating parameters to SM.  If the SM is currently streaming data, the params upload will not execute.

static void smset( PVPAR v, ULONG address, ULONG data )
{
    v->sm_reg_write( v, SM_PARAM_REG, (USHORT)( ( address << 8 ) | data ) );
}

#define CA(ATTR_ID)     v->attr_current[ATTR_ID]

static void sm_upload_params( PVPAR v )
{
    // Convert video bit rate from bps to kbps, which is the way the SM wants it.
    long video_kbps = v->attr_current[ATTR_video_bitrate] / 1000;

    // get the time
#ifndef WIN32
    SX16MOD_do_gettimeofday_secs( &v->last_upload_time[0]);
    SX16MOD_do_gettimeofday_usecs( &v->last_upload_time[1]);
#endif
    // If the SM is not streaming MPEG data ...
    if ( v->stream_state == STREAM_HALTED )
    {

        SX16MOD_sleep( 99 );        // IS THIS NEEDED?

        // Program all Stream Machine control registers.

        smset( v, SMP_AUDIO_BITRATE,    CA(ATTR_audio_bitrate)                                              );  // 0x00: Audio bit rate.
        smset( v, SMP_AUDIO_FORMAT,     CA(ATTR_audio_format) << 4                                          );  // 0x01: Audio encoding format.
        smset( v, SMP_STEREO_MODE,      CA(ATTR_audio_stereo_mode) << 6                                     );  // 0x02: MPEG stereo mode.
        smset( v, SMP_AUDIO_SMPRATE,    ( CA(ATTR_audio_sample_rate) << 4 ) | CA(ATTR_audio_pcm_resolution) );  // 0x03: Audio sampling rate, PCM resolution.
        smset( v, SMP_RESERVED_04,      0x00                                                                );  // 0x04:
        smset( v, SMP_AUDIO_ADVANCED,   0xB0                                                                );  // 0x05: CRC protection, copyright, original.
        smset( v, SMP_RESERVED_06,      0x00                                                                );  // 0x06:    Reserved.
        smset( v, SMP_RESERVED_07,      0x00                                                                );  // 0x07:    Reserved.
        smset( v, SMP_FILTER,           IS_FILT ? 0x00 /* 0x70 */ : 0x00                                    );  // 0x08: Preprocessing filter enable.
        smset( v, SMP_VIDEOCTRL,        ( IS_NTSC ? 0x40 : 0x20 ) | 0x12                                    );  // 0x09: Video format, top fld first, fmt=ITU656, ADC=SAA7111, loopback enab in record mode.
        smset( v, SMP_EVEN_OFFSET,      0x00                                                                );  // 0x0A: Even field offset.
        smset( v, SMP_ODD_OFFSET,       0x00                                                                );  // 0x0B: Odd field offset.

        // 0x0C: Picture resolution.
        switch( v->attr_current[ATTR_video_picsize])
        {
            case 7: // MPEG1 - DVD mux
                smset( v, SMP_RESOLUTION,       3);
                break;
            case 8: // MPEG2- 2/3 DVD mux
                smset( v, SMP_RESOLUTION,       1); // picsize 
                break;
            default:
                smset( v, SMP_RESOLUTION, v->attr_current[ATTR_video_picsize]);
                break;
        }
        
        smset( v, SMP_H_YC_OFFSET,      0x00                                                                );  // 0x0D: Y/C offset, horizontal line offset.
        smset( v, SMP_RESERVED_0E,      0x00                                                                );  // 0x0E:    Reserved.

        smset( v, SMP_GOPLEN_IPDIST,    ( CA(ATTR_video_m) << 3 ) | CA(ATTR_video_n)                        );  // 0x0F: GOP length, I-P distance.
        smset( v, SMP_VIDEORATELSB,     video_kbps & 0xFF                                                   );  // 0x10: Video rate in kbps.
        smset( v, SMP_VIDEORATEMSB,     video_kbps >> 8                                                     );  // 0x11:
        smset( v, SMP_MOTION_SEARCH,    0xE8                                                                );  // 0x12: 64 pel P search range, 32 pel B1 search range, 32 pel B2 search range.
        smset( v, SMP_VBRMODE,          CA(ATTR_video_vbr_enable)                                           );  // 0x13: Variable bit rate enable.
        smset( v, SMP_RESERVED_14,      0x00                                                                );  // 0x14:    Reserved.
        smset( v, SMP_VIDEO_MISC,       0x07                                                                );  // 0x15: DCT frame prediction = off, Q scale = non-linear, intra VLC format, alternate scan = slanted.
        smset( v, SMP_RESERVED_16,      0x00                                                                );  // 0x16:    Reserved.
        smset( v, SMP_RESERVED_17,      0x00                                                                );  // 0x17:    Reserved.
        smset( v, SMP_TEMPORAL_SEL,     0x00                                                                );  // 0x18:
        smset( v, SMP_TEMPORAL_THRESH,  0x00                                                                );  // 0x19:
        smset( v, SMP_RESERVED_1A,      0x00                                                                );  // 0x1A:    Reserved.
        smset( v, SMP_RESERVED_1B,      0x00                                                                );  // 0x1B:    Reserved.
        smset( v, SMP_AV_SELECT,        CA(ATTR_av_select) << 5                                             );  // 0x1C: Mux A/V select, nav msg off, sector hdr disable.
        smset( v, SMP_MUX_FORMAT,       CA(ATTR_mux_format)                                                 );  // 0x1D: Mux format.
        smset( v, SMP_RESERVED_1E,      0x00                                                                );  // 0x1E:    Reserved.
        smset( v, SMP_PROGRAM_NUM,      CA(ATTR_program_number)                                             );  // 0x1F: Program number.

        smset( v, SMP_CUSTOM_MUX_20,    0x00                                                                );  // 0x20:
        smset( v, SMP_CUSTOM_MUX_21,    0x00                                                                );  // 0x21:
        smset( v, SMP_CUSTOM_MUX_22,    0x00                                                                );  // 0x22:
        smset( v, SMP_CUSTOM_MUX_23,    0x00                                                                );  // 0x23:
        smset( v, SMP_INITIAL_SCR_24,   0x00                                                                );  // 0x24:
        smset( v, SMP_INITIAL_SCR_25,   0x00                                                                );  // 0x25:
        smset( v, SMP_INITIAL_SCR_26,   0x00                                                                );  // 0x26:
        smset( v, SMP_INITIAL_SCR_27,   0x00                                                                );  // 0x27:
        smset( v, SMP_INITIAL_PTS_28,   0x00                                                                );  // 0x28:
        smset( v, SMP_INITIAL_PTS_29,   0x00                                                                );  // 0x29:
        smset( v, SMP_INITIAL_PTS_2A,   0x00                                                                );  // 0x2A:
        smset( v, SMP_INITIAL_PTS_2B,   0x00                                                                );  // 0x2B:
        smset( v, SMP_SECTOR_CTRL,      0x00                                                                );  // 0x2C:
        smset( v, SMP_SYSTEM_FORMAT,    CA(ATTR_video_system_format)                                        );  // 0x2D: System format.
        smset( v, SMP_DMA_SIZE_LSB,     ( DFLT_DMA_SIZE >> 3 ) & 0xFF                                       );  // 0x2E: DRAM FIFO chunk size LSB, in 8-byte increments.
        smset( v, SMP_DMA_SIZE_MSB,     DFLT_DMA_SIZE >> 11                                                 );  // 0x2F: DRAM FIFO chunk size MSB, in 8-byte increments.

        smset( v, SMP_RESERVED_30,      0x00                                                                );  // 0x30:    Reserved.
        smset( v, SMP_RESERVED_31,      0x00                                                                );  // 0x31:    Reserved.
        smset( v, SMP_RESERVED_32,      0x00                                                                );  // 0x32:    Reserved.
        smset( v, SMP_RESERVED_33,      0x00                                                                );  // 0x33:    Reserved.
        smset( v, SMP_RESERVED_34,      0x00                                                                );  // 0x34:    Reserved.
        smset( v, SMP_MENU_CTRL,        0x00                                                                );  // 0x35:
        smset( v, SMP_RESERVED_36,      0x00                                                                );  // 0x36:    Reserved.
        smset( v, SMP_RESERVED_37,      0x00                                                                );  // 0x37:    Reserved.

        smset( v, SMP_PARAM_CTRL,       0x01                                                                );  // 0xB8: Param control reg: must write "1" here at the end of all param uploads.

        SX16MOD_sleep( 99 );        // IS THIS NEEDED?

        // Based on SM microcode program, finish the parameters upload handshake with the SM.
        switch ( v->active_microcode )
        {
        case MPEG_PLAY:

            // SM BUG WORKAROUND ----------------
            // SM doesn't send us a params upload response when in the playback mode, so we simply delay instead
            // of waiting for a command response handshake.  We must assume the upload was successful even though
            // we have no supporting evidence that this is true.

            // Delay to allow time for the SM to process the upload, and then blindly assume the upload succeeded.
            SX16MOD_sleep( 30 );

            break;
            
        case MPEG_RECORD:

            // Loop until we get a params upload ACK from the SM or we time out while waiting for a status packet.
            for ( ; ; )
            {
                // Exit loop if packet receiver timed out; the upload failed.
                if ( !sm_get_sized_packet( v, 1, SM_PARAM_ACK_TOUT ) )
                {
                    if ( v->debug_flags & DEBUG_PARAM_UPLOAD )
                        SX16MOD_text_message( "param upload failed !!!\n" );
                    break;
                }

                // Exit loop if we got an upload ACK, otherwise loop back for the next packet.
                if ( v->sm_responses[0] == 0xFEED )
                {
                    SX16MOD_sleep( 100 );   // IS THIS NEEDED ???
                    break;
                }
            }

            break;

        default:
            
            // Can't upload parameters to SM if microcode is not yet loaded.
            if ( v->debug_flags & DEBUG_PARAM_UPLOAD )
                SX16MOD_text_message( "microcode not yet loaded\n" );

            return;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Upload microcode to Stream Machine, dir specifies program: MPEG_PLAY or MPEG_RECORD.

#define MICROCODE_UPLOAD_ACK    0x2288      // Value of first word of every microcode upload status message.
#define MICROCODE_BLOCK_SIZE    CHUNK_SIZE

static void sm_upload_microcode( PVPAR v, int microcode_id )
{
    void    *h_microcode;                   // Handle to microcode binary source.
    ULONG   byte_count;                     // Byte count for upload operation.
    void    *microcode_buf;                 // Pointer to microcode buffer.
    int     count;
    // Abort upload if target microcode is already uploaded.
    if( v->active_microcode == microcode_id )
    {
        return;
    }

    // If it is supported, enable stream loopback while SM is being reconfigured.  The caller is responsible
    // for turning off the loopback after allowing sufficient time for the SM to start the loopback stream.
    v->set_stream_bypass( v, BYPASS_ON );
        
    // Clear any pending microcode upload error flags.
    v->errflags &= ~( ERR_UCODE_NOTFOUND | ERR_UCODE_NOUPLOAD );

    // Reset the SM (so it will start loading microcode) and the audio/video PLLs.  
v->sm_reset( v );

    // Disable SM's HIU (host interface unit) interrupt output.  Although the 516's GA does not properly complement the
    // HIU_INT* signal to convert it to active high (thus rendering it useless for generating host interrupts), we
    // disable it here in anticipation of someday fixing the gate array.
    v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

    // Make the SM's internal CPU the "DMA master" so that it will autonomously schedule data transfers
    // from MPEG data RAM to the DRAM fifo and it will set the dmaEn flag when ready for a fifo transfer.
    v->sm_reg_write( v, SM_DMAMASTER_REG, 1 );

    // Wait for the SM to stabilize before starting the microcode upload.
    SX16MOD_sleep( 100 );
    
    // Initialize SM's state image to reflect its reset state.
    v->stream_state             = STREAM_HALTED;    // SM is not streaming MPEG data.
    v->active_microcode         = MPEG_NONE;        // No microcode resides on the SM.
    v->microcode_version        = 0;                // Microcode version is not yet established.
        
    // UPLOAD THE MICROCODE TO THE SM --------------------------------------------

    // Allocate a buffer for microcode transfer.
    if ( ( microcode_buf = SX16MOD_malloc( MICROCODE_BLOCK_SIZE ) ) == NULL )
    {
        v->errflags |= ERR_MEM_ALLOC;
        return;
    }

    // Try to open the required SM binary microcode source for read access, and if it can't be opened, abort the upload.
    if ( ( h_microcode = SX16MOD_open_microcode(  microcode_id, v->attr_current[ATTR_stream_type],v->attr_current[ATTR_audio_format] ) ) == NULL )
    {
        v->errflags |= ERR_UCODE_NOTFOUND;
        
        // Free the microcode buffer.
        SX16MOD_free( microcode_buf );
        return;
    }

    // Repeat until the binary microcode source is exhausted ...
    do
    {
        // Fetch a chunk of microcode into mpeg_buf[] and get the size of the chunk.  We will request
        // a chunk of size MICROCODE_BLK_SIZE; if if we get less than we ask for it is assumed we have
        // reached the end of the microcode and this is the final chunk.  We can get away with using
        // the mpeg ring buffer here because it is not needed for mpeg data when the stream is off.
        byte_count = SX16MOD_read_microcode( h_microcode, microcode_buf, MICROCODE_BLOCK_SIZE );

        // Upload the microcode chunk to the SM.
        v->sm_reg_block_write( v, SM_MICROCODE_REG, microcode_buf, byte_count );

        // Call scheduler in case we are in kernel mode; this will allow other processes to run.
        SX16MOD_sleep( 0 );
    }
    while ( byte_count == MICROCODE_BLOCK_SIZE );

    // Close the binary microcode source.
    SX16MOD_close_microcode( h_microcode );     

    // Free the microcode buffer.
    SX16MOD_free( microcode_buf );

    // RECEIVE UPLOAD ACKNOWLEDGEMENT FROM THE SM ------------------------------------------

    // Loop until we get a microcode upload ACK from the SM or we timeout while waiting for the ACK.
        count = 0;
    for ( ; ; )
    {
        count++;
        // Exit if there is no response (the upload must have failed).  The ACK message must contain exactly three words.
        if ( !sm_get_sized_packet( v, 3, SM_UCODE_ACK_TOUT ) )
        {
            // Indicate upload failed.
            v->errflags |= ERR_UCODE_NOUPLOAD;
                SX16MOD_text_message( "sm ucode upload failed !!!\n" );
            break;
        }

        // If we got a microcode upload ACK ...
        if ( v->sm_responses[0] == MICROCODE_UPLOAD_ACK )
        {
            // Save the microcode version number, which is the second word of the ACK message.
            v->microcode_version = v->sm_responses[1];

            // SM BUG WORKAROUND ----------------
            // In early playback versions, and all newer record and playback microcode versions, the SM doesn't send us 0x0000
            // in the 3rd word of the microcode upload response (as the spec sheet says it should).  Consequently, we can't use
            // the 3rd word to detect whether the SM likes the microcode we sent to it, and we can only assume that microcode
            // was successfully installed.
            v->sm_responses[2] = 0;         // SM bug workaround: force 3rd response word to zero to indicate upload success.
                
            // If the microcode was successfully uploaded to the SM ...
            if ( v->sm_responses[2] == 0 )
            {

                    // sleep BEFORE setting the microcode active
                    // IS THIS NEEDED?
                    SX16MOD_sleep( 100 );

                    // Indicate microcode has been uploaded.
                    
                    v->active_microcode = microcode_id;
                    v->last_err[IDX_INFO_FW]++;
                    
                    // We are finished uploading the microcode, so exit.
                    break;
                }
        }
        if(count > 5) {
            v->errflags |= ERR_UCODE_NOUPLOAD;
                dbgprint("upload error\n");
                break;
        }
        
    }

}

/////////////////////////////////////////////////////////////////////////////
// Configure the SM's MPEG stream direction.

static void set_mpeg_mode( PVPAR v, int microcode_id )
{
    // If necessary, reset the SM and upload new microcode.
    sm_upload_microcode( v, microcode_id );

    // Upload all runtime configuration parameters to the SM.
    sm_upload_params( v );
}

////////////////////////////////////////////////////////////////////////////////////
// Write block of MPEG data to the SM.

// 516: Use GA indirect register access interface.
static void sm_mpeg_write_516( PVPAR v, USHORT *pdata, UINT nbytes )
{
    // Switch to big endian byte ordering for SM MPEG stream portal accesses.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_ON );

    // Set up GA's indirect address register so it points to the SM's mpeg data port.
    ga_set_indirect_adrs( v, SM_DRAM_REG_516 );

    // Copy one chunk of stream data from our mpeg buffer to the SM using block transfer of words.
    SX16MOD_block_write_io16( v->base_adrs + G_DATAIO, pdata, nbytes >> 1 );    // word count

    // Restore little endian byte ordering for SM register accesses.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_OFF );
}

// 616: Use SAA7146 DEBI interface.
static void sm_mpeg_write_616( PVPAR v, USHORT *pdata, UINT nbytes )
{
    // Switch to big endian byte ordering for SM MPEG stream portal accesses.
    SX16MOD_write_mem32( v->base_adrs + P_DEBICFG, DEBICFG_MPEG_ACCESS );

    // Copy data from our mpeg buffer to the SM using block transfer of words.
    sm_reg_block_write_616_mpeg( v, SM_DRAM_REG_616, pdata, nbytes );

    // Restore little endian byte ordering for SM register accesses.
    SX16MOD_write_mem32( v->base_adrs + P_DEBICFG, DEBICFG_REG_ACCESS );
}

////////////////////////////////////////////////////////////////////////////////////
// Read block of MPEG data from the SM.

// Use GA indirect register access interface.
static void sm_mpeg_read_516( PVPAR v, USHORT *pdata, UINT nbytes )
{
    // Switch to big endian byte ordering for SM MPEG stream portal accesses.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_ON );

    // Set up GA's indirect address register so it points to the SM's mpeg data port.
    ga_set_indirect_adrs( v, SM_DRAM_REG_516 );

    // Fetch one chunk of stream data from SM into our mpeg buffer using block transfer of words.
    SX16MOD_block_read_io16( v->base_adrs + G_DATAIO, pdata, nbytes >> 1 );     // word count

    // Restore little endian byte ordering for SM register accesses.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_OFF );
}

// Use SAA7146 DEBI interface.  Must transfer at least 3 words for DEBI block transfer!
static void sm_mpeg_read_616( PVPAR v, USHORT *pdata, UINT nbytes )
{
    // Switch to big endian byte ordering for SM MPEG stream portal accesses.
    SX16MOD_write_mem32( v->base_adrs + P_DEBICFG, DEBICFG_MPEG_ACCESS );

    // Copy data from SM to our mpeg buffer using block transfer of words.
    sm_reg_block_read_616( v, SM_DRAM_REG_616, pdata, nbytes );

    // Restore little endian byte ordering for SM register accesses.
    SX16MOD_write_mem32( v->base_adrs + P_DEBICFG, DEBICFG_REG_ACCESS );
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Save/restore hardware context so we can exert hardware control from a secondary polling thread.

static void context_save_516( PVPAR v )
{
    // Save the GA's current indirect register address.
    v->saved_lastaddr = v->lastaddr;
}

static void context_restore_516( PVPAR v )
{
    // Restore the GA's indirect register address.
    ga_set_indirect_adrs( v, v->saved_lastaddr );
}

static void vfunc_dummy_v( PVPAR v )
{
    // NOP - used for 616, which doesn't require hardware context save.
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// This function resets the interrupt and needs to be called from the interrupt handler for
// each interrupt.
// Return value: set of bit flags, one for each interrupt source.

int CORE_reset_interrupt( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    int sts, res = 0;

    // If the board is not open do nothing and return immediately.
    if ( !v->is_open || !v->irq)
        return 0;
    if( v->product_id != 616)
      return 1;
    // Read status reg.
    sts = SX16MOD_read_mem32( v->base_adrs + P_ISR );

    // Reset status reg.
    SX16MOD_write_mem32( v->base_adrs + P_ISR, sts );

    // Check status bits.
    if(sts & ISR_GPIO0)
        res |= INTERRUPT_STREAM;
    if(sts & ISR_RPS_I0)
        res |= INTERRUPT_VIDEO;
    return res;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
int CORE_acquire_mutex( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    return SX16MOD_acquire_mutex( v->mutex_object );
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// Core mutex release

void CORE_release_mutex( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    SX16MOD_release_mutex( v->mutex_object );
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// Video output shut off control

void CORE_encoder_out_ctrl( HBD board, int enable )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    
    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;

    if ( enable )
        eset( v, E_STANDARD_CTRL, IS_NTSC ? 0x15 : 0x16 );
    else
        eset( v, E_STANDARD_CTRL, IS_NTSC ? 0x55 : 0x56 );
    
    SX16MOD_release_mutex( v->mutex_object );

}


/////////////////////////////////////////////////////////////////////////////////////////////////
// Returns true if there is a pending DMA cycle request. Preriodically called from
// within a polling thread.

BOOL CORE_poll_dma_flag( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    int dma_flag;

    // If the board is not open do nothing and return immediately.
    if ( !v->is_open )
        return FALSE;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return FALSE;

    // Get dma flag.
    dma_flag = v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST;

    SX16MOD_release_mutex( v->mutex_object );

    return dma_flag;
}


/////////////////////////////////////////////////////////////////////////////////////////////
// Start the MPEG encoder if it isn't already running.  Call this before ReadStream().

static INLINE void start_encode( PVPAR v )
{
    // If MPEG stream is not already running ...
    if ( !stream_is_encoding( v ) )
    {
        //dbgprint("STREAM ENCODING %d\n", v->overlay_started);
        // DA: Oct 27, 2006.  race condition on startup.
        // DA: fixes overlay stop race condition when starting encoding app and capture app at same time
        // Workaround was previously to "move" the Overlay window to restart it.
        // Stop must be done because re-routed below.  If stop is not done here, then the capture/overlay window
        // will have very serious pink or garbage glitches.
        // capture with video buf works differently.  Capture will be stopped, but not restarted.
        // Capture timesout in the videobuf qs.  Reactivating is done through the videobuf framework.
      if( v->product_id == 616) {
        if( v->overlay_started || v->capture_started)
        {
            // turn off the RPS, DMA while loading
            SX16MOD_write_mem32( v->base_adrs + P_MC1, DSBL_RPS);
            v->cap_frozen = 1;
        }
        if( v->clip_started)
        {
            dbgprint("pausing clip\n");
            SX16MOD_write_mem32( v->base_adrs + P_MC2, 1 << 5 | 1 << 21);
            // disable video dma2
            SX16MOD_write_mem32( v->base_adrs + P_MC1, 1 << 21);
        }
      }

        //PRINTK("sx16core: start_encode\n");
        // Set mpeg byte count to indicate no data to transfer.
        v->num_dma_bytes = DFLT_DMA_SIZE;
        
        // Close the 7113's horizontal PLL; the horizontal frequency must be locked to the video input source.
//      dset( board, D_SYNC_CONTROL, 0x98 | ( ( v->system == SYS_NTSC ) << 6 ) | ( v->enc_loaded << 2 ), VAL_RUNTIME );
//      dset( board, D_SYNC_CONTROL, 0x18 | ( ( v->system == SYS_NTSC ) << 6 ) | ( v->enc_loaded << 2 ), VAL_RUNTIME );

        // Upload Encoder (record) microcode to SM if it is not already loaded, upload parameters, enable gate array loopback.
        set_mpeg_mode( v, MPEG_RECORD );

        // Command the SM to start the MPEG output stream if it is halted.
        set_stream_state( v, STREAM_RUNNING );

        // Disable stream bypass so SM can output the video.
        v->set_stream_bypass( v, BYPASS_OFF );

        // Disable SM DMA interrupt
        v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

        // Stream start time.
        v->stream_start_time = SX16MOD_time_now();
    
        // if vm overlay was started prior, restart it
        // Don't restart capture as this will be done through the videobuf frame work
        if( v->product_id == 616) {
        if(v->overlay_started)
        {
            // 
            SX16MOD_write_mem32( v->base_adrs + P_RPSADDR0, v->vpar_dma_cap.caprps);
            SX16MOD_write_mem32( v->base_adrs + P_RPSTO0, 0);                           // select port a
            SX16MOD_write_mem32( v->base_adrs + P_MC1, EN_DMA1_RPS0_OVL);
        }
        if( v->clip_started)
        {
            // start clip again
            SX16MOD_write_mem32( v->base_adrs + P_MC2, 1 << 5 | 1 << 21 | 1 << 3 | 1 << 19);
            SX16MOD_write_mem32( v->base_adrs + P_MC1, 1 << 5 | 1 << 21);
        }
        }
        v->cap_frozen = 0;
    }
    v->pend_encode = 0;
}

int CORE_get_pend_encode( HBD board) 
{
    PVPAR v = (PVPAR)board;
    return v->pend_encode;
}
// Public wrapper.
void CORE_start_encode( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;
    // get main mutex first
    if( SX16MOD_acquire_mutex( v->mutex_object ) < 0)
      return;
    v->b_upload = 1;
    start_encode( v );
    v->b_upload = 0;
    SX16MOD_release_mutex( v->mutex_object ); 
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// Start the MPEG decoder if it isn't already running.  Call this at some point before WriteStream().

static INLINE void start_decode( PVPAR v )
{
    // If MPEG stream is not already running ...
    if ( !stream_is_decoding( v ) )
    {
        // Set mpeg byte count to indicate no data to transfer.
        v->num_dma_bytes = DFLT_DMA_SIZE;

        // Upload Decoder (playback) microcode to SM if it is not already loaded, upload parameters, enable gate array loopback.
        set_mpeg_mode( v, MPEG_PLAY );

        // Open the 7113's horizontal PLL; the horizontal frequency is fixed in Play mode.
//      dset( board, D_SYNC_CONTROL, 0x25 | ( ( v->system == SYS_NTSC ) << 6 ) | 0x04, VAL_RUNTIME );
        eset( v, E_RTCEN_BURSTAMPL, IS_NTSC ? 0x63 : 0x45 );    // 0x62: Color burst amplitude. ~~Turn off RTCE for playback (to work w/o input video)

        // Command the SM to start processing the MPEG input stream if it is not already doing it.
        set_stream_state( v, STREAM_RUNNING );

        // Disable stream bypass so SM can output the video.
        v->set_stream_bypass( v, BYPASS_OFF );

        // Disable SM DMA interrupt
        v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

        // Stream start time.
        v->stream_start_time = SX16MOD_time_now();
    }
}

// Public wrapper.
void CORE_start_decode( HBD board )
{
    // Cache pointer to board's object.
    PVPAR v = (PVPAR)board;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;
    start_decode( v );
    SX16MOD_release_mutex( v->mutex_object ); 
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// RECORD MODE: copy up to nbytes of compressed video mpeg stream data from the SM
// into the user's buffer.  Returns number of bytes transferred, or zero if no bytes are available.
// Returns bytes requested or maximum remaining in dma cycle, whichever is less.

static INLINE int stream_read( PVPAR v, UCHAR *buffer, ULONG nbytes )
{
    ULONG bytes_transfered = 0; // Bytes transfered to/from SM
    ULONG bytes_to_transfer;    // Bytes to be transfered to/from SM.
    int chunk_size = CHUNK_SIZE;

    if( v->attr_current[ATTR_video_bitrate] > 4000000) {
        /* use a higher chunk size at the higher bitrates */
        chunk_size*=2;
    }

    
    // While bytes transfered is less than the buffer length...
    while ( bytes_transfered < nbytes )
    {
        // If we have reached the end of the dma transfer cycle...
        if ( v->num_dma_bytes == DFLT_DMA_SIZE )
        {
            // If dma flag is set, reset transfer count to zero.
            if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) != 0 )
                v->num_dma_bytes = 0;
        }

        // If dma flag is not set. make sure interrupt is enabled and exit.
        if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) == 0 )
        {
            //PRINTK("sx16core: stream_read: dma flag not set, bytes %d\n",
            //bytes_transfered);
            // Enable interrupt, if present.
            if ( v->irq )
                v->sm_reg_write( v, SM_INTCTRL_REG, SMICTL_DMA_ENABLE );
            break;
        }
        // If dma flag is set, disable further interrupts.
        else if ( v->irq )
            v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

        // If we need to tranfer more data than there is available is the dma cycle
        // adjust the transfer size so that is doesn't exceed the dma cycle.
        if ( ( v->num_dma_bytes + nbytes - bytes_transfered ) > DFLT_DMA_SIZE )
            bytes_to_transfer = DFLT_DMA_SIZE - v->num_dma_bytes;
        else
            bytes_to_transfer = nbytes - bytes_transfered;

        // Transfer data in chunk_size chunks...
        while ( bytes_to_transfer >= chunk_size  && ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) )
        {
            // Copy data chunk from SM to buffer.
            v->sm_mpeg_read( v, (USHORT *)&buffer[bytes_transfered], chunk_size );

            // Update counters
            bytes_to_transfer   -= chunk_size;
            bytes_transfered    += chunk_size;
            v->num_dma_bytes    += chunk_size;
        }

		// Transfer remaining data...
		if ( bytes_to_transfer > 0 && bytes_to_transfer < chunk_size  && ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) )
		{
			// Copy leftovers from SM to buffer.
			v->sm_mpeg_read( v, (USHORT *)&buffer[bytes_transfered], bytes_to_transfer );

			// Update counters
			v->num_dma_bytes 	+= bytes_to_transfer;
			bytes_transfered	+= bytes_to_transfer;
		}
	}

	// Return the number of bytes actually transfered.
	return bytes_transfered;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// RECORD MODE: copy up to nbytes of compressed video mpeg stream data from the SM into the user's
// buffer.

int CORE_stream_read( HBD board, UCHAR *buffer, ULONG nbytes )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
	{
		//PRINTK("acquire mutex\n");
		return 0;
	}
	// Start the mpeg encoder if it is not already running.
	start_encode( v );

	// Empty the status packet queue.
	sm_get_status_packet( v, 0 );

	// Do the SM read.
	nbytes = stream_read( v, buffer, nbytes );

	SX16MOD_release_mutex( v->mutex_object ); 

	// Return number of bytes transferred.
	return nbytes;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// PLAY MODE: copy up to nbytes of compressed video mpeg stream data to the SM
// from the user's buffer.  Returns number of bytes transferred, or zero if no bytes are available.
// Returns bytes requested or maximum remaining in dma cycle, whichever is less.

static INLINE int stream_write( PVPAR v, const UCHAR *buffer, ULONG nbytes )
{
	ULONG bytes_transfered = 0;	// Bytes transfered to/from SM
	ULONG bytes_to_transfer;	// Bytes to be transfered to/from SM.

	// While bytes transfered is less than the buffer length...
	while ( bytes_transfered < nbytes )
	{
		// If we have reached the end of the dma transfer cycle...
		if ( v->num_dma_bytes >= DFLT_DMA_SIZE )
		{
			// If dma flag is set, reset transfer count to zero.
			if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) != 0 )
				v->num_dma_bytes = 0;
		}

		// If dma flag is not set. make sure interrupt is enabled and exit.
		if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) == 0 )
		{
			// Enable interrupt, if present.
			if ( v->irq )
				v->sm_reg_write( v, SM_INTCTRL_REG, SMICTL_DMA_ENABLE );
			break;
		}
		// If dma flag is set, disable further interrupts.
		else if ( v->irq )
			v->sm_reg_write( v, SM_INTCTRL_REG, 0 );

		// If we need to tranfer more data than there is available is the dma cycle
		// adjust the transfer size so that is doesn't exceed the dma cycle.
		if ( ( v->num_dma_bytes + nbytes - bytes_transfered ) > DFLT_DMA_SIZE )
			bytes_to_transfer = DFLT_DMA_SIZE - v->num_dma_bytes;
		else
			bytes_to_transfer = nbytes - bytes_transfered;

		// Transfer data in CHUNK_SIZE_PLAY chunks...
		while ( bytes_to_transfer >= CHUNK_SIZE_PLAY  && ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) )
		{
			// Copy data chunk to SM from buffer.
			v->sm_mpeg_write( v, (USHORT *)&buffer[bytes_transfered], CHUNK_SIZE_PLAY );

			// Update counters
			bytes_to_transfer	-= CHUNK_SIZE_PLAY;
			bytes_transfered	+= CHUNK_SIZE_PLAY;
			v->num_dma_bytes 	+= CHUNK_SIZE_PLAY;
		}

		// Transfer remaining data...
		if ( bytes_to_transfer > 0 && bytes_to_transfer < CHUNK_SIZE_PLAY  && ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) )
		{
			// Copy leftovers to SM from buffer.
			v->sm_mpeg_write( v, (USHORT *)&buffer[bytes_transfered], bytes_to_transfer );

			// Update counters
			v->num_dma_bytes 	+= bytes_to_transfer;
			bytes_transfered	+= bytes_to_transfer;
		}
	}
	// Return the number of bytes actually transfered.
	return bytes_transfered;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// PLAY MODE: copy up to nbytes of compressed video mpeg stream data to the SM from the user's
// buffer.

int CORE_stream_write( HBD board, const UCHAR *buffer, ULONG nbytes )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return 0;

	// Start the mpeg decoder if it is not already running.
	start_decode( v );

	// Empty the status packet queue.
	sm_get_status_packet( v, 0 );

	// Do the SM write.
	nbytes = stream_write( v, buffer, nbytes );

	SX16MOD_release_mutex( v->mutex_object ); 

	// Return number of bytes transferred.
	return nbytes;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////  VIDEO OVERLAY  ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////
// Write parameter to overlay interface.
/*
static void ovset ( struct PVPAR v, ULONG address, ULONG data )
{
	v->i2c_write( v, I2CADDR_OVR_616, address, data );
}
*/
/////////////////////////////////////////////////////////
// Return overlay dimensions in pixels.

unsigned long CORE_overlay_get_width( HBD board )
{
	return 256;
}

unsigned long CORE_overlay_get_height( HBD board )
{
	return 32;
}

///////////////////////////////////////////////////////////
// Set overlay enable/disable: 0-off, 1-on.

void CORE_overlay_set_enable( HBD board, int enable )
{
//	struct PVPAR v = (PVPAR)board;
//	ovset( v, 0, v->mode );			   // 616 CODE -- Register 0: mode
}

//////////////////////////////////////////////////////////////
// Set overlay position in pixels (0,0 if upper left).

static void overlay_setpos( PVPAR v )
{
/*
	if ( BOARDTYPE == 616 )
	{
		ovset( v, 1, v->xstart ); 			// Register 1: xstart ( lo )
		ovset( v, 2, v->xstart >> 8 );		// Register 2: xstart ( hi )
		ovset( v, 3, v->ystart ); 			// Register 3: ystart ( lo )
		ovset( v, 4, v->ystart >> 8 );		// Register 4: ystart ( hi )
		ovset( v, 5, v->xend );				// Register 5: xend ( lo )
		ovset( v, 6, v->xend >> 8 );		// Register 6: xend ( hi )
		ovset( v, 7, v->yend );				// Register 7: yend ( lo )
		ovset( v, 8, v->yend >> 8 );		// Register 8: yend ( hi )
	}
*/
}

void CORE_overlay_set_xstart( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_position.xstart = val;
	overlay_setpos( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_ystart( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_position.ystart = val;
	overlay_setpos( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_xend( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_position.xend = val;
	overlay_setpos( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_yend( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_position.yend = val;
	overlay_setpos( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

//////////////////////////////////////////////////////////////////////////////////////////
// Set overlay colors.
// yfirst: Y - luminance component for first pixel after background.
// ynext:  Y - luminance component for next pixels after first.
// c1,c2:  C - chrominance components.

static void overlay_setcolors( PVPAR v )
{
/*
	if ( BOARDTYPE == 616 )
	{
		ovset( v, 9, v->yfirst );			// Register 9: yfirst
		ovset( v, 10, v->ynext );			// Register 10: ynext
		ovset( v, 11, v->c1 );				// Register 11: c1
		ovset( v, 12, v->c2 );				// Register 12: c2
	}
	else
*/
}

void CORE_overlay_set_yfirst( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_colors.yfirst = val;
	overlay_setcolors( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_ynext( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_colors.ynext = val;
	overlay_setcolors( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_c1( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_colors.c1 = val;
	overlay_setcolors( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_c2( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_colors.c2 = val;
	overlay_setcolors( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

//////////////////////////////////////////////////////////////////////////////////////////
// Write 8-bit data to overlay.  Return error code.
// row:    y position of byte
// column: x position of pixel

static void overlay_write( PVPAR v )
{
/*
	if ( BOARDTYPE == 616 )

	{
		if ( ( v->row < 0 ) || ( v->row > 3 ) || ( v->column < 0 ) || ( v->column > 255 ) )
			return ( ERR_ILLEGAL_INDEX );
		else
			v->i2c_write( v, I2CADDR_OVR_616 + ( v->row << 1 ) + 8, v->column, v->data );
	}
*/
}

void CORE_overlay_set_row( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_write.row = val;
	overlay_write( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_column( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_write.column = val;
	overlay_write( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}

void CORE_overlay_set_data( HBD board, ULONG val )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;

	v->overlay_write.data = val;
	overlay_write( v );

	SX16MOD_release_mutex( v->mutex_object ); 
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////  VIDEO CAPTION  //////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Calculate position parameters: x-stop and y-stop, for 516 caption.

static void captn_pos_calc_516( PVPAR v, void *pcapprms )
{
	BUF_CAPTN_PARAM *pcapparams = pcapprms;
	int txtlen = 0;
	int scale;

	txtlen = cstrlen( pcapparams->text );
	scale  = pcapparams->scale;

	switch ( scale )
	{
	case OVL_SCALE_X1:		// x1
	case OVL_SCALE_X2:		// x2
	case OVL_SCALE_X4:		// x4
		pcapparams->xstop = pcapparams->xstart + (scale+1)*8*txtlen;			// refer to HW spec, 8x16/char
		pcapparams->ystop = pcapparams->ystart + (scale+1)*16*(txtlen==0? 0:1);	// for the calculation.
		break;
	case OVL_SCALE_X2V:		// x2 (only vertical)
		pcapparams->xstop = pcapparams->xstart + (scale-1)*8*txtlen;			// refer to HW spec, 8x16/char
		pcapparams->ystop = pcapparams->ystart + scale*16*(txtlen==0? 0:1);		// for the calculation.
		break;
	default:
		pcapparams->xstop = pcapparams->xstart;
		pcapparams->ystop = pcapparams->ystart;
		break;
	}

	// trim xstop and ystop
	if ( pcapparams->xstop > MAX_XSTOP ) pcapparams->xstop = MAX_XSTOP;
	if ( v->attr_current[ATTR_video_system] == SYS_NTSC )
	{
		if ( pcapparams->ystop > MAX_YSTOP_NTSC ) pcapparams->ystop = MAX_YSTOP_NTSC;
	} else {
		if ( pcapparams->ystop > MAX_YSTOP_PAL ) pcapparams->ystop = MAX_YSTOP_PAL;
	}

}

////////////////////////////////////////////////////////////////////////
// Load caption parameters down to Regs and RAM-blocks in 516 Gate Array

static void captn_param_load_516( PVPAR v, void *pcapprms, ULONG action_id )
{
	static BUF_CAPTN_PARAM loaded_param[2] = {  { 1, 0, 0, 0, 0, 0, 0, 0, " \0", 0 },
												{ 2, 0, 0, 0, 0, 0, 0, 0, " \0", 0 }  };
	BUF_CAPTN_PARAM *pcapparams = pcapprms;
	USHORT			pos_x, pos_y;
	USHORT			attr = 0;
	USHORT			attrval, charpixel;
	int				txtlen		= 0;
	int				loadedlen	= 0;
	int				i, j;
	int				ascii_index;

	//PRINTK( "captn_param_load_516: action_id = %lu, window # %d = %d on/off\n",
	//								action_id, pcapparams->window, pcapparams->on );	// for debugging
	//PRINTK( "captn_param_load_516: xstart = %d, xstop = %d \n",  
	//								pcapparams->xstart, pcapparams->xstop );			// for debugging
	//PRINTK( "captn_param_load_516: ystart = %d, ystop = %d \n",  
	//								pcapparams->ystart, pcapparams->ystop );			// for debugging
	//PRINTK( "captn_param_load_516: scale = %d, color = %d, transp = %d \n",
	//								pcapparams->scale, pcapparams->color, pcapparams->transp );	// for debugging

	switch (action_id)
	{
	case CAPTN_window:
		break;
	case CAPTN_xstart:
	case CAPTN_xstop:
		pos_x = (pcapparams->xstart >= 4) ? pcapparams->xstop >> 2 : 1;						// refer to HW spec:
		pos_x = (pos_x << 8) + ( (pcapparams->xstart >= 4) ? pcapparams->xstart >> 2 : 1 );	// divided by 4 for x_start and x_stop
		ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1X : G_CAPWINPOSN_2X, pos_x );
		loaded_param[pcapparams->window - 1].xstart = pcapparams->xstart;
		loaded_param[pcapparams->window - 1].xstop  = pcapparams->xstop;
		break;
	case CAPTN_ystart:
	case CAPTN_ystop:
		pos_y = pcapparams->ystop >> 1;						// refer to HW spec:
		pos_y = (pos_y << 8) + (pcapparams->ystart >> 1);	// divided by 2 for y_start and y_stop
		ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1Y : G_CAPWINPOSN_2Y, pos_y );
		loaded_param[pcapparams->window - 1].ystart = pcapparams->ystart;
		loaded_param[pcapparams->window - 1].ystop  = pcapparams->ystop;
		break;
	case CAPTN_scale:
		pos_x = (pcapparams->xstart >= 4) ? pcapparams->xstop >> 2 : 1;						// refer to HW spec:
		pos_x = (pos_x << 8) + ( (pcapparams->xstart >= 4) ? pcapparams->xstart >> 2 : 1 );	// divided by 4 for x_start and x_stop
		ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1X : G_CAPWINPOSN_2X, pos_x );
		pos_y = pcapparams->ystop >> 1;						// refer to HW spec:
		pos_y = (pos_y << 8) + (pcapparams->ystart >> 1);	// divided by 2 for y_start and y_stop
		ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1Y : G_CAPWINPOSN_2Y, pos_y );
		ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINSCALE_1 : G_CAPWINSCALE_2, pcapparams->scale );
		loaded_param[pcapparams->window - 1].scale = pcapparams->scale;
		loaded_param[pcapparams->window - 1].xstop = pcapparams->xstop;
		loaded_param[pcapparams->window - 1].ystop = pcapparams->ystop;
		break;
	case CAPTN_color:
	case CAPTN_transp:
		txtlen = cstrlen( pcapparams->text );
		// change char-font, if opposite font is need.
		if ( ( ( pcapparams->transp > OVL_TRANSP_BBKGND ) && ( loaded_param[pcapparams->window - 1].transp <= OVL_TRANSP_BBKGND ) ) || 
			 ( ( pcapparams->transp < OVL_TRANSP_T_BKGND ) && ( loaded_param[pcapparams->window - 1].transp >= OVL_TRANSP_T_BKGND ) ) )
		{
			for ( i = 0; i < txtlen; i++ )
			{
				ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( ( pcapparams->window == 1 ) ? G_CAPRAMCHAR_1 : G_CAPRAMCHAR_2 ) + i*8 );
				for ( j = 0; j <8; j++ )
				{
					ascii_index = (int) *( pcapparams->text + i );
					charpixel = ( fnt_516[ascii_index][2*j] << 8 ) + fnt_516[ascii_index][2*j+1];
					if ( pcapparams->transp == OVL_TRANSP_T_BKGND || pcapparams->transp == OVL_TRANSP_B_BKGRD ) charpixel = ~charpixel;
					ga_indirect_reg_write ( v, G_CAPRAMDATA, charpixel );
					//PRINTK( "captn_param_load_516: ----- char = %c ( %i ) \t charpixel = 0x%x \n", 
					//											*( pcapparams->text + i ), ascii_index, charpixel );	// for debugging
				}
			}
		}
		// make up attributes.
		attr = (~pcapparams->transp) << 15;
		switch (pcapparams->color)		// refer to HW spec.
		{
		case OVL_WHITE:					// 0 -- white:   with 001
			attr = attr + 0x1240;		//			0+001+001+001+000000; 
			break;
		case OVL_BLACK:					// 1 -- black:   with 000
			attr = attr + 0x0000;		//			0+000+000+000+000000; 
			break;
		case OVL_RED:					// 2 -- red:     with 010
			attr = attr + 0x2480;		//			0+010+010+010+000000; 
			break;
		case OVL_GREEN:					// 3 -- green:   with 100
			attr = attr + 0x4900;		//			0+100+100+100+000000; 
			break;
		case OVL_BLUE:					// 4 -- blue:    with 011
			attr = attr + 0x36c0;		//			0+011+011+011+000000; 
			break;
		case OVL_YELLOW:				// 5 -- yellow:  with 101
			attr = attr + 0x5b40;		//			0+101+101+101+000000; 
			break;
		case OVL_CYAN:					// 6 -- cyan:    with 111
			attr = attr + 0x7fc0;		//			0+111+111+111+000000;
			break;
		case OVL_MAGENTA:				// 7 -- magenta: with 110
			attr = attr + 0x6d80;		//			0+110+110+110+000000;
			break;
		default:
			return;
		}
		ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( pcapparams->window == 1 ) ? G_CAPRAMATTR_1 : G_CAPRAMATTR_2 );
		for ( i = 0; i < txtlen; i++ )
		{
			attrval = attr + ( ( pcapparams->window == 2 ) ? 32 : 0 ) + (USHORT) i;
			ga_indirect_reg_write ( v, G_CAPRAMDATA, attrval );
			//PRINTK( "captn_param_load_516: ----- window = %d : attrval = 0x%x ( i = %d )\n", 
			//												pcapparams->window, attrval, i );	// for debugging
		}
		loaded_param[pcapparams->window - 1].color  = pcapparams->color;
		loaded_param[pcapparams->window - 1].transp = pcapparams->transp;
		break;
	case CAPTN_text:
		txtlen    = cstrlen( pcapparams->text );
		loadedlen = cstrlen( loaded_param[pcapparams->window - 1].text );
		for ( i = 0; i < txtlen; i++ )
		{
			if ( i < loadedlen )
			{
				if ( *( pcapparams->text + i ) != *( loaded_param[pcapparams->window - 1].text + i ) )
				{
					// load font to RAM block in GA516
					ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( ( pcapparams->window == 1 ) ? G_CAPRAMCHAR_1 : G_CAPRAMCHAR_2 ) + i*8 );
					for ( j = 0; j <8; j++ )
					{
						ascii_index = (int) *( pcapparams->text + i );
						charpixel = ( fnt_516[ascii_index][2*j] << 8 ) + fnt_516[ascii_index][2*j+1];
						if ( pcapparams->transp == OVL_TRANSP_T_BKGND || pcapparams->transp == OVL_TRANSP_B_BKGRD ) charpixel = ~charpixel;
						ga_indirect_reg_write ( v, G_CAPRAMDATA, charpixel );
						//PRINTK( "captn_param_load_516: ----- char = %c ( %i ) \t charpixel = 0x%x \n", 
						//											*( pcapparams->text + i ), ascii_index, charpixel );	// for debugging
					}
				}
			} else {
				// load font to RAM block in GA516
				ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( ( pcapparams->window == 1 ) ? G_CAPRAMCHAR_1 : G_CAPRAMCHAR_2 ) + i*8 );
				for ( j = 0; j <8; j++ )
				{
					ascii_index = (int) *( pcapparams->text + i );
					charpixel = ( fnt_516[ascii_index][2*j] << 8 ) + fnt_516[ascii_index][2*j+1];
					if ( pcapparams->transp == OVL_TRANSP_T_BKGND || pcapparams->transp == OVL_TRANSP_B_BKGRD ) charpixel = ~charpixel;
					ga_indirect_reg_write ( v, G_CAPRAMDATA, charpixel );
					//PRINTK( "captn_param_load_516: ----- char = %c ( %i ) \t charpixel = 0x%x \n", 
					//											*( pcapparams->text + i ), ascii_index, charpixel );	// for debugging
				}
				if ( i == loadedlen ) 
				{
					// make up attributes.
					attr = (~pcapparams->transp) << 15;
					switch (pcapparams->color)		// refer to HW spec.
					{
					case OVL_WHITE:					// 0 -- white:   with 001
						attr = attr + 0x1240;		//			0+001+001+001+000000; 
						break;
					case OVL_BLACK:					// 1 -- black:   with 000
						attr = attr + 0x0000;		//			0+000+000+000+000000; 
						break;
					case OVL_RED:					// 2 -- red:     with 010
						attr = attr + 0x2480;		//			0+010+010+010+000000; 
						break;
					case OVL_GREEN:					// 3 -- green:   with 100
						attr = attr + 0x4900;		//			0+100+100+100+000000; 
						break;
					case OVL_BLUE:					// 4 -- blue:    with 011
						attr = attr + 0x36c0;		//			0+011+011+011+000000; 
						break;
					case OVL_YELLOW:				// 5 -- yellow:  with 101
						attr = attr + 0x5b40;		//			0+101+101+101+000000; 
						break;
					case OVL_CYAN:					// 6 -- cyan:    with 111
						attr = attr + 0x7fc0;		//			0+111+111+111+000000; 
						break;
					case OVL_MAGENTA:				// 7 -- magenta: with 110
						attr = attr + 0x6d80;		//			0+110+110+110+000000; 
						break;
					default:
						return;
					}
				}
				ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( pcapparams->window == 1 ) ? (G_CAPRAMATTR_1+i) : (G_CAPRAMATTR_2+i) );
				attrval = attr + ( ( pcapparams->window == 2 ) ? 32 : 0 ) + (USHORT) i;
				ga_indirect_reg_write ( v, G_CAPRAMDATA, attrval );
				//PRINTK( "captn_param_load_516: ----- window = %d : attrval = 0x%x ( i = %d )\n", 
				//												pcapparams->window, attrval, i );	// for debugging
			}
		}
		// reload xstop and ystop, if text length changed.
		if ( txtlen != loadedlen ) {
			captn_param_load_516 ( v, pcapparams, CAPTN_xstop );
			captn_param_load_516 ( v, pcapparams, CAPTN_ystop );
		}
		// update loaded text.
		//strcpy ( loaded_param[pcapparams->window - 1].text, pcapparams->text );
		cmemcpy( loaded_param[pcapparams->window - 1].text, pcapparams->text, sizeof(pcapparams->text) );
		break;
	case CAPTN_on:
		if ( pcapparams->on == CAPTION_ON )
		{
			// load xstart and xstop.
			pos_x = (pcapparams->xstart >= 4) ? pcapparams->xstop >> 2 : 1;						// refer to HW spec:
			pos_x = (pos_x << 8) + ( (pcapparams->xstart >= 4) ? pcapparams->xstart >> 2 : 1 );	// divided by 4 for x_start and x_stop
			ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1X : G_CAPWINPOSN_2X, pos_x );
			loaded_param[pcapparams->window - 1].xstart = pcapparams->xstart;
			loaded_param[pcapparams->window - 1].xstop  = pcapparams->xstop;
			// load ystart and ystop.
			pos_y = pcapparams->ystop >> 1;						// refer to HW spec:
			pos_y = (pos_y << 8) + (pcapparams->ystart >> 1);	// divided by 2 for y_start and y_stop
			ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1Y : G_CAPWINPOSN_2Y, pos_y );
			loaded_param[pcapparams->window - 1].ystart = pcapparams->ystart;
			loaded_param[pcapparams->window - 1].ystop  = pcapparams->ystop;
			// load scale.
			ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINSCALE_1 : G_CAPWINSCALE_2, pcapparams->scale );
			loaded_param[pcapparams->window - 1].scale = pcapparams->scale;
			// load text font into RAM.
			txtlen = cstrlen( pcapparams->text );
			for ( i = 0; i < txtlen; i++ )
			{
				ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( ( pcapparams->window == 1 ) ? G_CAPRAMCHAR_1 : G_CAPRAMCHAR_2 ) + i*8 );
				for ( j = 0; j <8; j++ )
				{
					ascii_index = (int) *( pcapparams->text + i );
					charpixel = ( fnt_516[ascii_index][2*j] << 8 ) + fnt_516[ascii_index][2*j+1];
					if ( pcapparams->transp == OVL_TRANSP_T_BKGND || pcapparams->transp == OVL_TRANSP_B_BKGRD ) charpixel = ~charpixel;
					ga_indirect_reg_write ( v, G_CAPRAMDATA, charpixel );
					//PRINTK( "captn_param_load_516: ----- char = %c ( %i ) \t charpixel = 0x%x \n", 
					//											*( pcapparams->text + i ), ascii_index, charpixel );	// for debugging
				}
			}
			//strcpy ( loaded_param[pcapparams->window - 1].text, pcapparams->text );
			cmemcpy( loaded_param[pcapparams->window - 1].text, pcapparams->text, sizeof(pcapparams->text) );
			// load attributes for caption window into RAM.
			attr = (~pcapparams->transp) << 15;
			switch (pcapparams->color)		// refer to HW spec.
			{
			case OVL_WHITE:					// 0 -- white:   with 001
				attr = attr + 0x1240;		//			0+001+001+001+000000; 
				break;
			case OVL_BLACK:					// 1 -- black:   with 000
				attr = attr + 0x0000;		//			0+000+000+000+000000; 
				break;
			case OVL_RED:					// 2 -- red:     with 010
				attr = attr + 0x2480;		//			0+010+010+010+000000; 
				break;
			case OVL_GREEN:					// 3 -- green:   with 100
				attr = attr + 0x4900;		//			0+100+100+100+000000; 
				break;
			case OVL_BLUE:					// 4 -- blue:    with 011
				attr = attr + 0x36c0;		//			0+011+011+011+000000; 
				break;
			case OVL_YELLOW:				// 5 -- yellow:  with 101
				attr = attr + 0x5b40;		//			0+101+101+101+000000; 
				break;
			case OVL_CYAN:					// 6 -- cyan:    with 111
				attr = attr + 0x7fc0;		//			0+111+111+111+000000; 
				break;
			case OVL_MAGENTA:				// 7 -- magenta: with 110
				attr = attr + 0x6d80;		//			0+110+110+110+000000; 
				break;
			default:
				return;
			}
			ga_indirect_reg_write ( v, G_CAPRAMINDEX, ( pcapparams->window == 1 ) ? G_CAPRAMATTR_1 : G_CAPRAMATTR_2 );
			for ( i = 0; i < txtlen; i++ )
			{
				attrval = attr + ( ( pcapparams->window == 2 ) ? 32 : 0 ) + (USHORT) i;
				ga_indirect_reg_write ( v, G_CAPRAMDATA, attrval );
				//PRINTK( "captn_param_load_516: ----- window = %d : attrval = 0x%x ( i = %d )\n", 
				//												pcapparams->window, attrval, i );	// for debugging
			}
			loaded_param[pcapparams->window - 1].color  = pcapparams->color;
			loaded_param[pcapparams->window - 1].transp = pcapparams->transp;
			// done.
		} else {
			// set and load xstop = xstart.
			pos_x = (pcapparams->xstart >= 4) ? pcapparams->xstart >> 2 : 1;						// refer to HW spec:
			pos_x = (pos_x << 8) + ( (pcapparams->xstart >= 4) ? pcapparams->xstart >> 2 : 1 );		// divided by 4 for x_start and x_stop
			ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1X : G_CAPWINPOSN_2X, pos_x );
			loaded_param[pcapparams->window - 1].xstart = pcapparams->xstart;
			loaded_param[pcapparams->window - 1].xstop  = pcapparams->xstart;
			// set and load ystop = ystart.
			pos_y = pcapparams->ystart >> 1;						// refer to HW spec:
			pos_y = (pos_y << 8) + (pcapparams->ystart >> 1);		// divided by 2 for y_start and y_stop
			ga_indirect_reg_write ( v, ( pcapparams->window == 1 ) ? G_CAPWINPOSN_1Y : G_CAPWINPOSN_2Y, pos_y );
			loaded_param[pcapparams->window - 1].ystart = pcapparams->ystart;
			loaded_param[pcapparams->window - 1].ystop  = pcapparams->ystart;
		}
		break;
	default:
		return;
	}

	//PRINTK( "captn_param_load_516: Loaded once, with caption text = %s \n", pcapparams->text );	// for debugging

}

/////////////////////////////////////////////////////////////////////
// Calculate position parameters: x-stop and y-stop, for 616 caption.
// OVL_SCALE_XX not supported!
static void captn_pos_calc_616( PVPAR v, void *pcapprms )
{
	BUF_CAPTN_PARAM *pcapparams = pcapprms;
	int txtlen;

	txtlen = cstrlen( pcapparams->text );

    pcapparams->xstop = pcapparams->xstart + 8 * txtlen;
	pcapparams->ystop = pcapparams->ystart + 16;

	// trim xstop and ystop
	if ( pcapparams->xstop > MAX_XSTOP ) pcapparams->xstop = MAX_XSTOP;
	if ( v->attr_current[ATTR_video_system] == SYS_NTSC )
	{
		if ( pcapparams->ystop > MAX_YSTOP_NTSC )
            pcapparams->ystop = MAX_YSTOP_NTSC;
	}
    else
    {
		if ( pcapparams->ystop > MAX_YSTOP_PAL )
            pcapparams->ystop = MAX_YSTOP_PAL;
	}


}

////////////////////////////////////////////////////////////////////////
// Load caption parameters down to Regs in 616 Gate Array

static void captn_param_load_616( PVPAR v, void *pcapprms, ULONG action_id )
{
	BUF_CAPTN_PARAM *pcapparams = pcapprms;

    //dbgprint("%d %d %d %d %s\n", pcapparams->xstart, pcapparams->xstop,
    //   pcapparams->ystart, pcapparams->ystop, pcapparams->text);

	switch (action_id)
	{
	case CAPTN_window:
		break;
	case CAPTN_text:
        {
            int col, chr, c, d, vect;

            for(col = 0; col < 256; col++)            // count columns
              {
              chr = col / 8; d = col % 8;             // split to font table indexes
              c = pcapparams->text[chr];
              if(c == 0)
                  break;
              vect = fnt_616[c][d];                    // extract from font table
              i2c_write_616( v, I2CADDR_OVR_616 + 8, col, vect & 0xff);
              i2c_write_616( v, I2CADDR_OVR_616 + 2 + 8, col, (vect >> 8) & 0xff);
              }
        }
	case CAPTN_xstart:
	case CAPTN_xstop:
	case CAPTN_ystart:
	case CAPTN_ystop:
        i2c_write_616( v, I2CADDR_OVR_616, 1, pcapparams->xstart & 0xff);         // Register 1: xstart (lo)
        i2c_write_616( v, I2CADDR_OVR_616, 2, (pcapparams->xstart >> 8) & 0xff);  // Register 2: xstart (hi)

        i2c_write_616( v, I2CADDR_OVR_616, 3, pcapparams->ystart & 0xff);         // Register 3: ystart (lo)
        i2c_write_616( v, I2CADDR_OVR_616, 4, (pcapparams->ystart >> 8) & 0xff);  // Register 4: ystart (hi)

        i2c_write_616( v, I2CADDR_OVR_616, 5, pcapparams->xstop & 0xff);           // Register 5: xstop (lo)
        i2c_write_616( v, I2CADDR_OVR_616, 6, (pcapparams->xstop >> 8) & 0xff);    // Register 6: xstop (hi)

        i2c_write_616( v, I2CADDR_OVR_616, 7, pcapparams->ystop & 0xff);           // Register 7: yend (lo)
        i2c_write_616( v, I2CADDR_OVR_616, 8, (pcapparams->ystop >> 8) & 0xff);    // Register 8: yend (hi)
		break;

	case CAPTN_scale:
	case CAPTN_transp:
		break;

	case CAPTN_color:
        {
            int y, yf, cr, cb;
            switch (pcapparams->color)
            {
            case OVL_WHITE:					// 0 -- white
                y = 235; cb = 128; cr = 128;
                break;
            case OVL_BLACK:					// 1 -- black
                y = 20; cb = 128; cr = 128;
                break;
            case OVL_RED:					// 2 -- red
                y = 81; cb = 90; cr = 240;
                break;
            case OVL_GREEN:					// 3 -- green
                y = 145; cb = 54; cr = 34;
                break;
            case OVL_BLUE:					// 4 -- blue
                y = 41; cb = 240; cr = 110;
                break;
            case OVL_YELLOW:				// 5 -- yellow
                y = 210; cb = 16; cr = 146;
                break;
            case OVL_CYAN:					// 6 -- cyan
                y = 170; cb = 166; cr = 16;
                break;
            case OVL_MAGENTA:				// 7 -- magenta
                y = 106; cb = 202; cr = 222;
                break;
            default:
                y = 235; cb = 128; cr = 128;
                break;
            }
            if( y > 127 )
              yf = y + 20;
            else
              yf = y - 20;
            i2c_write_616( v, I2CADDR_OVR_616, 9, yf);              // Register 9: yfirst
            i2c_write_616( v, I2CADDR_OVR_616, 10, y);              // Register 10: ynext
            i2c_write_616( v, I2CADDR_OVR_616, 11, cr);             // Register 11: c1
            i2c_write_616( v, I2CADDR_OVR_616, 12, cb);             // Register 12: c2
    		break;
        }
	case CAPTN_on:
        i2c_write_616( v, I2CADDR_OVR_616, 0, pcapparams->on == CAPTION_ON);  // Register 0: mode
        break;
	default:
		return;
	}

}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////  VIDEO CONFIGURATION  ////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// These functions may be executed anytime to invoke immediate mode changes, including when streaming is active.

///////////////////////////////////////////////////////////
// Program colorbars enable/disable.

static void set_colorbars_enable( PVPAR v )
{
	eset( v, E_INPUTCTRL, ( v->attr_current[ATTR_video_colorbars_enable] != 0 ) ? 0x80 : 0x13 );
}

///////////////////////////////////////////////////////////
// Program video brightness level.

static void set_brightness( PVPAR v )
{
	dset( v, D_LUM_BRIGHTNESS, (UCHAR)v->attr_current[ATTR_video_brightness] );
}

///////////////////////////////////////////////////////////
// Program video contrast level.

static void set_contrast( PVPAR v )
{
	dset( v, D_LUM_CONTRAST, (UCHAR)v->attr_current[ATTR_video_contrast] );
}

///////////////////////////////////////////////////////////
// Program video saturation level.

static void set_saturation( PVPAR v )
{
	dset( v, D_CHROMA_SAT, (UCHAR)v->attr_current[ATTR_video_saturation] );
}

///////////////////////////////////////////////////////////
// Program video input and type.

#define CHAN		( v->attr_current[ATTR_video_chan] )
#define IS_CVID		( v->attr_current[ATTR_video_type] == CVIDEO )
#define IS_PLAY		( v->active_microcode == MPEG_PLAY )

// 516: SAA7113 decoder.
static void set_video_in_516( PVPAR v )
{
	dset( v, D_AIN_CTRL_1,									// 0x02: AIN control 1: video input.
		0xC0 | (											//	Activate both amplifier and anti-alias filter.
		IS_PLAY ? 0x05 : (									//	Play mode: select unused input to prevent sm interference from decoder syncs.
		IS_CVID ? CHAN : ( ( CHAN & 1 ) | 0x06 ) ) ) );		//	Rec mode: cvideo-select specified chan, svideo-limit to ch0-1 and select separate Y/C paths.

/*
	dset( v, D_LUM_CTRL,									// 0x09: Luminance control.
		0x39												//	1 (ntsc std) or 2 (pal std) video lines additional processing delay.
		| ( IS_CVID ? 0x00 : 0x80 )							//	C-trap or Y-comb filter: active for cvideo, bypassed for svideo.
		| ( IS_FILT ? 0x40 : 0x00 ) );						//	Luminance filter: on=low pass filt 3dB@1.9MHz, notch at 3.0MHz.
*/

	dset( v, D_LUM_CTRL, 0x40 );
}

// 616: SAA7114 decoder.
static void set_video_in_616( PVPAR v )
{
	dset( v, D_AIN_CTRL_1,									// 0x02: AIN control 1: video input:
		0xC0 |												//	Activate both amplifier and anti-alias filter.
		( IS_PLAY ? 0x03 :									//	In playback mode, select unused input to prevent interference from input syncs.
		( IS_CVID ? 0x00 : 0x06 ) ) );						//	In record mode, select fixed cvideo or svideo source path, as appropriate.

	dset( v, D_LUM_CTRL,									// 0x09: Luminance control.
		0x00 												//	1 (ntsc std) or 2 (pal std) video lines additional processing delay.
		| ( IS_CVID ? 0x40 : 0x80 )							//	C-trap or Y-comb filter: active for cvideo, bypassed for svideo.
		| ( IS_FILT ? 0x0E : 0x04 ) );						//	Luminance filter: on=low pass filt 3dB@1.9MHz, notch at 3.0MHz.
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////  GPIO FUNCTIONS  /////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////
// Read state of active-low GPIO pins: when input pulled down bit is set to 1.

// Fetch GPIO bits from GA.
static int gpio_read_516( PVPAR v )
{
	return ga_indirect_reg_read( v, G_GPIO ) & 3;
}

// Fetch GPIO bits from PCA9554<1:0>.
static int gpio_read_616( PVPAR v )
{
	return v->i2c_read( v, I2CADDR_IO_616, PCA_PORT_IN ) & PCA_MASK_DIN;
}

// Unified GPIO read.
int CORE_gpio_read( HBD board )
{
	PVPAR v = (PVPAR)board;
	int retval;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return 0;

	retval = v->gpio_read( v );

	SX16MOD_release_mutex( v->mutex_object );

	return retval;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Set states of active-low GPIO pins.  In case of 516, pin is asserted low when bit is set to 1.

// Write GPIO bits to GA.
static void gpio_write_516( PVPAR v, int data )
{
	// Write data directly to GPIO pins.
	ga_indirect_reg_write( v, G_GPIO, (USHORT)( data & 3 ) );
}

// Write GPIO bits to PCA9554 output pins.
static void gpio_write_616( PVPAR v, int data )
{
	// Preserve states of all PCA9554 I/O port bits that are not general-purpose outputs.
	int base_image = v->i2c_read( v, I2CADDR_IO_616, PCA_PORT_OUT ) & ~PCA_MASK_DOUT;

	// Write data to GPO pins (which are physically driven by PCA9554<3:2>) via I2C bus.
	v->i2c_write( v, I2CADDR_IO_616, PCA_PORT_OUT, base_image | ( ( data << PCA_BIT_DOUT ) & PCA_MASK_DOUT ) );
}

// Unified GPIO write.
void CORE_gpio_write( HBD board, int data )
{
	PVPAR v = (PVPAR)board;

	if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
		return;
	v->gpio_write( v, data );
	SX16MOD_release_mutex( v->mutex_object ); 
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////  DIAGNOSTICS  /////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
// Return TRUE if 516 hardware signature detected.

BOOL CORE_board_is_516( USHORT base_adrs )
{
	BOOL is516;

	// Detect 516 board's hardware signature, return BoardNotDetected if signature not valid.
	is516 = ( ( SX16MOD_read_io16( base_adrs + G_SIGNATURE ) & 0xFFF0 ) == 0x5160 );

	return is516;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Return the currently loaded microcode version number.

USHORT CORE_microcode_version( HBD board )
{
	PVPAR v = (PVPAR)board;
	return v->microcode_version;
}

/////////////////////////////////////////////////////////////////////////////
// Enable/disable video loopback in gate array.

static void set_stream_bypass_516( PVPAR v, BOOL enable )
{
	ga_indirect_reg_modify( v, G_RESETCONTROL, (USHORT)( ( enable != BYPASS_OFF ) ? RST_LOOPBACK : 0 ), (USHORT)( ~RST_LOOPBACK ) );
}

static void set_stream_bypass_616( PVPAR v, BOOL enable )
{
    // 616 is not able to bypass the SM.
}

// Exported, unified loopback enable.
void CORE_stream_bypass( HBD board, BOOL enable )
{
    PVPAR v = (PVPAR)board;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;

    v->set_stream_bypass( v, enable );

    SX16MOD_release_mutex( v->mutex_object ); 
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////  BOARD OPEN, RESET AND CLOSE  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Copy memory block.

static void mem_copy( void *dest, const void *source, int nbytes )
{
    register int    i;
    register char   *src = (char *)source;
    register char   *dst = (char *)dest;
    for ( i = 0; i < nbytes; i++ )
        *dst++ = *src++;
}

//////////////////////////////////////////////////////////////////////////
// SM reset.  Resets SM, audio/video PLLs, but not encoder/decoder.

#define MSEC_RESET_PLL_AND_SM   1       // Duration of PLL and SM resets asserted.
#define MSEC_RESET_SM_CONTINUE  1       // Duration of SM reset continuation after PLL reset is released.
#define MSEC_RESET_SM_STABILIZE 20      // Wait this long after SM reset before talking to it.
#define MSEC_RESET_GA_AUDIO     10      // Wait this long after SM reset before talking to it.

static void sm_reset_516( PVPAR v )
{
    // Assert the following resets: audio PLL, video PLL, SM hard reset.  This does not affect the state of the
    // gate array video loopback (516 only).  Hold these asserted for at least the 500us required by the SM.
    ga_indirect_reg_modify( v, G_RESETCONTROL, RST_HARD | RST_APLL | RST_VPLL, RST_LOOPBACK );
    SX16MOD_sleep( MSEC_RESET_PLL_AND_SM );

    // Negate audio PLL and video PLL resets.  Continue to assert the SM hard reset for another 500us, minimum.
    ga_indirect_reg_modify( v, G_RESETCONTROL, RST_HARD, RST_LOOPBACK );
    SX16MOD_sleep( MSEC_RESET_SM_CONTINUE );

    // Negate the SM hard reset.  We add a short delay after this so the SM's internal ARC processor has sufficient
    // time to run the microcode's init functions before we try to upload parameters to it.
    ga_indirect_reg_modify( v, G_RESETCONTROL, 0, RST_LOOPBACK );
    SX16MOD_sleep( MSEC_RESET_SM_STABILIZE );

    // additional reset for GA-audio: 0x08 <- 0, 0x8100 //
    ga_indirect_reg_write( v, G_AUDIOCONTROL, 0x0000 );
    SX16MOD_sleep( MSEC_RESET_GA_AUDIO );
    ga_indirect_reg_write( v, G_AUDIOCONTROL, 0x8100 );
    SX16MOD_sleep( MSEC_RESET_GA_AUDIO );
}

static void sm_reset_616( PVPAR v )
{
    // Assert the following resets: video PLL, SM hard reset.  Because the audio PLL reset is driven by SM's GPIO_1 (which
    // is pulled up during SM hard reset), it is not possible to satisfy the SM spec that requires the audio PLL reset to
    // be driven active during SM hard reset.  Consequently, we will refrain from asserting the audio PLL reset and hope
    // that it behaves properly even though it is never reset.  On the SAA7146, GPIO_2 controls the video PLL reset and
    // GPIO_1 controls the SM hard reset.  We now program GPIO<3:0> = ( in, 0, 0, in ) to force the video PLL and SM
    // hard resets active.  Hold the video PLL and SM resets asserted for at least the 500us required by the SM.
    SX16MOD_write_mem32( v->base_adrs + P_GPIO, /*0x00404000*/ 0x00404020 );
    SX16MOD_sleep( MSEC_RESET_PLL_AND_SM );

    // Negate video PLL reset.  This is done by programming GPIO<3:0> = ( in, 1, 0, in ).  Continue to assert the SM
    // hard reset for another 500us, minimum.
    SX16MOD_write_mem32( v->base_adrs + P_GPIO, /*0x00504000*/ 0x00504020 );
    SX16MOD_sleep( MSEC_RESET_SM_CONTINUE );

    // Negate the SM hard reset.  This is done by programming GPIO<3:0> = ( in, 1, 1, in ).  We add a short delay after this so
    // SM's internal ARC processor has sufficient time to run the microcode's init functions before we try to upload parameters to it.
    SX16MOD_write_mem32( v->base_adrs + P_GPIO, /*0x00505000*/ 0x00505020 );
    SX16MOD_sleep( MSEC_RESET_SM_STABILIZE );
}

////////////////////////////////////////////////////////////////////////////
// Program audio/video codecs to the runtime configuration settings.

static void program_av_codecs( PVPAR v )
{
#ifdef TEST_PCI_616
  return;
#endif
  // Program audio codec to runtime configuration.
    set_volume( v );            // Program audio volume.
    set_bass_treble( v );       // Program bass and treble levels.
    set_mute( v );              // Program audio muting enable.

    set_audio_agc_ingain( v );
    
    // Upload all runtime configuration parameters to the audio codec and video encoder and decoder.
    v->enc_config( v );
    v->dec_config( v );
}

/////////////////////////////////////////////////////////////////////////////////////////
// Bring board to the default module state.

static void board_reset( PVPAR v )
{
#ifdef TEST_PCI_616
  return;
#endif
    
  
    // DA: Oct 27, 2006.  board reset problem.
    // If MPEG application is called and then live video called at same time, then
    // this reset causes problems with the overlay or capture.  
    // DA: fixes overlay stop race condition when starting encoding app and capture app at same time
    // Workaround was previously to "move" the Overlay window to restart it.
    // Stop must be done because re-routed below.  If stop is not done here, then the capture/overlay window
    // will have very serious pink or garbage glitches.
    // capture is stopped but not restarted.  See comments in start_encode for details.
    if( v->product_id == 616 )
    {
        if( v->overlay_started || v->capture_started) 
        {
            // turn off the RPS, DMA while loading
            SX16MOD_write_mem32( v->base_adrs + P_MC1, DSBL_RPS);
            v->cap_frozen = 1;
        }
        if( v->clip_started)
        {
            dbgprint("pausing clip\n");
            SX16MOD_write_mem32( v->base_adrs + P_MC2, 1 << 5 | 1 << 21);
            // disable video dma2
            SX16MOD_write_mem32( v->base_adrs + P_MC1, 1 << 21);
        }

    }


    // If MPEG stream is running, halt the stream now and force a microcode upload during next set_mpeg_mode().
    set_stream_state( v, STREAM_HALTED );

    // Program audio/video codecs to the runtime configuration.
    program_av_codecs( v );

    // IS THIS NEEDED ???
//  ga_indirect_reg_write( v, G_CAPWINPOSN_1X, 0x5040 );    // Turn on subtitles.
//  ga_indirect_reg_write( v, G_CAPWINPOSN_1Y, 0x2010 );

    // If necessary, reset the SM and configure it for Record mode, enable gate array loopback.
    set_mpeg_mode( v, MPEG_RECORD );

    // Disable SM stream bypass (implemented on 516 only) so SM can output the video.
    v->set_stream_bypass( v, BYPASS_OFF ); //Was OFF

    if( v->product_id == 616)
    {
        // if vm overlay was started prior, restart it
        if(v->overlay_started)
        {
            SX16MOD_write_mem32( v->base_adrs + P_RPSADDR0, v->vpar_dma_cap.caprps);
            SX16MOD_write_mem32( v->base_adrs + P_RPSTO0, 0);                           // select port a
            SX16MOD_write_mem32( v->base_adrs + P_MC1, EN_DMA1_RPS0_OVL);
        }
        if( v->clip_started)
        {
            dbgprint("restarting clip\n");
            SX16MOD_write_mem32( v->base_adrs + P_MC2, 1 << 5 | 1 << 21 | 1 << 3 | 1 << 19);
            SX16MOD_write_mem32( v->base_adrs + P_MC1, 1 << 5 | 1 << 21);
        }
        v->cap_frozen = 0;
    }

}

// Public wrapper.
void CORE_board_reset( HBD board )
{
    PVPAR v = (PVPAR)board;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;
    board_reset( v );
    SX16MOD_release_mutex( v->mutex_object ); 
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Shut down hardware in preparation for driver shutdown.

static void destructor_516( PVPAR v )
{
    // Nothing to do...
}

static void destructor_616( PVPAR v )
{
    // Nothing to do...
}

void CORE_board_shutdown( HBD board )
{
    PVPAR v = (PVPAR)board;
    dbgprint("board shutboard\n");
    if( !v) {
        return;
    }
    // disable DMA, etc..
    if( v->product_id == 616)
    {
        // stop dma,etc
        SX16MOD_write_mem32(v->base_adrs + P_MC1, 0x00ff0000);
        // Invoke SAA7146 soft reset.
        SX16MOD_write_mem32( v->base_adrs + P_MC1, MC1_SOFT_RESET );
    }
    

    // Disable board interrupt
    v->disable_irq( v );
    v->shutdown_irq( v );
    // Reset all general-purpose I/O bits.
    v->gpio_write( v, 0 );

    // Do the model-dependent destruction for the board's object and hardware.
    v->destructor( v );

    // Halt the board and terminate its application connection.
    CORE_board_close( v );

    // Free the mutex object
    SX16MOD_delete_mutex( v->mutex_object );
    SX16MOD_delete_mutex( v->i2c_lock );

    // Free the VPAR structure.
    SX16MOD_free( v );
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Init the board's object and hardware.  Call this first, once only, from module-level init.
// Returns: 1 = successful, 0 = error.

static void constructor_516( PVPAR v )
{
    // Initialize board object's constants.
    v->i2caddr_dec          = I2CADDR_DEC_516;
    v->i2caddr_enc          = I2CADDR_ENC_516;
    v->L3bit_data           = L3BIT_DATA_516;
    v->L3mask_data          = 1 << L3BIT_DATA_516;
    v->L3mask_mode          = 1 << L3BIT_MODE_516;
    v->L3mask_clk           = 1 << L3BIT_CLK_516;
    v->lastaddr             = -1;                       // Set GA's Index Register image to illegal value so it will be init upon first use.

    // Initialize board object's Vfunc pointers.
    v->context_save         = context_save_516;
    v->context_restore      = context_restore_516;
    v->dec_config           = dec_config_516;
    v->enc_config           = enc_config_516;
    v->destructor           = destructor_516;
    v->gpio_read            = gpio_read_516;
    v->gpio_write           = gpio_write_516;
    v->i2c_is_busy          = i2c_is_busy_516;
    v->i2c_read             = i2c_read_516;
    v->i2c_write            = i2c_write_516;
    v->L3_read              = L3_read_516;
    v->L3_write             = L3_write_516;
    v->set_stream_bypass    = set_stream_bypass_516;
    v->set_video_in         = set_video_in_516;
    v->sm_mpeg_read         = sm_mpeg_read_516;
    v->sm_mpeg_write        = sm_mpeg_write_516;
    v->sm_reg_block_write   = sm_reg_block_write_516;
    v->sm_reg_read          = sm_reg_read_516;
    v->sm_reg_write         = sm_reg_write_516;
    v->sm_reset             = sm_reset_516;
    v->captn_pos_calc       = captn_pos_calc_516;
    v->captn_param_load     = captn_param_load_516;
    v->enable_irq           = enable_irq_516;
    v->disable_irq          = disable_irq_516;
    v->shutdown_irq         = shutdown_irq_516;
    v->product_id           = 516;

    // Initialize board object's pointer to parameter constants.
    v->attr                 = attr_516;

    // Initialize board object's pointer to caption parameter constants.
    v->captn_prm            = captn_prm_516;

    // Assert the encoder/decoder resets for the minimum required time (which is unknown, so we make a guess).
    ga_indirect_reg_modify( v, G_RESETCONTROL, RST_ENC_DEC, RST_LOOPBACK ); // Assert encoder/decoder resets.
    SX16MOD_sleep( 100 );                                                   // Hold for minimum time.
    ga_indirect_reg_modify( v, G_RESETCONTROL, 0, RST_LOOPBACK );           // Negate resets.

    // Wait for encoder/decoder to stabilize following the reset.
    SX16MOD_sleep( 20 );

    // Enable little endian byte ordering for SM register accesses.  This is the default endian configuration
    // that is in effect at all times except when accessing MPEG steam data during interrupt servicing.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_OFF );

    // Init the GA's audio clock steering.
    ga_indirect_reg_write( v, G_AUDIOCONTROL, 0 );
    
}

static void constructor_616( PVPAR v )
{
    // Initialize board object's constants.
    v->i2caddr_dec          = I2CADDR_DEC_616;
    v->i2caddr_enc          = I2CADDR_ENC_616;
    v->L3bit_data           = L3BIT_DATA_616;
    v->L3mask_data          = 1 << L3BIT_DATA_616;
    v->L3mask_mode          = 1 << L3BIT_MODE_616;
    v->L3mask_clk           = 1 << L3BIT_CLK_616;

    // Initialize board object's Vfunc pointers.
    v->context_save         = vfunc_dummy_v;
    v->context_restore      = vfunc_dummy_v;
    v->dec_config           = dec_config_616;
    v->enc_config           = enc_config_616;
    v->destructor           = destructor_616;
    v->gpio_read            = gpio_read_616;
    v->gpio_write           = gpio_write_616;
    v->i2c_is_busy          = i2c_is_busy_616;
    v->i2c_read             = i2c_read_616;
    v->i2c_write            = i2c_write_616;
    v->L3_read              = L3_read_616;
    v->L3_write             = L3_write_616;
    v->set_stream_bypass    = set_stream_bypass_616;
    v->set_video_in         = set_video_in_616;
    v->sm_mpeg_read         = sm_mpeg_read_616;
    v->sm_mpeg_write        = sm_mpeg_write_616;
    v->sm_reg_block_write   = sm_reg_block_write_616;
    v->sm_reg_read          = sm_reg_read_616;
    v->sm_reg_write         = sm_reg_write_616;
    v->sm_reset             = sm_reset_616;
    v->captn_pos_calc       = captn_pos_calc_616;
    v->captn_param_load     = captn_param_load_616;
    v->enable_irq           = enable_irq_616;
    v->disable_irq          = disable_irq_616;
    v->shutdown_irq         = shutdown_irq_616;
    v->product_id           = 616;

    // Initialize board object's pointer to parameter constants.
    v->attr                 = attr_616;

    // Initialize board object's pointer to caption parameter constants.
    v->captn_prm            = captn_prm_616;

    // Invoke SAA7146 soft reset.
    SX16MOD_write_mem32( v->base_adrs + P_MC1, MC1_SOFT_RESET );

    // Wait for the 7146 to stabilize before configuring it.
    SX16MOD_sleep( 100 );

    // Disable board's master interrupt.
    SX16MOD_write_mem32( v->base_adrs + P_INTEN, INT_OFF );

    // Note: 616 can't assert encoder/decoder resets under program control; these are asserted only during PCI hard reset.

    // Enable I2C and DEBI port pins, and enable DEBI transfers.
    SX16MOD_write_mem32( v->base_adrs + P_MC1, MC1_I2C_ENABLE | MC1_DEBI_ENABLE | MC1_TR_E_DEBI );

    // Initialize DEBI interface.
    SX16MOD_write_mem32( v->base_adrs + P_DEBICFG, DEBICFG_REG_ACCESS );
    SX16MOD_write_mem32( v->base_adrs + P_DEBIPAGE, DEBIPAGE_TABLE( DEBIPAGE_DISABLE, 0 ) );

    // Initialize the I2C interface ---------

    // Abort any I2C transaction that may be in progress.
    SX16MOD_write_mem32( v->base_adrs + P_I2CSTAT, I2C_CLKSEL | I2C_ABORT );
    SX16MOD_write_mem32( v->base_adrs + P_MC2, MC2_I2C_UPLD_CMD );
#ifndef TEST_PCI_616
    while ( ( SX16MOD_read_mem32( v->base_adrs + P_MC2 ) & MC2_I2C_UPLD_DONE ) == 0 );
#endif
    // Clear status register twice to reset all I2C error flags, per SAA7146 data sheet.
    SX16MOD_write_mem32( v->base_adrs + P_I2CSTAT, I2C_CLKSEL );
    SX16MOD_write_mem32( v->base_adrs + P_MC2, MC2_I2C_UPLD_CMD );
#ifndef TEST_PCI_616    
    while ( ( SX16MOD_read_mem32( v->base_adrs + P_MC2 ) & MC2_I2C_UPLD_DONE ) == 0 );
#endif
    SX16MOD_write_mem32( v->base_adrs + P_I2CSTAT, I2C_CLKSEL );
    SX16MOD_write_mem32( v->base_adrs + P_MC2, MC2_I2C_UPLD_CMD );
#ifndef TEST_PCI_616        
    while ( ( SX16MOD_read_mem32( v->base_adrs + P_MC2 ) & MC2_I2C_UPLD_DONE ) == 0 );
#endif
    // Initialize the PCA9554 8-bit I/O port.  Only bits 1 and 0 are
    // configured as inputs, per the following diagram of the port's bit functions:
    //
    //      +--------+--------+--------+--------+--------+--------+--------+--------+
    //      |    x   | L3CLK  | L3MODE | L3DATA | GPOUT1 | GPOUT0 | GPIN1  | GPIN0  |
    //      +--------+--------+--------+--------+--------+--------+--------+--------+
    //
    
    // Initialize states of all output bits on the PCA9554.
    i2c_write_616( v, I2CADDR_IO_616, PCA_PORT_OUT, L3MASK_616 );
    
    // Enable the output pins on the PCA9554.
    i2c_write_616( v, I2CADDR_IO_616, PCA_PORT_CONFIG, PCA_MASK_DIN );
}

static HBD board_init( int product_id, unsigned long base_adrs, int irq, PVOID dma_adrs_virt, ULONG dma_adrs_phys, int debug_flags )
{
    int i;
    PVPAR v;
    
    if( (product_id != 516) && (product_id != 616))
    {
        // probably have some stack calling problem with the compilation
        // return or risk segmentation fault later
        
        return NULL;
    }
    // Allocate the VPAR structure.
    v = (PVPAR)SX16MOD_malloc( sizeof(VPAR) );
    if ( v == NULL )
    {
        return NULL;
    }

    // Initialize model-independent parts of the board's object.
    v->base_adrs            = base_adrs;        // Save the base address of the application registers.
    v->irq                  = irq;              // Save the board's interrupt number.
    v->is_open              = FALSE;            // Board has not yet been opened.
    v->errflags             = 0;                // No errors logged.
    v->active_microcode     = MPEG_NONE;        // No microcode is loaded on the SM.
    v->stream_state         = STREAM_HALTED;    // MPEG stream flow is halted.
    v->debug_flags          = debug_flags;      // Set the debug message options.
    v->b_upload             = 0;

    // DMA buffer addresses (for 616 only.)
    v->dmabuf_virt          = (void *) dma_adrs_virt;   // Virtual address.
    v->dmabuf_phys          = dma_adrs_phys;            // Physical (DMA) address.
    v->pend_encode = 0;
    // Create mutex object
    v->mutex_object = SX16MOD_create_mutex();
    // i2c mutex
    v->i2c_lock = SX16MOD_create_mutex();
    if ( v->mutex_object == NULL )
    {
        SX16MOD_free( v );
        return NULL;
    }
    if ( v->i2c_lock == NULL )
    {
        SX16MOD_free( v );
        return NULL;
    }
    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
    {
//        SX16MOD_text_message( "x16: board_init: acquire mutex fail\n" );
        SX16MOD_delete_mutex( v->mutex_object );
        SX16MOD_free( v );
        return NULL;
    }
#ifndef WIN32
#ifdef SX16_BUILD_SRAY
    // RTP init
    v->rtp.h_rtp = NULL;
    v->rtp.ssrc = SX16MOD_get_jiffies() * SX16MOD_get_jiffies() + base_adrs;
    //PRINTK(" ssrc %d \n", v->rtp.ssrc);
    v->rtp.last_transfer = 0;
    v->rtp.transfer_count = 0;
    v->rtp.rawdata = 0;
#endif
    // write sensoray parameters
    v->TempBuf = NULL;
    v->PacHdr = NULL;
    v->fifo = NULL;
    v->SequenceNumber = 0;
    v->LostPackets = 0;
    for(i=0;i<2;i++)
    {
        v->last_sync_time[i] = 0;
        v->last_lostpacket_time[i] = 0;
        v->last_upload_time[i] = 0;
        v->last_dbgtime1[i] = 0;
        v->last_dbgtime2[i] = 0;
    }
    for(i=0;i<11;i++)
    {
        v->last_err[i] = 0;
    }
    v->dblBufferCount = 0;
    v->PacketCount = 0;
    v->write_step = 0;
    v->write_step_mux = 0;
    // end write sensoray parameters
#endif
    v->bCapInit =0;
    v->capture_started = 0;
    v->overlay_started = 0;
    v->clip_started = 0;

    // Do the model-dependent initialization for the board's object and hardware.
    switch ( product_id )
    {
        case 516:   constructor_516( v );   
                    break;
        case 616:   constructor_616( v );   
                    break;
        default:    SX16MOD_release_mutex( v->mutex_object );
                    SX16MOD_delete_mutex( v->mutex_object );
                    SX16MOD_release_mutex( v->i2c_lock );
                    SX16MOD_delete_mutex( v->i2c_lock );
                    SX16MOD_free( v );
                    // SX16MOD_text_message( "x16: board_init: invalid model number\n" );
                    return NULL;            // Illegal model number.
    }

    // Initialize the board object's standard default device configuration images.
    // These configuration parameters are allowed to change until the board is opened,
    // at which time they will be treated as constants.
    for ( i = 0; i < NUM_ATTR_IDS; i++ )
        v->attr_default[i] = v->attr[i].default_value;
    
    // Initialize runtime parameter images by setting them to their default module-level values.
    mem_copy( v->attr_current, v->attr_default, sizeof( v->attr_default ) );

    // Initialize the board object's caption default value.
    // The parameters are allowed to change until the board is opened,
    // at which time they will be treated as constants.
    for ( i = 0; i < NUM_CAPTN_IDS; i++ )
        v->captn_default[i] = v->captn_prm[i].default_value;
    
    // Initialize runtime caption parameters by setting them to their default module-level values.
    mem_copy( v->captn_current, v->captn_default, sizeof( v->captn_default ) );
    for ( i = 0; i < 2; i++ )
    {
        v->captn_parameters[i].window   = i+1;                                              // Caption window #.
        v->captn_parameters[i].xstart   = (USHORT)  v->captn_default[CAPTN_xstart];         // Caption x start.
        v->captn_parameters[i].xstop    = (USHORT)  v->captn_default[CAPTN_xstop];          // Caption x stop.
        v->captn_parameters[i].ystart   = (USHORT)  v->captn_default[CAPTN_ystart];         // Caption y start.
        v->captn_parameters[i].ystop    = (USHORT)  v->captn_default[CAPTN_ystop];          // Caption y stop.
        v->captn_parameters[i].scale    = (USHORT)  v->captn_default[CAPTN_scale];          // Caption scale.
        v->captn_parameters[i].color    = (int)     v->captn_default[CAPTN_color];          // Caption color.
        v->captn_parameters[i].transp   = (int)     v->captn_default[CAPTN_transp];         // Caption transparency.
//      sprintf( v->captn_parameters[i].text, "%c", (char) v->captn_default[CAPTN_text] );  // Caption text.
        num2txt( (char) v->captn_default[CAPTN_text], v->captn_parameters[i].text );

        v->captn_parameters[i].on       = (BOOL)    v->captn_default[CAPTN_on];             // Caption on/off.
    }

    v->vpar_dma_cap.capdma.hDma     = 0;
    v->vpar_dma_cap.capdma.RpsAddr  = NULL;

    v->clip_dma_cap.capdma.hDma     = 0;
    v->clip_dma_cap.capdma.RpsAddr  = NULL;

    // Initialize the audio interface.
    AudioInit( v );
    
    // Bring the board to its default state.
    board_reset( v );
    
    // Enable board interrupt
    if(v->irq)
    {
        //dbgprint("enabling IRQ %d!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n", v->irq);
        v->enable_irq( v );
    }

    SX16MOD_release_mutex( v->mutex_object );

    // Indicate success.
    return v;
}

// Public wrapper using default debug flags.
HBD CORE_board_init( ULONG product_id, ULONG base_adrs, ULONG irq, PVOID dma_adrs_virt, ULONG dma_adrs_phys )
{
    return board_init( product_id, base_adrs, irq, dma_adrs_virt, dma_adrs_phys, DFLT_DEBUG_FLAGS );
}

// Public wrapper using specified debug flags.
HBD CORE_board_init_debug( ULONG product_id, ULONG base_adrs, ULONG irq, PVOID dma_adrs_virt, ULONG dma_adrs_phys, ULONG debug_flags )
{
    return board_init( product_id, base_adrs, irq, dma_adrs_virt, dma_adrs_phys, debug_flags );
}


/////////////////////////////////////////////////////////////////////////////////////////
// Init board hardware and object to module-level's defaults.

int CORE_board_open( HBD board )
{
    PVPAR v = (PVPAR)board;

    // Can't open a board that is already open.
    //dbgprint("board open\n");
    if ( v->is_open ) {
        dbgprint("board already open\n");
        return FALSE;
    }

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return FALSE;

    // Initialize runtime parameter images by setting them to their default module-level values.
    mem_copy( v->attr_current, v->attr_default, sizeof( v->attr_default ) );

    // Reset SM and bring board to module's default state.
    board_reset( v );

    // Enable IRQ line.
    if( v->irq)
    {
        v->enable_irq( v );
    }

    // initialize the capture settings
    if ( v->product_id == 616 )
    {
        capture_init_params( v );
    }

#ifndef WIN32
    /* initialize write sensoray buffers */
    v->TempBuf = SX16MOD_malloc( TEMP_BUF_SIZE_LARGER);
    v->TempBufSize = TEMP_BUF_SIZE_LARGER;
    // min water mark to flush the temp buf
    v->TempBufHW = TEMP_BUF_SIZE_LARGER - (MIN_WRITE_SIZE_MUX*12);
    if (v->TempBuf == NULL) {
        dbgprint("temp buf too large, using smaller tempbuf\n");
        v->TempBuf = SX16MOD_malloc( TEMP_BUF_SIZE_LARGE);
        v->TempBufSize = TEMP_BUF_SIZE_LARGE;
        v->TempBufHW = TEMP_BUF_SIZE_LARGE - (MIN_WRITE_SIZE_MUX*12);
        if (v->TempBuf == NULL) {
            dbgprint("temp buf still too large, using smallest tempbuf\n");
            v->TempBuf = SX16MOD_malloc( TEMP_BUF_SIZE);
            v->TempBufSize = TEMP_BUF_SIZE;
            v->TempBufHW = TEMP_BUF_SIZE - (MIN_WRITE_SIZE_MUX*12);
            if (v->TempBuf == NULL) {
                dbgprint("out of memory error!\n");
                return FALSE;
            }
        }
    }
    v->PacHdr = SX16MOD_malloc( PACHEADER_SIZE);

    if (v->PacHdr == NULL) {
        return FALSE;
    }

    v->fifo = SX16MOD_malloc( FIFO_SIZE);

    if (v->fifo == NULL) {
        return FALSE;
    }

#endif
    // The board is now open for business.
    v->is_open = TRUE;

    CORE_reset_interrupt( v);
    SX16MOD_release_mutex( v->mutex_object ); 

    // Indicate success.
    return TRUE;
}

///////////////////////////////////////////////////////////////
// Halt the board and terminate its application connection.

int CORE_board_close( HBD board )
{
    PVPAR v = (PVPAR)board;

    if( v == NULL)
        return -1;

    // Can't close a board that is already closed.
    if ( !v->is_open ) {
        dbgprint("not open!\n");
        return FALSE;
    }

    // Disable IRQ line.
    v->disable_irq( v );
    v->active_microcode = MPEG_NONE;
#if 0
    if( SX16MOD_acquire_mutex( v->mutex_object ) < 0)
    {
        dbgprint("mutex failed!\n");
        return -1;
    }
#endif


    SX16MOD_release_threads( board );

#ifndef WIN32
#ifdef SX16_BUILD_SRAY
    CloseRTP( v->rtp.h_rtp );
    v->rtp.h_rtp = 0;
#endif

    /* remove temp buf */
    if (v->TempBuf)
    {
        SX16MOD_free( v->TempBuf);
        v->TempBuf = NULL;
    }

    if (v->PacHdr)
    {
        SX16MOD_free( v->PacHdr);
        v->PacHdr = NULL;
    }

    if (v->fifo)
    {
        SX16MOD_free( v->fifo);
        v->fifo = NULL;
    }


#endif
    // Disable IRQ line.
    v->disable_irq( v );

    // The board is now closed.
    v->is_open = FALSE;

    // Reset the current settings to their default values.
    mem_copy( v->attr_current, v->attr_default, sizeof( v->attr_default ) );

    // Halt stream if running, reset SM and return board to module's default state.
        // DA 01/18/2008.  Why reset?  changing this to halt because the next board open, it's reloaded anyways!
        //board_reset( v );
    set_stream_state( v, STREAM_HALTED  );

        v->capture_started = 0;
        v->overlay_started = 0;

    
    //SX16MOD_release_mutex( v->mutex_object ); 

    // Indicate success.
    return TRUE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////  GET STATUS OF VIDEO INPUT  ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////
// Return status.

static int in_status_get( PVPAR v )
{
  v->i2c_write( v, v->i2caddr_dec, D_OUT_CTRL_3_7113, 0 );
  return v->i2c_read( v, v->i2caddr_dec, D_STATUS ); 
}

// Public wrapper

int CORE_in_status_get( HBD board )
{
    PVPAR v = (PVPAR)board;
    ULONG retval;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return 0;

    retval = in_status_get( v );

    SX16MOD_release_mutex( v->mutex_object );

    return retval;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////  GET/SET OPERATIONAL ATTRIBUTES  //////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Map ATTR_stream_type and ATTR_mpeg into associated ATTR_mux_format.
// NOTE: DA: 10-13-04: changing MUX_FORMAT_SVCD to MUX_FORMAT_DVD for MPEG 2/3(streamtype 1).
//       RTP backback of MPEG 2/3 in SVCD is currently buggy.
// 10-18-04: changing MPEG 2-2/3 back to SVCD, but adding new MPEG format with DVD mux.
#define NUM_STREAM_TYPES    4
#define NUM_MPEGS           9
static const ULONG muxfmt[NUM_STREAM_TYPES][NUM_MPEGS] = {
    { MUX_FORMAT_DVD, MUX_FORMAT_SVCD, MUX_FORMAT_DVD, MUX_FORMAT_VCD, MUX_FORMAT_VCD, MUX_FORMAT_DVD, MUX_FORMAT_DVD, MUX_FORMAT_DVD, MUX_FORMAT_DVD },
    { MUX_FORMAT_VES, MUX_FORMAT_VES,  MUX_FORMAT_VES, MUX_FORMAT_VES, MUX_FORMAT_VES, MUX_FORMAT_VES, MUX_FORMAT_VES, MUX_FORMAT_VES, MUX_FORMAT_VES },
    { MUX_FORMAT_AES, MUX_FORMAT_AES,  MUX_FORMAT_AES, MUX_FORMAT_AES, MUX_FORMAT_AES, MUX_FORMAT_AES, MUX_FORMAT_AES, MUX_FORMAT_AES, MUX_FORMAT_AES },
    { MUX_FORMAT_TRANSPORT, 0, 0, 0, 0, 0, 0, 0, 0 }
};

/////////////////////////////////////////////////////////
// Return attribute.

static ULONG attr_get( PVPAR v, ULONG attr_id )
{
    // Return current setting.
    return v->attr_current[attr_id];
}

// Public wrapper.  Returns zero if attribute identifier is not legal.
ULONG CORE_attr_get( HBD board, ULONG attr_id )
{
    PVPAR v = (PVPAR)board;
    ULONG retval;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return 0;

    retval = ( attr_id < NUM_ATTR_IDS ) ? attr_get( v, attr_id ) : 0;

    SX16MOD_release_mutex( v->mutex_object ); 

    return retval;
}

////////////////////////////////////////////////////////////////////////
// Modify attribute's memory image, but don't write to codecs.
// Returns true if image was modified.

static int attr_set_image( PVPAR v, ULONG attr_id, ULONG value )
{
    int st;

    // Change parameter's default setting if board is closed.
    if ( !v->is_open )
    {
        v->attr_default[attr_id] = value;
        v->attr_current[attr_id] = value;
    }

    // Change parameter's runtime setting if board is open and we are not streaming or parameter is allowed to change while streaming.
    else if ( ( v->stream_state != STREAM_RUNNING ) || ( v->attr[attr_id].change_anytime ) )
        v->attr_current[attr_id] = value;

    // Abort if parameter can't be changed now.
    else
        return FALSE;

    // Modify any fundamental attributes that are associated with the target attribute.
    switch ( attr_id )
    {
    case ATTR_mpeg:

        // System format must always be set to Other in the standard operating modes.
        attr_set_image( v, ATTR_video_system_format, SYSFMT_OTHER );

        // Force ATTR_mux_format as appropriate for the new ATTR_video_picsize and the current ATTR_stream_type.
        attr_set_image( v, ATTR_mux_format, muxfmt[ (st = attr_get( v, ATTR_stream_type )) ][ value ] );

        //printk("ATTR_mpeg streamtype %ld value %ld \n", attr_get(v, ATTR_stream_type), value);
                
        // Set ATTR_video_picsize, which is the underlying fundamental attribute of ATTR_mpeg.
        if (st != STREAMTYPE_TRANSPORT)
        {
            attr_set_image( v, ATTR_video_picsize, value );
        }
        else
        {
            attr_set_image( v, ATTR_video_picsize, SIZE704_2 );
        }

        break;

    case ATTR_stream_type:

        // System format must always be set to Other in the standard operating modes.
        attr_set_image( v, ATTR_video_system_format, SYSFMT_OTHER );

        if (value == STREAMTYPE_TRANSPORT)
        {
            attr_set_image( v, ATTR_video_picsize, SIZE704_2 );
        }
        //printk("stream type %ld %ld \n", value, attr_get(v, ATTR_video_picsize));
        // Force ATTR_mux_format, which is the underlying fundamental attribute of ATTR_stream_type.
        attr_set_image( v, ATTR_mux_format, muxfmt[ value ][ attr_get( v, ATTR_video_picsize ) ] );

        break;
    }

    return TRUE;
}

// Public function: set attribute.
void CORE_attr_set( HBD board, ULONG attr_id, ULONG value )
{
    ATTR_ITEM   *a;
    PVPAR       v = (PVPAR)board;

    // Abort if attribute identifier is not legal.
    if ( attr_id >= NUM_ATTR_IDS )
    {
        return;
    }

    // Cache pointer to parameter's attributes.
    a = &( v->attr[attr_id] );

    // Clamp the new parameter value to its legal range.
    if ( value < a->min_legal_value )       value = a->min_legal_value;
    else if ( value > a->max_legal_value )  value = a->max_legal_value;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;

    // Attempt to modify the parameter's image; do nothing if image is not allowed to change now.
    if ( attr_set_image( v, attr_id, value ) )
    {
        // Change parameter if board is open.
        if ( v->is_open )
        {
            // Program affiliated devices via the fastest method.
            if ( a->change_anytime )
            {
                a->change_anytime( v );     // Invoke "quick change" on affiliated device(s).
            }
            else
            {
                program_av_codecs( v );     // Program all audio/video/mpeg codecs to the runtime configuration.
                sm_upload_params( v );      // Upload all runtime configuration parameters to the SM.
            }
        }
    }

    SX16MOD_release_mutex( v->mutex_object ); 
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////    GET/SET CAPTION ATTRIBUTES    //////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// subroutine for Return caption parameters.

static ULONG caption_get( PVPAR v, ULONG cap_attr_id )
{
    // Return runtime setting if board is open, or default setting if board is closed.
    // return v->is_open ? v->captn_current[cap_attr_id] : v->captn_default[cap_attr_id];
    if ( v->is_open ) 
    {
        if ( v->captn_current[CAPTN_on] )
        {
            switch (cap_attr_id)
            {
            case CAPTN_window:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].window;
            case CAPTN_xstart:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].xstart;
            case CAPTN_xstop:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].xstop;
            case CAPTN_ystart:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].ystart;
            case CAPTN_ystop:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].ystop;
            case CAPTN_scale:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].scale;
            case CAPTN_color:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].color;
            case CAPTN_transp:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].transp;
            case CAPTN_text:
                // Can't pass a kernel pointer to user space.
                // return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].text;
                return 0;
            case CAPTN_on:
                return (ULONG) v->captn_parameters[v->captn_current[CAPTN_window]-1].on;
            default:
                return 0;
            }
        } else {
            return v->captn_current[cap_attr_id];
        }
    } else {
        return v->captn_default[cap_attr_id];
    }
}

// Public wrapper.  Returns zero if attribute identifier is not legal.
ULONG CORE_caption_get( HBD board, ULONG cap_attr_id )
{
    PVPAR v = (PVPAR)board;
    ULONG retval;

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return 0;

    retval = ( cap_attr_id < NUM_ATTR_IDS ) ? caption_get( (PVPAR)board, cap_attr_id ) : 0;

    SX16MOD_release_mutex( v->mutex_object );

    return retval;
}


// Public function: set caption attribute.
void CORE_caption_set( HBD board, ULONG cap_attr_id, ULONG value )
{
    PVPAR       v = (PVPAR)board;
    CAPTN_PARAM *cp;

    // Abort if caption attribute identifier is not legal.
    if ( cap_attr_id >= NUM_CAPTN_IDS )
        return;

    // Cache pointer to parameter's attributes.
    cp = &( v->captn_prm[cap_attr_id] );

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return;

    if ( cap_attr_id == CAPTN_text )
    {
        // copy over text to captn_txt_current.
        cmemcpy( v->captn_txt_current, (char *) value, sizeof(v->captn_txt_current) );
    } 
    else 
    {
        // Clamp the new parameter value to its legal range.
        if ( value < cp->min_legal_value )      value = cp->min_legal_value;
        else if ( value > cp->max_legal_value ) value = cp->max_legal_value;
        // update current parameters.
        if ( cap_attr_id == CAPTN_window && v->captn_current[CAPTN_window] != value )
        {
            // save current parameters to the window operated last.
            v->captn_parameters[v->captn_current[CAPTN_window]-1].window    = v->captn_current[CAPTN_window];
            v->captn_parameters[v->captn_current[CAPTN_window]-1].xstart    = (USHORT)v->captn_current[CAPTN_xstart];
            v->captn_parameters[v->captn_current[CAPTN_window]-1].ystart    = (USHORT)v->captn_current[CAPTN_ystart];
            v->captn_parameters[v->captn_current[CAPTN_window]-1].scale     = (USHORT)v->captn_current[CAPTN_scale];
            v->captn_parameters[v->captn_current[CAPTN_window]-1].color     = v->captn_current[CAPTN_color];
            v->captn_parameters[v->captn_current[CAPTN_window]-1].transp    = v->captn_current[CAPTN_transp];
            cmemcpy( v->captn_parameters[v->captn_current[CAPTN_window]-1].text, v->captn_txt_current, sizeof(v->captn_txt_current) );
            v->captn_parameters[v->captn_current[CAPTN_window]-1].on        = v->captn_current[CAPTN_on];
            // update current over from board database.
            v->captn_current[CAPTN_window] = value;
            v->captn_current[CAPTN_xstart] = v->captn_parameters[v->captn_current[CAPTN_window]-1].xstart;
            v->captn_current[CAPTN_xstop]  = v->captn_parameters[v->captn_current[CAPTN_window]-1].xstop;
            v->captn_current[CAPTN_ystart] = v->captn_parameters[v->captn_current[CAPTN_window]-1].ystart;
            v->captn_current[CAPTN_ystop]  = v->captn_parameters[v->captn_current[CAPTN_window]-1].ystop;
            v->captn_current[CAPTN_scale]  = v->captn_parameters[v->captn_current[CAPTN_window]-1].scale;
            v->captn_current[CAPTN_color]  = v->captn_parameters[v->captn_current[CAPTN_window]-1].color;
            v->captn_current[CAPTN_transp] = v->captn_parameters[v->captn_current[CAPTN_window]-1].transp;
            cmemcpy( v->captn_txt_current, v->captn_parameters[v->captn_current[CAPTN_window]-1].text, sizeof(v->captn_txt_current) );
            v->captn_current[CAPTN_on]     = v->captn_parameters[v->captn_current[CAPTN_window]-1].on;
        }
        else
        {
            v->captn_current[cap_attr_id] = value;
        }
    }

    // Exit if board is closed.
    if ( !v->is_open )
    {
        SX16MOD_release_mutex( v->mutex_object ); 
        return;
    }

    if ( v->captn_current[CAPTN_on] == CAPTION_ON ) 
    {
        // Update:
        switch (cap_attr_id)
        {
        case CAPTN_window:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].window = v->captn_current[CAPTN_window]; break;
        case CAPTN_xstart:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].xstart = (USHORT)v->captn_current[CAPTN_xstart];
            v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] ); break;
        case CAPTN_xstop:
            v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] ); break;
        case CAPTN_ystart:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].ystart = (USHORT)v->captn_current[CAPTN_ystart];
            v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] ); break;
        case CAPTN_ystop:
            v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] ); break;
        case CAPTN_scale:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].scale = (USHORT)v->captn_current[CAPTN_scale];
            v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] ); break;
        case CAPTN_color:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].color = v->captn_current[CAPTN_color]; break;
        case CAPTN_transp:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].transp = v->captn_current[CAPTN_transp]; break;
        case CAPTN_text:
            cmemcpy( v->captn_parameters[v->captn_current[CAPTN_window]-1].text, v->captn_txt_current, sizeof(v->captn_txt_current) );
            //PRINTK( "CORE_caption_set: text = %s \n", v->captn_parameters[v->captn_current[CAPTN_window]-1].text );   // for debugging
            v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] ); break;
        case CAPTN_on:
            if( v->product_id == 616)
            {
                v->captn_parameters[v->captn_current[CAPTN_window]-1].on = v->captn_current[CAPTN_on]; 
                v->captn_parameters[v->captn_current[CAPTN_window]-1].window = v->captn_current[CAPTN_window];
                v->captn_parameters[v->captn_current[CAPTN_window]-1].xstart = (USHORT)v->captn_current[CAPTN_xstart];
                v->captn_parameters[v->captn_current[CAPTN_window]-1].ystart = (USHORT)v->captn_current[CAPTN_ystart];
                v->captn_parameters[v->captn_current[CAPTN_window]-1].scale = (USHORT)v->captn_current[CAPTN_scale];
                v->captn_parameters[v->captn_current[CAPTN_window]-1].color = v->captn_current[CAPTN_color];
                v->captn_parameters[v->captn_current[CAPTN_window]-1].transp = v->captn_current[CAPTN_transp];
                cmemcpy( v->captn_parameters[v->captn_current[CAPTN_window]-1].text, v->captn_txt_current, sizeof(v->captn_txt_current) );
                v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] );
            }
            else
            {
                if ( v->captn_parameters[v->captn_current[CAPTN_window]-1].on == CAPTION_OFF ) 
                {
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].on = v->captn_current[CAPTN_on]; 
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].window = v->captn_current[CAPTN_window];
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].xstart = (USHORT)v->captn_current[CAPTN_xstart];
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].ystart = (USHORT)v->captn_current[CAPTN_ystart];
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].scale = (USHORT)v->captn_current[CAPTN_scale];
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].color = v->captn_current[CAPTN_color];
                    v->captn_parameters[v->captn_current[CAPTN_window]-1].transp = v->captn_current[CAPTN_transp];
                    cmemcpy( v->captn_parameters[v->captn_current[CAPTN_window]-1].text, v->captn_txt_current, sizeof(v->captn_txt_current) );
                    v->captn_pos_calc( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1] );
                }
            }
            break;
        default:
            SX16MOD_release_mutex( v->mutex_object ); 
            return;
        }

        // Load new parameters down to HW:
        v->captn_param_load( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1], cap_attr_id );

    }

    if( v->product_id != 616)
    {
        if ( v->captn_current[CAPTN_on] == CAPTION_OFF && v->captn_parameters[v->captn_current[CAPTN_window]-1].on == CAPTION_ON ) 
        {
            // Update:
            v->captn_parameters[v->captn_current[CAPTN_window]-1].on = v->captn_current[CAPTN_on];
            // Load new parameters down to HW:
            v->captn_param_load( v, &v->captn_parameters[v->captn_current[CAPTN_window]-1], cap_attr_id );
        }
    }

    SX16MOD_release_mutex( v->mutex_object ); 
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////  STATISTICS FUNCTIONS  /////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Return number of mpeg KB transferred to/from internal mpeg buffer since stream started.

ULONG CORE_get_stat_KB( HBD board )
{
    PVPAR v = (PVPAR)board;
    // Fetch the mpeg byte count.
    return v->KB_transferred;
}

int CORE_get_frame_rate (HBD board)
{
    PVPAR v = (PVPAR)board;

    return IS_NTSC ? 30 : 25;
}


#ifndef WIN32
#ifdef SX16_BUILD_SRAY
/* RTP is not supported outside of the 2416 */

/////////////////////////////////////////////////////////////////////////////
// Function:    ReadStreamParse
// Parameters:  vp - video parameter structure
//              StreamBuffer - buffer to place MPEG data
// Returns:     Number of headers parsed on success, -1 on error
// Description: This function reads MPEG data from the MPEG encoder and
//              generates a table of headers
//
int ReadStreamRTPParse( HBD board, int RTP_handle )
{
    PVPAR v = (PVPAR)board;
    int i;
    unsigned short  ad = v->base_adrs + G_DATAIO;

    //PRINTK("rtp stream parse\n");
    if ( RTP_handle == 0 || RTP_handle == -1 )
    {
        //PRINTK("sx16core: ReadStreamRTPParse: RTP handle\n");
        return -1;
    }

    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
    {
        //printk("acquire mutex err\n");
        return 0;
    }
    
    // If we have reached the end of the dma transfer cycle...
    if ( v->num_dma_bytes >= DFLT_DMA_SIZE )
    {
        // If dma flag is set, reset transfer count to zero.
        if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) != 0 )
        {
            //printk("reseting dma bytes\n");
            v->num_dma_bytes = 0;
        }
    }

    // If dma flag is not set. make sure interrupt is enabled and exit.
    if ( ( v->sm_reg_read( v, SM_INTINFO_REG ) & SMSTAT_DMA_REQUEST ) == 0 )
    {
        //PRINTK("sx16core: stream_read: dma flag not set, bytes %d\n",
        //bytes_transfered);
        // Enable interrupt, if present.
        //printk("dma %d\n", v->num_dma_bytes);
        if ( v->irq )
            v->sm_reg_write( v, SM_INTCTRL_REG, SMICTL_DMA_ENABLE );
        SX16MOD_release_mutex( v->mutex_object );
        return 0;
    }
    // If dma flag is set, disable further interrupts.
    else if ( v->irq )
    {
        //printk("irq\n");
        v->sm_reg_write( v, SM_INTCTRL_REG, 0 );
    }

    if ( v->num_dma_bytes >= DFLT_DMA_SIZE )
    {
        // DMA_Request must have just gone high, return
        SX16MOD_release_mutex( v->mutex_object );
        return 0;
    }
    
    v->context_save (v);                //need to do that before SM access ???
    
    // Switch to big endian byte ordering for SM MPEG stream portal accesses.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_ON );
    // Set up GA's indirect address register so it points to the SM's mpeg data port.
    ga_set_indirect_adrs( v, SM_DRAM_REG_516 );

    i=0;
     
    // Transfer data in CHUNK_SIZE chunks into RTP buffer...
    while ( i < ( (CHUNK_SIZE / 32)) )
    {
        // Parse 32 bytes of MPEG data into RTP buffer
        ParseRTP_IO_32( RTP_handle, ad );
        // Update counters
        i++;
        v->num_dma_bytes += 32;
    }
    // Update counters

    // Restore little endian byte ordering for SM register accesses.
    ga_indirect_reg_write( v, G_MISC, MISC_SM_BSWAP_OFF );

    // Restore the original hardware context.
    v->context_restore( v );

    // release mutex
    SX16MOD_release_mutex( v->mutex_object );
    // return total number of bytes read
    return i;
    
}

int CORE_rtp_plus (HBD board, unsigned char *RawBuffer)
{
    PVPAR v = (PVPAR)board;
    v->rtp.rawdata = RawBuffer;
    if( v->rtp.h_rtp)
    {
        RTPSetRTPRawBuffer ( v->rtp.h_rtp, v->rtp.rawdata);
    }
    return 0;   //success
}


/* Read RTP method - Attempts to read 5 RTP packets from the 512.
The actual number of packets read is returned. This function blocks
untils at least five packets are available. If this is the first
read the encoder is initialized and started.
NOTE: buffer should be checked and VALID before this is called 
on Error, an negative number is returned 
*/
int CORE_read_rtp( HBD board, unsigned char *buffer )
{
#define TIMEOUT_INITIAL_RTP 5000  // 5 seconds max waiting for data
#define RTP_PACKETS_DESIRED 5
    PVPAR   v = (PVPAR)board;
    int     rc = 0;
    int     FrameRate;
    int     MpegType;
    int     attr_mpeg;
    //  must check counter for the RTP parsing loop.  Otherwise, we may overflow RTP raw data buffer.
    // This can happen when 2416 sitting idle (encoding RTP and data), and PAL input plugged in DURING the RTP parsing loop
    // without changing input from NTSC to PAL.
    // Please see CR2 of app-241x-web project
    unsigned int BytesRead;  // each time RTP parse called

    // internal loop counter for RTP.  used to prevent lockup when changing
    // from NTSC to PAL or vice-versa.
    unsigned long loopcount2 = 0;
    // Start the encoder if it is not already running.
    if ( !CORE_stream_is_encoding( board ) )
    {
        //PRINTK("START ENCODE %d\n", board_num);
        CORE_start_encode( board );
#ifdef STREAM_TIME
        stream_time[board_num] = SX16MOD_time_now();
#endif
        // Open the RTP system
        v->rtp.h_rtp = OpenRTP( v->rtp.ssrc, (void *) &SX16MOD_do_gettimeofday, (void *)&SX16MOD_get_jiffies, (void *)&SX16MOD_malloc,
            (void *)&SX16MOD_free, SX16MOD_get_hz(), NULL );
        RTPSetRTPRawBuffer ( v->rtp.h_rtp, v->rtp.rawdata);
        // wait for data before proceeding
        SX16MOD_sleepq( board, TIMEOUT_INITIAL_RTP );
    }


    // DA: open RTP system if no handle. 
    if (v->rtp.h_rtp == 0)
    {
        // Open the RTP system
        v->rtp.h_rtp = OpenRTP( v->rtp.ssrc, (void *) &SX16MOD_do_gettimeofday, (void *)&SX16MOD_get_jiffies, (void *)&SX16MOD_malloc,
            (void *)&SX16MOD_free, SX16MOD_get_hz(), NULL );
        RTPSetRTPRawBuffer ( v->rtp.h_rtp, v->rtp.rawdata);
    }




    InitGetRTPPackets( v->rtp.h_rtp, buffer + 4, RTP_PACKETS_DESIRED);
    FrameRate = CORE_get_frame_rate( board );
    attr_mpeg = CORE_attr_get( board, ATTR_mpeg );
    if ( (attr_mpeg == SIZE176_1) || (attr_mpeg == SIZE352_1)) // not picsize
    {
        MpegType = 1;
    }
    else
    {
        MpegType = 2;
    }

    BytesRead = 0;
    while((( rc = GetRTPPackets( v->rtp.h_rtp, FrameRate, MpegType ) ) == 0) )
    {
        loopcount2=0;
        while ( ( rc = ReadStreamRTPParse( board, v->rtp.h_rtp ) ) <= 0 )
        {
            loopcount2++;
            if ( rc < 0 )
            {
                return -1;
            }
            if( loopcount2 > 5 ) 
            {
                SX16MOD_text_message( "read RTP timeout\n" );
                return -1;
            }
            // in blocking mode, wait for more data(reduces CPU)
            //  added a timeout( 5 second max)
            // 
            // Necessary to fix bug when changing webpage settings from
            // NTSC to PAL with a PAL input connected
            //  System may lock-up without the timeout.
            // This is a separate bug from the overflow RTP problem, but
            // related to the same CR.
            SX16MOD_sleepq( board, 1000);  
        }
        BytesRead += rc;    //this is not the number of bytes, but rather 1/32 of that
        // this check is mandatory, because if PHYSICAL input
        // is changed from NTSC to PAL or vice-versa while in this loop,
        // we could overflow the raw data buffer
        
        if(  BytesRead > ((2*32768)/32) )
        {
            SX16MOD_text_message( "read RTP overflow. Possible change of input\n" );
            return 0;
        }
    }
    
    // Save number of packets
    *((int *) buffer) = rc;
    // Save number of bytes read
    v->rtp.last_transfer = BytesRead + 4;
    // Increment transfer count
    v->rtp.transfer_count++;
    //  PRINTK("rc %d, BytesRead %d\n", rc, BytesRead);

    return rc;
}


int CORE_get_rtp_handle( HBD board)
{
    PVPAR   v = (PVPAR)board;
    return v->rtp.h_rtp;
}

#endif
#endif /* ifndef WIN32 (RTP code)*/



// Note, this Capture DMA approach works for Linux, but should ideally be changed for the WDM driver.
// not for 64 bit...
// See DDK help
//  MmGetPhysicalAddress returns the physical address that corresponds to the given virtual address.
//  Do not use this routine to obtain physical addresses for use with DMA operations. 
//  For information about the proper techniques for performing DMA operations, see Adapter Objects and DMA.

/* begin new capture functions for 616 */
/* capture settings */

// get sx16 specific capture format
int CORE_capture_get_fmt( HBD board, void *val )
{
    PVPAR v = (PVPAR)board;
    if( v->b_upload)
    {
        SX16MOD_acquire_mutex_nonint(v->mutex_object);
    }
    else
    {
        if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
            return -1;
    }
    //cmemcpy( val, (void *) &v->cap_pixfmt, sizeof (struct vidcap_pix_format));    ????????
    *(struct vidcap_pix_format *)val = v->cap_pixfmt;
    //dbgprint("get format %d %d\n", v->cap_pixfmt.width, v->cap_pixfmt.height);
    SX16MOD_release_mutex( v->mutex_object );
    return 0;
}


// set sx16 specific capture format
int CORE_capture_set_fmt(HBD board, void *vformataddr, int set)
{
     PVPAR   v = (PVPAR)board;
    struct vidcap_pix_format *vformat_new = (struct vidcap_pix_format *) vformataddr;
    struct vidcap_pix_format *vformat;
    int err = 0;

    if( v->b_upload)
    {
        SX16MOD_acquire_mutex_nonint(v->mutex_object);
    }
    else
    {
        if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
            return -1;
    }

    /* no changes here.  we accept the format.  changes are done in try_fmt */
    vformat = &v->cap_pixfmt;
    vformat = vformat_new;
    SX16MOD_release_mutex( v->mutex_object );
    return err;
}


// constants for scaler
#define N_VS_NORMS  10

static struct { int cy, dcgy; } vs_norm [N_VS_NORMS] =
 {
   {0x0100, 0}, {0x0102, 1}, {0x0300, 1}, {0x0106, 2}, {0x0502, 2},
   {0x0708, 2}, {0x0F00, 2}, {0x011E, 3}, {0x031C, 3}, {0x0738, 3}
 };


// initialize the 616 capture RPS program

int CORE_capture_init(HBD board, int height, int width, int bpp, int fourcc)
{
    int lines, pixels;
    int inlpf, inppl, interlaced, yacm, yacl, ypo, ype, xpsc, xacm;
    int pfy, pfuv, dcgy, cy, dcgx, matrix, outformat;
    SX_DMA *dma;
    int *ptvirt, ptphys;
    int *RPS;                   // virtual address of current RPS command
    int pitch;
    unsigned int xsci, ysci;
    int vflip = 0;
    //    int prot_addr;
    unsigned long page_size = SX16MOD_get_PAGESIZE();
    struct vidcap_pix_format *vformat;
    int targetlines;

    PVPAR   v = (PVPAR)board;

    if( page_size != 4096 )
    {
        dbgprint( "wrong page size %d - must be 4096\n", page_size );
        return -1;
    }

    if( v->bCapInit)
    {
        SX16MOD_write_mem32( v->base_adrs + P_MC1, DSBL_RPS);   // disable RPS's
        v->bCapInit = 0;
        v->capture_started = 0;
    }

    if( v->b_upload)
    {
        SX16MOD_acquire_mutex_nonint(v->mutex_object);
    }
    else
    {
        if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
            return -1;
    }

    vformat = &v->cap_pixfmt;
    dma = &v->vpar_dma_cap.capdma;

    inlpf = IS_NTSC ?  SYS_YACTIVE_MAX_NTSC / 2 : SYS_YACTIVE_MAX_PAL / 2;
    inppl = IS_NTSC ?  SYS_XACTIVE_MAX_NTSC : SYS_XACTIVE_MAX_PAL;

    if( v->overlay_started || v->capture_started) {
        //dbgprint("overlay or capture already started. stop first\n");
        SX16MOD_release_mutex( v->mutex_object );
        return -1;
    }

    pixels = width;
    lines = height;


    if( IS_NTSC) {
        // don't allow invalid scaling
        if( lines == 288)
        {
            dbgprint("info: 288 is for PAL half size, use height 240 if you want NTSC half size\n");
        }
    }

    switch (fourcc) {
    case VIDCAP_PIX_FMT_RGB332:
        /*  8  RGB-3-3-2     */
            bpp = 1;
            matrix = 2; outformat = 0x7;
            break;
    case VIDCAP_PIX_FMT_RGB555:
        /* 16  RGB-5-5-5     */
            bpp = 2;
            matrix = 2; outformat = 0x3;
            break;
    case VIDCAP_PIX_FMT_RGB565:
        /* 16  RGB-5-6-5     */
            bpp = 2;
            matrix = 2; outformat = 0x0;
            break;
    case VIDCAP_PIX_FMT_RGB555X:
        /* 16  RGB-5-5-5 BE  */
            bpp = 2;
            matrix = 2; outformat = 0x4;
            break;
    case VIDCAP_PIX_FMT_RGB24:
        /* 24  RGB-8-8-8     */
        dbgprint("rgb24\n");
            bpp = 3;
            matrix = 2; outformat = 0x1;
            break;
    case VIDCAP_PIX_FMT_RGB32:
        /* 32  RGB-8-8-8-8   */
    case VIDCAP_PIX_FMT_BGR32:
        /* 32  RGB-8-8-8-8   */
            bpp = 4;
            matrix = 2; outformat = 0x2;
            break;
    case VIDCAP_PIX_FMT_GREY:
        /*  8  Greyscale     */
            bpp = 1;
            matrix = 0; outformat = 0x6;
            break;
    case VIDCAP_PIX_FMT_UYVY:
            bpp = 2;
            matrix = 0;
            outformat = 0;
            break;
    case VIDCAP_PIX_FMT_YVU420:
            bpp = BPP_32;
            matrix = 0;
            outformat = 0x0a;
            break;
    default:
            bpp = 3;
            matrix = 2; outformat = 0x1;
            break;
    }

    //dbgprint("pixels %d lines %d bufbase %x left %d top %d\n", pixels, lines, bufbase, left, top);

    vformat->width = pixels;
    vformat->height = lines;
    

    SX16MOD_write_mem32( v->base_adrs + P_MC1, DSBL_RPS);                       // disable RPS's

    if(dma -> hDma)
    {
        dma -> hDma = 0;
    }

    if(dma->RpsAddr)
    {
        SX16MOD_free_dma( board, dma->RpsSize, (unsigned long *) dma->RpsAddr, (unsigned long ) dma->RpsPhAddr);
        dma->RpsAddr = NULL;
        dma->RpsSize = 0;
        dma->RpsPhAddr = 0;
    }

    
    // Changing "malloc" to consistent dma allocation for PCI
    //dma -> RpsAddr = SX16MOD_malloc( page_size * 2 ); // pages: 1 for RPS + 1 for page table
    dma->RpsSize = page_size * 2;
    SX16MOD_alloc_dma( board, dma->RpsSize, (unsigned long **)&dma->RpsAddr, (unsigned long *) &dma->RpsPhAddr);

    if(dma->RpsAddr == NULL)
    {
        dbgprint("kmalloc ERROR\n");
        SX16MOD_release_mutex( v->mutex_object );
        return -1; // (-EFAULT);
    }
    RPS = dma->RpsAddr;
    // Page table starts at the 2nd page boundary in the RPS
    ptvirt = dma -> PtAddr = (void *)(((long)dma -> RpsAddr / page_size + 1) * page_size);  // 2nd page

    // Physical start of RPS
    //dbgprint("dma addresses CPU %p, phys %ld\n", dma->RpsAddr, dma->RpsPhAddr);
    v->vpar_dma_cap.caprps = dma -> RpsPhAddr;

    // Physical start of PT
    ptphys = dma ->PtPhAddr;
    pitch = vformat->bytesperline * 2;  //-pixels * bpp;

    interlaced = (vformat -> field == VIDCAP_FIELD_INTERLACED);

    if(!interlaced)
    {
        //dbgprint("not interlaced\n");
        pitch /= 2;
        yacm = 1;
    }
    else
    {
        lines /= 2;
        //dbgprint("interlaced\n");
        yacm = 0;
    }

    if(vflip) {
        pitch = -pitch;
    }

    xacm = inppl / pixels >= 4;

    if(xacm)
    {
        xpsc = inppl / pixels / 4;
        if(xpsc > 1)
            dcgx = 2;
        else
            dcgx = 1;
        xsci = inppl * 1024 / (xpsc + 1) / pixels;
    }
    else
    {
        xpsc = 0;
        dcgx = 1;
        //1024 * NIP / NOP
        xsci = inppl * 1024 / pixels; 
    }

    //dbgprint("xacm %d  xpsc %d  xsci %d\n", xacm, xpsc, xsci);
    if(yacm)
    {
        ysci = 1024 - 1024 * lines / inlpf;
        if(ysci > 1023)
            ysci = 1023;
        yacl = inlpf / lines - 1;
        pfy = 0; pfuv = 0;
        if(yacl >= N_VS_NORMS)
            yacl = N_VS_NORMS - 1;
        dcgy = vs_norm[yacl].dcgy;
        cy = vs_norm[yacl].cy;
        ypo = ype = (ysci + 15) / 16;
    }
    else
    {
        //ysci = inlpf * 1024 / lines - 1024;
        //if(ysci > 1023)
        // ysci = 1023;
        // no scaling
        ysci = 0;
        yacl = 0;
        pfy = 0; pfuv = 0;
        dcgy = 0;
        cy = 1; //0xff;
        //ype = ysci / 16;
        //ypo = ype + ysci / 64;
        ypo = ysci /16;
        ype = 3*ypo/2;
    }
    
    //dbgprint("xsci %d ysci %d yacl %d ypo %d dcgy %d cy %x\n", xsci, ysci, yacl, ypo, dcgy, cy);
    // BEGIN RPS program capture
    *RPS++ = RPS_CLRSIGNAL | RPS_HS | RPS_HT | RPS_E_FIDA | RPS_VBIA | RPS_SIG2 | RPS_O_FIDA | RPS_VBIB;
    *RPS++ = RPS_SETSIGNAL | RPS_SIG0  ;                  // signals RPS0 active
    // source line counter threshold. 
    *RPS++ = RPS_LDREG | (2 << 8) | (P_LCTS0 >> 2);
   
//~~    *RPS++ = inlpf + 16;  //~~!!+4

    targetlines = inlpf;// - 8;
    *RPS++ = ((targetlines) << 16) + (targetlines);  //~~!!+4
    *RPS++ = 0;

    *RPS++ = RPS_LDREG | (6 << 8) | (P_BASEODD1 >> 2);  // BaseODD1,BaseEven1,ProtAddr1,Pitch1,PageMap,Format regs
    // Note: these addresses are set in set_dma function in hw.cpp
    // ADDRESS 1 (Base Odd)
    *RPS = 0;//base_odd;
    // save the RPS address
    dma->rps_base_odd = RPS;
    RPS++;
    
    //ADDRESS 2 (Base Even)
    *RPS = 0;// base_even;
    dma->rps_base_even = RPS;
    RPS++;

    // ADDRESS 3 (protection address)
    *RPS = 0;//prot_addr;                                 // protectionaddr;
    dma->rps_prot_addr = RPS;
    RPS++;

    // ADDRESS 4 (pitch)
    *RPS = 0;//pitch;
    dma->rps_pitch = RPS;
    RPS++;
    // ADDRESS 5 (page map)
    *RPS = 0; //(ptphys & 0xfffff000); // | DMA_MAP | DMA_NOSWAP;
    dma->rps_ptphys = RPS;
    RPS++;
    // ADDRESS 6 (format reg)
    *RPS = ((inlpf + 4) << 16) | inppl;     //lines at the input of HPS ~~used 17 at some point
    dma->rps_numlines = RPS;
    RPS++;
    dma->rps_waitcond = NULL;
    *RPS++ = RPS_UPLOAD | 0x664;
    *RPS++ = RPS_LDREG | (1 << 8) | (P_MC2 >> 2);
    *RPS++ = UPLD_TR_E_1;
    *RPS++ = RPS_LDREG | (1 << 8) | (P_MC2 >> 2);
    *RPS++ = UPLD_DMA1;
    if(interlaced) 
    {
//      *RPS++ = RPS_PAUSE | RPS_VBIA;
        *RPS++ =  RPS_PAUSE | RPS_O_FIDA;   //~~ Could be anything: odd or even field, or VBIA
    }
    else
    {
       *RPS++ =  RPS_PAUSE | RPS_E_FIDA;
    }

    *RPS++ = RPS_LDREG | (2 << 8) | (P_MC1 >> 2);
    *RPS++ = UPLD_TR_E_1;


    *RPS++ = 0;
    *RPS++ = RPS_PAUSE | RPS_HT;            //A.K.: wait for target events (scaler output hor lines)
    *RPS++ = RPS_CLRSIGNAL | RPS_HT;

    if( interlaced ) 
    {
        *RPS++ = RPS_PAUSE | RPS_HT ;       //wait again
    }
    // Set SIG2 for the polling thread( if non-recommended polling method(no irqs) is used for V4L)
    *RPS++ = RPS_SETSIGNAL | RPS_SIG2;
    *RPS++ = RPS_LDREG | (1 << 8) | (P_MC1 >> 2);          //stop acquisition
    *RPS++ = UPLD_TR_D_1;
    *RPS++ = RPS_IRQ;
    // Clear SIG0 indicating that the program is stopped
    *RPS++ = RPS_CLRSIGNAL | RPS_SIG0;
    *RPS++ = RPS_STOP;

    // Note: See Sensoray forum for comments on D1SYNC.  Also see comments above on CL chip and VS period
//    SX16MOD_write_mem32( v->base_adrs + P_D1SYNC, (5<<24) | (5 << 8)); //D1_INIT_CAPTURE_HALF); //interlaced ? D1_INIT_CAPTURE_FULL : D1_INIT_CAPTURE_HALF); //  /*0x0 << 24*/);             //0x050; Init 7146 D1 interface
    SX16MOD_write_mem32( v->base_adrs + P_D1SYNC, interlaced ? D1_INIT_CAPTURE_FULL : D1_INIT_CAPTURE_HALF);
    SX16MOD_write_mem32( v->base_adrs + P_D1VIDEO, IS_NTSC ? 0 : (1 << 10));  //0x054; ~~may req mod for PAL (see 613) DA. 6/26/07 invert field PAL
    SX16MOD_write_mem32( v->base_adrs + P_PCIBTV,  BURST_INIT);           //0x048; burst, threshold
    // vertical offset after v_sync; A.K. 6/12/07 - increased from 7 to 18 to eliminate black band on top ***PAL*** 
    // PAL increased to 23 from 12. D.A. 6/26/07
    SX16MOD_write_mem32( v->base_adrs + P_HPSCTL, (0 << 30) | (0 << 28) | (((IS_NTSC ? 18 : 23) << 12))); // 6, 12
    SX16MOD_write_mem32( v->base_adrs + P_BCSCTL, BCS_INIT);              //0x070;
    SX16MOD_write_mem32( v->base_adrs + P_CKEY, 0);
    SX16MOD_write_mem32( v->base_adrs + P_HPSFORM, matrix << 30 | outformat << 24 | interlaced << 3);     // 78 colorformat
    SX16MOD_write_mem32( v->base_adrs + P_HPSHPS, xacm << 17 | xpsc << 18 | dcgx << 27);
    SX16MOD_write_mem32( v->base_adrs + P_HPSHFS, xsci << 12 | (IS_NTSC ? 0 : 10 ));           // 6c; A.K. 6/12/07 - added hor offset (4) to get rid of black vert band on the left ***PAL***(D.A. 6/26/07 10 for PAL)
    SX16MOD_write_mem32( v->base_adrs + P_HPSVS, 0);//~~yacm << 31 | ysci << 21 | yacl << 15 | ypo << 8 | ype << 1);  // 60
    SX16MOD_write_mem32( v->base_adrs + P_HPSVSG, 0xff);//~~pfy << 28 | pfuv << 24 | dcgy << 16 | cy);       // 64
#ifdef DEBUG_SCALER
    // Note if change the above values, change the debug output below
    dbgprint("xsci %d xacm %d xpsc %d dcgx %d \n", xsci, xacm, xpsc, dcgx);
    dbgprint("ysci %d yacm %d yacl %d ypo %d ype %d \n", ysci, yacm,yacl, ypo, ype);
    dbgprint("pfy %d pfuv %d dcgy %d cy %d\n", pfy,pfuv, dcgy, cy);

    dbgprint("P_D1SYNC %04x\n", (5<<24)); //D1_INIT_CAPTURE_HALF); //interlaced ? D1_INIT_CAPTURE_FULL : D1_INIT_CAPTURE_HALF); //  /*0x0 << 24*/);             //0x050; Init 7146 D1 interface
    dbgprint("P_D1VIDE %04x\n", IS_NTSC ? 0 : (1 << 10));                    //0x054; ~~may req mod for PAL (see 613) DA. 6/26/07 invert field PAL
    
    dbgprint("P_PCIBTV %04x\n", BURST_INIT);           //0x048; burst, threshold
    
    dbgprint("P_HPSCTL %04x\n",  (0 << 30) | (0 << 28) | (((IS_NTSC ? 7 : 12) << 12)));
    dbgprint("P_BCSCTL %04x\n",  BCS_INIT);
    dbgprint("P_CKEY   %04x\n", 0);
    dbgprint("P_HPSFORM %04x\n",  matrix << 30 | outformat << 24 | interlaced << 3);     // 78 colorformat
    dbgprint("P_HPSHPS %04x\n", xacm << 17 | xpsc << 18 | dcgx << 27);
    dbgprint("P_HPSHFS %04x\n", xsci << 12);           // 6c
    dbgprint("P_HPSVS %04x\n", yacm << 31 | ysci << 21 | yacl << 15 | ypo << 8 | ype << 1);  // 60
    dbgprint("P_HPSVSG %04x\n", pfy << 28 | pfuv << 24 | dcgy << 16 | cy);       // 64  
#endif

    v->bCapInit = 1;
    SX16MOD_release_mutex( v->mutex_object );
    return 0;
}


int CORE_get_dma_vformat( HBD board, SX_DMA **ppDma, struct vidcap_pix_format **ppVformat)
{
    PVPAR   v = (PVPAR)board;
    *ppDma =  &v->vpar_dma_cap.capdma;
    if( ppVformat) {
        *ppVformat = &v->cap_pixfmt;
    }
    return 0;
}




// start DMA capture for single frame
// assumes RPS program already ready(capture_init)
// will restart the next RPS program.
int CORE_capture_start(HBD board, int bLock)
{
    PVPAR   v = (PVPAR)board;
    if( bLock)
    {
        if( v->b_upload)
        {
            SX16MOD_acquire_mutex_nonint(v->mutex_object);
        }
        else
        {
            if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
                return -1;
        }
    }
    if( !v->bCapInit)
    {
        //dbgprint("capture not initialized\n");
        if( bLock)
            SX16MOD_release_mutex( v->mutex_object );

        return -1;
    }

    if( v->cap_frozen)
    {
        //dbgprint("capture frozen. not starting\n");
        if( bLock)
            SX16MOD_release_mutex( v->mutex_object );
        return -1;
    }
    //dbgprint("capture start2\n");
#ifdef DEBUG_VIDEOBUF
    dbgprint("capture start\n");
    dbgprint("MC1 %x\n", SX16MOD_read_mem32( v->base_adrs + P_MC1));
    dbgprint("MC2 %x\n", SX16MOD_read_mem32( v->base_adrs + P_MC2));
    dbgprint("PSR %x\n", SX16MOD_read_mem32( v->base_adrs + P_PSR));
    dbgprint("SSR %x\n", SX16MOD_read_mem32( v->base_adrs + P_SSR));
    dbgprint( "caprps %d\n", v->vpar_dma_cap.caprps);
#endif

    SX16MOD_write_mem32( v->base_adrs + P_RPSADDR0, v->vpar_dma_cap.caprps);
    SX16MOD_write_mem32( v->base_adrs + P_RPSTO1, 0);                        // select port B
    SX16MOD_write_mem32( v->base_adrs + P_MC1, EN_RPS0);
    v->capture_started = 1;
    if( bLock)
       SX16MOD_release_mutex( v->mutex_object );

    return 0;
}

int CORE_capture_stop( HBD board)
{
    PVPAR   v = (PVPAR)board;
    //SX_DMA  *dma = &v->vpar_dma_cap.capdma;
    // stop the RPS    
    SX16MOD_write_mem32( v->base_adrs + P_MC1, DSBL_RPS); 
    SX16MOD_write_mem32( v->base_adrs + P_MC2, 0xf8000000); // clear all the signals
    // capture may still be running.  (Q's are running RPS)
    // set capture_started = 0 in capture_close
    //v->capture_started = 0;
    return 0;
}



// return true if frame captured
int CORE_capture_get_status(HBD board)
{
    PVPAR   v = (PVPAR)board;
    int retval;
#if 0
    // protected by videoq s_lock 
    if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
        return -1;
#endif
    retval = SX16MOD_read_mem32(v->base_adrs + P_MC2);
    if( retval & STATUS_MC2_SIG2)
    {
        // clear signal 2
        SX16MOD_write_mem32(v->base_adrs + P_MC2, 0x20000000);
        return 1;
    }
    else return 0;
}



int CORE_capture_get_MC1(HBD board)
{
    PVPAR   v = (PVPAR)board;
    int retval;
    retval = SX16MOD_read_mem32(v->base_adrs + P_MC1);
    return retval;
}


int CORE_capture_get_MC2(HBD board)
{
    PVPAR   v = (PVPAR)board;
    int retval;
    retval = SX16MOD_read_mem32(v->base_adrs + P_MC2);
    return retval;
}



int CORE_capture_get_PSR( HBD board) 
{
    int retval;
    PVPAR   v = (PVPAR)board;
    retval = SX16MOD_read_mem32( v->base_adrs + P_PSR);
    return retval;
}

int CORE_capture_get_SSR( HBD board)
{
    int retval;
    PVPAR   v = (PVPAR)board;
    retval = SX16MOD_read_mem32( v->base_adrs + P_SSR);
    return retval;
}


// stop DMA and free memory
int CORE_capture_close(HBD board)
{
    PVPAR   v = (PVPAR)board;
    SX_DMA *dma;

    if( v == NULL )
        return -1;

    if( v->mutex_object == NULL) {
        return -1;
    }

    if( SX16MOD_acquire_mutex( v->mutex_object ) < 0)
    {
    dbgprint("capture close failed!\n");
        return -1;
    }

    if( !v->bCapInit)
    {
        dbgprint("capture close not initialized\n");
        SX16MOD_release_mutex( v->mutex_object);
        return -1;   
    }

    dma = &v->vpar_dma_cap.capdma;
    v->capture_started = 0;
    if(dma == NULL) {
        //dbgprint("dma null close\n");
        SX16MOD_release_mutex( v->mutex_object );
        return -1;
    }
    SX16MOD_write_mem32( v->base_adrs + P_MC1, DSBL_RPS);                       // disable RPS's

    if(dma -> hDma)
    {
        dma -> hDma = 0;
    }


    v->bCapInit = 0;
    SX16MOD_release_mutex( v->mutex_object );
    return 0;
}


// initialize the capture and vmoverlay settings
static void capture_init_params( PVPAR v )
{
    // board specific
    v->vpar_dma_cap.capbuf = 0;
    v->vpar_dma_cap.caprps = 0;
    v->vpar_dma_cap.capdma.UserAddr = 0;
    v->vpar_dma_cap.capdma.hDma = 0;
    v->vpar_dma_cap.capdma.RpsAddr = 0;
    v->vpar_dma_cap.capdma.PtAddr = 0;
    v->vpar_dma_cap.capdma.RpsPhAddr = 0;
    v->vpar_dma_cap.capdma.PtPhAddr = 0;
    v->vpar_dma_cap.capdma.Bytes = 0;
    v->vpar_dma_cap.capdma.Pages = 0;

    v->clip_dma_cap.capbuf = 0;
    v->clip_dma_cap.caprps = 0;
    v->clip_dma_cap.capdma.UserAddr = 0;
    v->clip_dma_cap.capdma.hDma = 0;
    v->clip_dma_cap.capdma.RpsAddr = 0;
    v->clip_dma_cap.capdma.PtAddr = 0;
    v->clip_dma_cap.capdma.RpsPhAddr = 0;
    v->clip_dma_cap.capdma.PtPhAddr = 0;
    v->clip_dma_cap.capdma.Bytes = 0;
    v->clip_dma_cap.capdma.Pages = 0;

    // Init capture format structure
    v->cap_pixfmt.top = 0;
    v->cap_pixfmt.left = 0;
    v->cap_pixfmt.width = SYS_XACTIVE_MAX_NTSC;
    v->cap_pixfmt.height = SYS_YACTIVE_MAX_NTSC;
    v->cap_pixfmt.pixelformat = VIDCAP_PIX_FMT_UYVY;//BGR24;
    v->cap_pixfmt.field = VIDCAP_FIELD_INTERLACED;
    v->cap_pixfmt.bytesperline = 3 * SYS_XACTIVE_MAX_NTSC;
    v->cap_pixfmt.sizeimage = SYS_XACTIVE_MAX_NTSC * SYS_YACTIVE_MAX_NTSC * 3;
    v->cap_pixfmt.colorspace = VIDCAP_COLORSPACE_SRGB;

    // Init vmoverlay format structure
    v->ovr_pixfmt.capability = 0;
    v->ovr_pixfmt.flags = 0;
    v->ovr_pixfmt.width = 256;
    v->ovr_pixfmt.height = 192;
    v->ovr_pixfmt.pixelformat = VIDCAP_PIX_FMT_BGR24;
    v->ovr_pixfmt.field = VIDCAP_FIELD_INTERLACED;
    v->ovr_pixfmt.bytesperline = 1024;
    v->ovr_pixfmt.sizeimage = 256 * 192 * 3;
    v->ovr_pixfmt.colorspace = VIDCAP_COLORSPACE_SRGB;
}

int CORE_pend_encode( HBD board)
{
    PVPAR v = (PVPAR) board;
    v->pend_encode = 1;
    return 0;
}

// set the capture buffer
int CORE_capture_set_capbuffer( HBD board, void *val)
{
    PVPAR v = (PVPAR) board;
    if( v->b_upload)
    {
        SX16MOD_acquire_mutex_nonint(v->mutex_object);
    }
    else
    {
        if ( SX16MOD_acquire_mutex( v->mutex_object ) < 0 )
            return -1;
    }
    v->vpar_dma_cap.capdma.UserAddr = val;
    SX16MOD_release_mutex( v->mutex_object );
    return 0;
}



#ifndef WIN32

/****************************************************************/
// BEGIN write sensoray code

// Struct's used for sync'ing to stream during playback
typedef struct {
    unsigned int            pattern;                    //pattern, bytes in reverse order     0xb3010000
    int                     length;                     //valid bytes in pattern
    int                     fbytes;                     //# of bytes already found
    int                     reset;                      //1 if fbytes is reset at least once in current call
} SEARCH;

typedef struct {
    unsigned short          word;                       //2-byte buffer; bytes in stream order
    int                     bytes;                      //valid bytes in buffer
} GETWORD;

typedef struct {
    char                    buf[16];
    int                     bytes;
    int                     reqbytes;
} GETBYTES;

typedef struct {
    SEARCH                  search;
    union {
        GETWORD             getword;
        GETBYTES            getbytes;
    } get;
} PARSE;






int CORE_get_write_params( HBD board, write_param_t *pParam)
{
    PVPAR v = (PVPAR) board;
    int i;
    pParam->SequenceNumber = v->SequenceNumber;
    pParam->LostPackets = v->LostPackets;
    for(i=0;i<11;i++)
    {
        pParam->last_error[i] = v->last_err[i];
    }
    for(i=0;i<2;i++)
    {
        pParam->last_packetlost_time[i] = v->last_lostpacket_time[i];
        pParam->last_sync_time[i] = v->last_sync_time[i];
        pParam->last_upload_time[i] = v->last_upload_time[i];
        pParam->last_dbgtime1[i] = v->last_dbgtime1[i];
        pParam->last_dbgtime2[i] = v->last_dbgtime2[i];

    }
    pParam->PacketCount = v->PacketCount;
    pParam->dblBufferCount = v->dblBufferCount;
    return 0;
}

/** this function resets the state machine for sensoray write and sensoray write mux
*/
void CORE_reset_write_step( HBD board )
{
    PVPAR v; 
    if( board == NULL)
        return;
    v = (PVPAR) board;
    v->write_step      = 0;
    v->write_step_mux  = 0;
    return;
}


#endif



